# Configuration file for IMACS makefile system
# Included by most makefiles in the system
#
# Define the system kernel name
syskn := $(shell uname -s)
ifeq "$(syskn)" "Linux"
    sysvn := $(shell uname -r | cut -d. -f1,2)
#    ifeq "$(sysvn)" "2.6"
#        syskn := "Linux"
#    else
#        syskn := "Redhat"
#    endif
endif
ifeq "$(syskn)" "Darwin"
    sysvn := $(shell uname -p)
# For Darwin, sysvn is "powerpc" or "i386" (platform)
    sysvx := $(shell uname -r | cut -d. -f1,2)
endif
#- syspf := $(shell uname -p)
#- syspf := $(shell kspn)
syspf := $(shell $(BASE)/kspn)
#- syshome := $(shell echo $HOME)
syshome := $(shell echo ${HOME})
#
# Define the project directory locations
#
# BASE is the master directory for the project, and should be
#  defined in each individual makefile before including this.
#
#  Directories where subroutine and header files are found
UTILS    = $(BASE)/utils
MASKG    = $(BASE)/progs
#
#  The name of the local link library
LIBNAME  = $(BASE)/libimacs.a
#
# Define the components used outside their home directories
FITS     = $(UTILS)/fits
BTREE    = $(UTILS)/btree
IOUTILS  = $(UTILS)/ioutils
KDCUTIL  = $(UTILS)/kdcutil
STATPAK  = $(UTILS)/statpack
#MASKDAT  = $(UTILS)/maskdat
MASKDAT  = $(MASKG)/maskdat
MGUTILS  = $(MASKG)/mgutils
# MGUTILS  = $(UTILS)/mgutils
OPTICS   = $(MASKG)/optics
# OPTICS   = $(UTILS)/optics
#OPTUTIL  = $(UTILS)/optutils
OPTUTIL  = $(MASKG)/optutils
CXT      = $(MASKG)/cxtfiles
FEATS    = $(MASKG)/mgfeats
#
#  Include Files definitions ---------------------------------------------
#  We could alternately add -I<directory> and define file names...
# IFILES  = -DIOUTILH="\"$(IOUTILS).h\"" -DMGUTILH="\"$(MGUTILS).h\"" -DKDCUTILH="\"$(KDCUTIL).h\"" -DOPTUTILH="\"$(OPTUTIL).h\"" -DMASKDATH="\"$(MASKDAT).h\"" -DSTATPAKH="\"$(STATPAK).h\"" -DFEATSH="\"$(FEATS).h\"" -DBTREEH="\"$(BTREE).h\""
# Removed  -DOPTICSH="\"$(OPTICS).h\"" 
#
# Above replaced 08/05-06 by the following copying from imacs/config.mk
#
IFILES  = -I$(UTILS) -DIOUTILH="\"ioutils.h\"" -DMGUTILH="\"$(MGUTILS).h\"" -DKDCUTILH="\"kdcutil.h\"" -DOPTUTILH="\"$(OPTUTIL).h\"" -DMASKDATH="\"$(MASKDAT).h\"" -DSTATPAKH="\"statpak.h\"" -DFEATSH="\"$(FEATS).h\"" -DBTREEH="\"btree.h\"" 
#
# General system flags -------------------------------------------------

# Defines the local subroutine library for the link steps
LOCLIB  = -L$(BASE) -limacs

# test
#CC 	= gcc
#CFLAGS	= -DSOLARIS  # -I /usr/openwin/include

#uncomment for SOLARIS
# CC 	= cc
# CFLAGS	= -DSOLARIS -v -xCC -I$(CXT) # -I /usr/openwin/include
# OPT	=  # -O
# LIBS	= $(LOCLIB) -lm -lposix4 -lnsl -lsunmath -lsocket -lpthread
# LIBX	= -lX11 -L$(CXT) -lcxt # -L/usr/openwin/lib
# cc options:
#  -s = strip symbol table from executable
#  -xCC accepts // comments, -lsunmath adds "nint" and others.
# See ~/source/makefile for that and btree.c stuff.

#  Definitions for all systems
CC	= gcc
#LIBS    = $(LOCLIB) -lm -lpthread -lcurl -lcrypto #
# lcurl needed for ioutils using libcurl to read FTP file
# lcrypto is needed (at least on Darwin) for openssl programs
# Above is normal, below needed for xmlrpc things, now standard
# XMLRPC  = $(MASKG)/xmlrpc  # placeholder - that's not where it is
XMLRPC  = $(syshome)/local/$(syspf)/lib
IXMLRPC = $(syshome)/local/include
# Above were:  (but trailing spaces were taken literally)
#XMLRPC  = $(syshome)/local/$(syspf)/lib  # /usr/local/lib
#IXMLRPC = $(syshome)/local/include  # /usr/local/include
#LIBXMLRPC = -L$(XMLRPC)  -lkxmlrpc_client -lkxmlrpc_util -lkxmlrpc
#- LIBRPC	= -L$(XMLRPC) -lxmlrpc_client -lxmlrpc_util -lxmlrpc #
LIBRPC  = -L$(XMLRPC) -lxmlrpc_client -lxmlrpc -lxmlrpc_util -lxmlrpc_xmlparse -lxmlrpc_xmltok
# The above actually works, and is very sensitive to the order in which
# the libraries are searched.
LIBS    = $(LOCLIB) -lm -lpthread -lcurl -lcrypto # $(LIBRPC) #
# We place LIBRPC only where it is actually needed in the makefile
#
LIBX    = -L/usr/X11R6/lib -lX11
LIBC	= -L$(CXT) -lcxt

#  Special definition for version support
VERfile := mgvers.dat

# Definitions for Linux systems
ifeq  "$(syskn)" "Linux"
 CFLAGS  = -DLINUX -DSYS_$(syskn) -Wall -DFONTSIZE=18 -I$(CXT) $(IFILES) -I$(IXMLRPC)
# {Formerly had..} # -I /usr/openwin/include
 OPT     =  -g # -O
# PGLIBS	= -L/usr/lib/pgplot -lpng -lz -lcpgplot -lpgplot
#  The png library is non-standard here...
# PGLIBS	= -L/usr/lib/pgplot /usr/lib/libpng.so.3 -lz -lcpgplot -lpgplot
 PGLIBS	= -lpgplot -lcpgplot -lpng
# Above changed 08/10-22 per PMC; -lX11 is defined separately above.

else

# Definitions for Mac-OS-X systems
 ifeq "$(syskn)" "Darwin"

# Using -g in mgutils.c causes an internal compiler error; temporatily
# remove it.  Try it again later, or if debugging is needed.
  OPT   =  # -g # -O
#  Special pgplot stuff for shimura:

  PGDIR = /Users/clardy/pgplot
  PGLIBS	= ${PGDIR}/libpgplot.a ${PGDIR}/libcpgplot.a
  CFLAGS  = -DLINUX -DSYS_$(syskn) -DNEED_SINCOS -Wall -DFONTSIZE=18 -I$(CXT) $(IFILES) -I$(IXMLRPC) -DCPGPLT="\"$(PGDIR)/cpgplot.h\""
# Fix up the definition of LIBS, as the general one was insufficient.
# LIBRPC	= -lxmlrpc_client -lxmlrpc_util -lxmlrpc #  -L/sw/lib # -lxml2 #
LIBS    = $(LOCLIB) -lm -lpthread -lcurl -lcrypto # only when needed $(LIBRPC) #


 else
  $(warning  Definitions for $(syskn) not known)
 endif
endif
# ----------------------------------------------------------------------
