# Generalized makefile statements
#
#  Definitions for most symbols have been made already, either in
#  the config.mk file in this directory, or in the individual 
#  makefile which includes these statements.
#
# This file slightly modified for Linux (GNU make) use.
#
# targets --------------------------------------------------------------

all:: $(TARGET)

# To add an application, add it to definition of TARGET in the
# individual makefile, and to BUILD_TARGET if it is to be built
# using the usual object suffix rule.  Define dependency in the
# makefile for each component of TARGET.

# To add a subroutine file, add it to the library dependencies
#  below, similar to the existing ones.

#  Build (link, actually) target applications from their object files:
#- $(BUILD_TARGET) : $$@.o
#$(BUILD_TARGET) : $(BUILD_TARGET).o
#	$(CC) -o $@ $@.o  $(LIBS) $(LIBX) 
#	@echo $@ linked
# This is how applications are finally linked
# Local subroutines are included from the local library via LIBX

# The object file is built from the C source by a suffix rule, below.

rebuild recompile :
	@make all FRC=force_rebuild
# Recompile is done by setting FRC non-null; and using
# macro $(FRC) in all dependencies; here those with $(HEADERS)
# and others explicitly including $(FRC).  Not using touch
# allows the true date to remain on header files.

# Dummy target for above
force_rebuild:

clean:
	@ rm -f *.o  core core.*  dummy.*

#howto:
#	@cat  $(BASE)/how2make.txt

# suffix rules ---------------------------------------------------------
.SUFFIXES  :  .c .o .a .ln .lx .sstest
# the .sstest above is for the test commands below.

.c.o :
	$(CC) $(CFLAGS) $(OPT) -c $<
# This is similar to the default suffix rule, but includes
# additional OPT macro.
# -c causes no link, object file output

# Following rule should never be used, it is probably flawed:
.o :
	$(CC) -o $@  $< $(LIBS) $(LIBX)
	@echo $@ linked by suffix rule o
# The old form below is flawed, including both $@ and $@.o
#	$(CC) -o $@ $@.o $< $(LIBS) $(LIBX)
#	@echo $@ linked by suffix rule
# This is how applications are finally linked, the $< above
# includes the object files upon which it has been said to depend

# Following is for default programs not otherwise mentioned
.c :
	$(LINK.c) -o $@  $< $(LIBS) $(LIBX)
	@echo $@ linked by suffix rule c
# Similar to default rule; LINK.C adds OPT, and definition of LDLIBS
# is also supplied above.  Add echo to indicate where it came from.
# Same as rule in make.rules usually.
# Rule is used to compile simply from a xxx.c file, with no further
# dependencies
# Copied from the rule in ~/imacs/general.mk

# -- Taking default .c.a rule, which is approximately this:
.c.a :
	$(CC) $(CFLAGS) $(OPT) -c -o $% $<
	ar rv $@ $%
	ar s  $@
	@ echo "$% entered by suffix rule."
#	rm -f $%
# -- Modifying .c.a rule to omit removing object file
# -- The IFILES is now included in CFLAGS by config.mk
# -- Modified 05/05-12 KDC for MacOS-X to add "ar s" step to
#   re-index the library which is not automatic there.

# For using splint on Fedora machines, use this:
LINT = splint
SPLINTFLAGS = -skipisoheaders -warnposix +trytorecover

LINTFLAGS = $(IFILES) -I$(CXT) -DLINUX -nestcomment -nullassign $(SPLINTFLAGS)

.c.ln:
	$(LINT.c) $<

# The above should use splint in a moderately similar fashion to the
# lint on solaris.  May need to add more flags to suppress some
# non-needed checks.
# NOTE that this fails on macx (Darwin) which has no splint...

# Lint-like functionality

#LDFLAGS = $(IFILES) -I$(CXT) $(OPT)
#LCFLAGS = -Wall -Wextra

#LXFLAGS = $(LDFLAGS) $(LCFLAGS)

.c.lx:
	$(CC) -S $(CFLAGS) -Wextra -o /dev/null $<

#	gcc -S $(LXFLAGS) -o /dev/null $<

# This should give warnings without producing anything;
# to use, do  "make  source.lx"
# Does not find fall through in case statement


# application dependencies ---------------------------------------------

# We have--> anything.o : anything.c depends.h $(FRC)
# The depends.h is added in individual makefiles, this is the general part

#- $(BUILD_TARGET) : $$@.c $(LIBNAME) $(FRC)
$(BUILD_TARGET) : $(BUILD_TARGET).c $(LIBNAME) $(FRC)

# library dependencies -------------------------------------------------

#-- Utility programs from the UTILS directory:

## ${LIBNAME} : ${LIBNAME}(ioutils.o) ${LIBNAME}(btree.o) ${LIBNAME}(fits.o) ${LIBNAME}(kdcutil.o) ${LIBNAME}(mgutils.o) ${LIBNAME}(dbplot.o) ${LIBNAME}(charplt.o)
# Removed  ${LIBNAME}(optics.o) 
# And, that was a comment anyway...

${LIBNAME} : ${LIBNAME}(ioutils.o) ${LIBNAME}(kdcutil.o) ${LIBNAME}(mgutils.o) ${LIBNAME}(charplt.o) ${LIBNAME}(optutils.o) ${LIBNAME}(btree.o) 
# Removed:  ${LIBNAME}(optics.o) 

## ${LIBNAME} : ${LIBNAME}(ioutils.o) ${LIBNAME}(kdcutil.o) ${LIBNAME}(mgutils.o) ${LIBNAME}(charplt.o) ${LIBNAME}(optutils.o)

${LIBNAME}(ioutils.o) : $(UTILS)/ioutils.c $(UTILS)/ioutils.h
${LIBNAME}(kdcutil.o) : $(UTILS)/kdcutil.c $(UTILS)/kdcutil.h
${LIBNAME}(btree.o)   : $(UTILS)/btree.c   $(UTILS)/btree.h
# ${LIBNAME}(fits.o)    : $(UTILS)/fits.c    $(UTILS)/fits.h
# ${LIBNAME}(mgutils.o) : $(MASKG)/mgutils.c $(MASKG)/mgutils.h
# ${LIBNAME}(mgutils.o) : $(UTILS)/mgutils.c $(UTILS)/mgutils.h
${LIBNAME}(mgutils.o) : $(MGUTILS).c  $(MGUTILS).h $(MASKG)/mgfeats.h
# ${LIBNAME}(optics.o)  : $(MASKG)/optics.c  $(MASKG)/optics.h
# ${LIBNAME}(optics.o)  : $(OPTICS).c  $(OPTICS).h
# ${LIBNAME}(dbplot.o)  : $(MASKG)/dbplot.c  $(MASKG)/dbplot.h
${LIBNAME}(charplt.o) : $(UTILS)/charplt.c $(UTILS)/charplt.h
# ${LIBNAME}(optutils.o) : $(UTILS)/optutils.c $(UTILS)/optutils.h
${LIBNAME}(optutils.o) : $(OPTUTIL).c  $(OPTUTIL).h

# When made from a different directory, the include file in the above
# xxx.c files will probably fail...

# NOTE -- if there are problems with building the library, put these
# statements into the UTILS directory's makefile, so that the
# make of the library.o files only occurs in that sub-directory.

#-- Utility programs from the MASKG directory:

# ${LIBNAME} : ${LIBNAME}(mgutils.o) 

# ${LIBNAME}(mgutils.o): $(MASKG)/mgutils.c $(MASKG)/mgutils.h $(UTILS)/ioutils.h

# List library content
list :
	ar -tv  $(LIBNAME)

# was: ar -t -v $(LIBNAME)

# tests ----------------------------------------------------------------

TSTRG	= String Substitution x
test :	target_as target_is directory/file dummy
	@echo This \($@\) should read \(test\) - substitution in echo
	@echo $(TSTRG:x=supported!)
	@echo Make shell is $(SHELL), should be /bin/sh
	@make cshell SHELL=/bin/csh
	@rm -f dummy.sstest

# Test target as dependency, the $$@ syntax
# target_as:  $$@_dependency
# 	@echo $? supported.

# target_as_dependency:
target_as:
	@echo $@ dependency not supported.

# NOTE -- The following dependency is in the Linux (GNU make) style,
# and is what causes an error exit if this file is used (wrongly) by
# a Solaris "make" command.  This is considered a *good* thing.
# target_is_dependency:
target_is: %: %_dependency
	echo $? IS supported by Gnu make.

target_is_dependency:

# In Sun, and System V make, use the syntax
#   targets: $$@.xx   to get target.xx;
# But in Gnu make, use the equivalent syntax
#   targets: %: %.xx   to get target.xx.
# Then use $@ in the rest of the syntax to match target.

# Test directory and file syntax
directory/file :
	@echo Both $(@F) and $(@D) should work,
	@echo and do, if file and directory appear above.

# Test for single suffix
.sstest :
	@echo Prerequisite $< makes target $@.

dummy.sstest :
	@touch dummy.sstest

# dummy : dummy.sstest
# Remove to make dummy prereq work.
# Single suffix rule only works if no explicit dependency is given
# for the target.  If a depencency exists, the suffix rule
# is no longer used.

# Test ability to set shell
cshell:
	@echo cshell shell is $(SHELL) \(set to /bin/csh\)
	@if (-f makefile ) then ; echo \(explicit shell setting works\) ; endif

# ----------------------------------------------------------------------
