/* DBplot == debug plot tool tasks */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// Standard includes

//#include        <stdio.h>
//#include        <string.h>
//#include        <stdlib.h>

// Global data - window stuff, etc.


// Include the standard mask generation things and standard system

#include	MGUTILH


/* Create a window, and interace with it. */

#ifdef  MASKONLY
#define P_title "SMDF plot"
#else
//#define	P_title	"Mask Generation Debug"
#define	P_title	"MaskGen Debug"
#endif
//#define	I_title	"NdP-tool"
#define	I_title	"dbP"
#define	P_vers	"2.13"

// Version 1.65 02-11-07 - remove extra calls to repaing window
// Version 1.76 05-2-3 -- support different size mask and detector
// Version 2.0  05-5-13+ -- Translate to new plotting library
// Version 2.01 05-60-16 -- Starting window is parameter of pwin
// Version 2.02 05-06-29 -- Support virtual object for gap
// Version 2.03 05-11-17 -- Return with handle_vmenu to replot window
// Version 2.05 06-04-13 -- Plot feature names, optionally
// Version 2.06 06-10-26 -- Plot GISMO subfields
// Version 2.10 07-09-27 -- Symbol MASKONLY for only mask plot
// Version 2.11 07-10-02 -- New title display sequence
// Version 2.12 08-02-27 -- Add inch mode support and wfccd support
// Version 2.13 10-03-08 -- Include order needed revision
// Version 2.14 11-10-26 -- Add debug print of mouse button, state; add SHFTCLK
// 11-11-10	-- Window position changed from CBS_BOTTOMRIGHT to CBX_MIDRIGHT

/*  Other system includes  */

//#include	<math.h>

//#include	<sys/time.h>

#include	"dbplot.h"	/* Plot tools - Includes mgutils.h - */
//#include	"ndplot.h"	/* Plot tools */
// ndplot.h is reall the same as dbplot.h for now.

/*  dbplot.h includes mgutils.h, if not already included, 
  which in turn includes all system includes and the proper things
  for mask generation includes.  This should be done first, before
  any other includes are done.  */

/*  NOTE
  System includes should *ALWAYS* come before local programming includes.
  In this instance, dbplot.h included mgutils.h, which in turn included
  <openssl/evp.h>.  Several layers into that, the <openssl/bn.h> has in
  MAC OS 1.6 a function prototype which include (optional) variable names,
  one of which is "Xp" which is used as a defined symbol in intgui.h,
  included (now) below.  Since it was defined as "4", the function prototype
  resulted in a syntax error since a numeric constant was now used instead
  of a variable name.  When system includes are done before local includes,
  this will not happen.
  
  The system includes define global things needed for proper compile and
  linking; they never need things defined by the program using them.  On
  the other hand, the local includes might well need those things defined
  by the system includes.
  
  A better method of insuring that all needed system includes preceede all
  local includes is being developed in the future.
                                                                     */

/* Global Data and structures here */

//#include	"ccbxlib.c"	/* Just put whole thing in... */
#include	"cxt.h"		/* From library directory; new plotting */
#include	"utils.h"	/* No longer included by ccb lib */
#include	"intgui.h"	/* Definition of interface */
// contains all the general routines

// LOOPDEBUG specifies the output of event information to find any
// problems with the loop
//#define  LOOPDEBUG
#undef  LOOPDEBUG

// DEBUG specifies the bounding box orange plot.
//#define  DEBUG
#undef   DEBUG

/* Symbol  MASKONLY  may be defined, and if it is, only the mask
  plane is active.  */

/* For non-solaris, rather than including kdcutil.h... */

#ifndef __sun
#define  nint(a)  (int)round(a)
#endif

// Time utility "tsince" now in complot.c included below
// Now complot is renamed cwplot

static struct timeval	previous;
static struct timeval	current;
static	int	bpressed=0;

static Application  *app;
static MainWindow  mwin;

//static Display  *display;
static Window	win;	// Short for the main window
static Button	dunbut;
#ifndef  MASKONLY
static Menu	vmenu;
#endif
static Button	zibut;
static Button	zobut;
static Button	rzbut;

static Button	tupbut;
static Button	tlfbut;
static Button	trtbut;
static Button	tdnbut;

static	Bool	iconized=False;

/* And, a list of windows... */
wble*	winlist = NULL;

GC	gc;
GC	lgc;
static	GfC	gc2;
Bool	done=False;
char	locbuf[64];

/* We have the following properties for each view/window
	> Scaling parameters BX, BY, ZX, ZY
	> Color of background
  These are indexed by MASK, DETECTOR and kept in a
  structure.  */

typedef	struct  vprops {
	double	BX, BY, ZX, ZY;
	u_long	color;
	} viewprop;

static	int	wx = 500;
static	int	wy = 550;
static	int	cwx, cwy;

static	viewprop  V;

static	viewprop  vp[2];
/* These get initialized when the window is started */
int	w_active;

#define	bw	60
#define	bh	20

#define	kw	36
#define	kh	18

#define zw	18
#define zh	18
#define zd	4

// Above are in screen coordinates
// Below are in raster coordinates

#define  kby  2
#define  kbh  6

// handler function prototypes
#ifndef  MASKONLY
// int handle_vmenu(void* param);
void handle_vmenu(void* param);
#endif

/* --- New setup; blocks for windows, buttons */

/* Button blocks for button widgets */
/* butblock:  xloc,yloc, width,height, label, *data,
	*button, *mwin, aligntype, laboff, callback  */

butblock	dbfield = {
	150, kby,  1, kbh,  "Done", NULL,
	&dunbut, &mwin, Rightside, 0, NULL };

butblock	zifield = {
	33,  kby, 0, kbh, "In", NULL,
	&zibut, &mwin, Leftside, 0, NULL };
// An approximation to a button block, needs lotsa work...

butblock	zofield = {
	46,  kby, 0, kbh, "Out", NULL,
	&zobut, &mwin, Leftside, 0, NULL };

butblock	rzfield = {
	100,  kby, 1, kbh, "ReZero", NULL,
	&rzbut, &mwin, Leftside, 0, NULL };
// callback field in buttons above was ToggleButton, has been
// removed as we are using manual rather than auto buttons here.


/* The list of buttons */
bble*	butlist = NULL;

/* Menu Block:  xloc,yloc, width,height, label, *data,
	type, *menu, *mwin, align, laboff, callback, color */

#ifndef  MASKONLY
mnublock	vfield = {
	2, kby,  6, kbh, "View", NULL,
	M_POP, &vmenu, &mwin, Leftside, 0, handle_vmenu, 0 };
#endif


	/* - - - - - - - - - - - - - - *\
	|				|
	|  Include common plot things	|
	|				|
	\* - - - - - - - - - - - - - - */


#define  ISDBPLOT

// #include	"complot.c"
#include	"cwplot.c"


void	plot_edge (sp_edge e, u_long color)
/* Plot the given edge */
{
	v2line (color, e.r, e.p);
	v2line (color, e.p, e.u);
	fcircle (app->red, e.r, 3, 0);
	fcircle (app->blue, e.u, 3, 0);
	fcircle (app->green, e.p, 3, 0);
// Could use coeficients of curve if needed
}

void	plot_spect (spect* sp, u_long color, int type)
/* Plot the sides and ends of a real spectrum */
{
    if (type == 0) {	/* edges */
	v2line (color, sp->e1.r, sp->e2.r);
	plot_edge (sp->e1, color);
	v2line (color, sp->e1.u, sp->e2.u);
	plot_edge (sp->e2, color);
	fcircle (color, sp->center, 2, 0);   /* DEBUG */
    } else {		/* bounds */
//	rectangle (color | DOTTED, sp->x1, sp->x2, sp->y1, sp->y2);
	rectangle (color | DOTTED, sp->bb.x.lo, sp->bb.x.hi, 
				   sp->bb.y.lo, sp->bb.y.hi);
    }
}

#ifdef  OBJNAM
void	ob_label (vect2 p, char* name)
/* Put the name centered above the point */
{
int	x, y;
int	k;

// Find length of label to plot for offsetting
    k = strlen (name);

// Find location within window...
    x = scale (V.BX, V.ZX, p.x, cwx);
    y = scale (V.BY, V.ZY, p.y, cwy);

// Set the color to black for the label
//    XSetForeground (mwin.disp, gc, app->black);
    XSetForeground (mwin.disp, mwin.gfs.gc, app->black);

//    XDrawString (mwin.disp, win, gc,  x - (k*7)/2,  y-4, name, k );
//    XDrawString (mwin.disp, mwin.win, mwin.gfs.gc, 
//		 x - (k*7)/2,  y-4, name, k );
    XDrawString (mwin.disp, mwin.win, gc2.gc, 
		 x - (k*6)/2,  y-4, name, k );

/* If we use lgc or gc, we don't get the font which has been set
  using CBX_CreateGfC below.  This is a general problem.  */

}
#endif


/* -------------------Plot things--------------------- */

/* Mask dimensions for wfccd copied from maskcut.c file.  */
#define  XSize   (5.980 * Inch)
#define  YSize   (4.480 * Inch)

void	draw_window (Obs* a, int df)
/* Actually do the plotting */
{
//double	x, y;
vect2	el, er, ew;
vect2	p1, p2;
//vect2	p3;
vect2	q1, q2, q3, q4;
//vect2	c1, c2;
vect2	pm;
//#ifdef  GISMO
vect2	pf;
//#endif
objq	*r;
//int	j, k;
int	i;
int	d;
int	type;
spect	*sp;
u_long	pclr;
Mindx	imx;
double	bsize;
sfset	*gset;
#ifdef  OBJNAM
//char	fontname[128];
//Font	font;
#endif

/* We will either plot objects on the mask, or spectra on
  the detector.  Depends on which page is active.  The menu will
  change which page is active, change the plot, and color.  This
  is all handled by a set of globals which contain the mode, and
  current scale for each window.  */

//    imx = (strncmp(a->instrument, "IMACS", 5) == 0);
//    imx = (strncmp(a->Instr_mt->name, "IMACS", 5) == 0);
// Only compare 5, can be IMACS or IMACS-short here...
    imx = MaskIndx (a);

/* Put fiducal marks at center on edges for zoom reference */
    lgc = gc;	// a local gc for color drawing...
    XSetForeground (mwin.disp, lgc, app->black);
    XDrawLine (mwin.disp, win, lgc, 0, cwy, 20, cwy);
    XDrawLine (mwin.disp, win, lgc, wx, cwy, wx-20, cwy);
    XDrawLine (mwin.disp, win, lgc, cwx, 50, cwx, 70);
    XDrawLine (mwin.disp, win, lgc, cwx, wy, cwx, wy-20);

    cross (app->black, 0.0, 0.0, 8, 0);	/* black cross at origin */

    if (w_active == MASK) {
//	cross (app->green, 150.0, 0.0, 8, 0);	/* Green cross at 150 */
//	cross (app->green, 0.0, 150.0, 8, 0);	/* Green cross at 150 */
// Plot the mask bounds
/* Fully obsolete...
	x = a->Telescope->field_size / 2.0;
	y = x;
	rectangle (app->red | DOTTED, -x, x, -y, y);
.. not plotted at all... */

// Plot the mask boundary (IMACS specific)
//	bsize = imx ? 313.0 : 93.3;	// imacs or ldss here
/* Change to use full index for imacs, ldss, gismo */
// Setting the size for mask boundry (probably obsolete)
    switch  (imx) {
	case  M_imacs:	bsize = 313.0;		break;
	case  M_lds2:
	case  M_lds3:	bsize = 93.3;		break;
	case  M_gismo:	bsize = 175.0;		break;
	case  M_wfcd:	bsize = XSize/2.0;	break;
	default:	bsize = 500.0;		break;
    }

    if (imx == M_imacs || imx == M_lds3) {
	scircle (app->red | DOTTED, make2vect(0.0, 0.0), bsize, 0);
    } else if (imx == M_gismo) {
//	rectangle (app->red, -175.0, 175.0, -163.25, 163.25);
	rectangle (app->red, -175.0, 175.0, -143.25, 143.25);
// Above reduced in y by 2 slide intervals top and bottom; the avoids
// will also need to be reduced similarly, but only for maskgen!
/* Also, plot the individual boundries */
	gset = (sfset*)(a->Frontend->data);
	if (gset != NULL) {
	    for (i=0; i < gset->n; i++) {	// Over subfields
		rectangle (app->orange,
		    gset->s[i].field.x.lo, gset->s[i].field.x.hi,
		    gset->s[i].field.y.lo, gset->s[i].field.y.hi );
	    }
	}
    } else if (imx == M_wfcd) {
//	rectangle (app->red, -175.0, 175.0, -163.25, 163.25);
// That is the full rectangle.
	rectangle (app->red | DOTTED,  -XSize/2.0,  XSize/2.0,
		-YSize/2.0,  YSize/2.0);
// A smaller rectangle outlining the wfccd mask itself.
// We don't show the corner cuts or the alignment notches.
    }

// Loop for objects
	lgc = gcsetup (gc, app->blue);
	d = 3;
// Loop over objects...GISMO
	C_loop (r, a->ob) {
/* Skip any virtual objects in the mask view, unless labeling them */
#ifdef  OBJNAM
	    if ((a->Flag & OFpon) == 0)
#endif
//		{ if (r->flag & OBJECT_VIRTUAL) continue; }
// what we really mean is to skip gap objects here
		{ if (r->type == OBJ_GAP) continue; }

/* Skip any inactive objects... */
    /* Plot the slit with width, etc. */
	    pm = r->smpos;
//#ifdef  GISMO
	    pf = r->fpos;
//#endif
	    if (r->flag & OBJECT_ACTIVE) {
//	    if ( (r->flag & OBJECT_ACTIVE) &&
//		((r->flag & OBJECT_VIRTUAL) == 0) ) {
// The wslit, lslit and rslit routines return vectors for the
// slit width and length edge centers.  Summing appropriately
// yields the slit corners here.
		ew = wslit (r->slit);
    /* Plot real slit by shape... */
		p1 = sum2vect (pm, ew);   // top center
		p2 = sub2vect (pm, ew);   // bottom center
		if (r->slit.shape == RECTANGLE) {
		    el = lslit (r->slit);
		    er = rslit (r->slit);
		    q1 = sum2vect (p1, el);
		    q2 = sum2vect (p1, er);
		    q3 = sum2vect (p2, er);
		    q4 = sum2vect (p2, el);
		    quadralat (app->blue, q1, q2, q3, q4);
/* Add green ghost slit on input for gismo */
//#ifdef  GISMO
// should recompute ew, p1, p2 for any rotations
		    if (imx == M_gismo) {
			ew = wslit (r->islit);
			p1 = sum2vect (pf, ew);
			p2 = sub2vect (pf, ew);
			el = lslit (r->islit);
			er = rslit (r->islit);
			q1 = sum2vect (p1, el);
			q2 = sum2vect (p1, er);
			q3 = sum2vect (p2, er);
			q4 = sum2vect (p2, el);
			quadralat (app->green, q1, q2, q3, q4);
		    }
//#endif
		} else if (r->slit.shape == SQUARE) {
		    er.x =  ew.y;
		    er.y = -ew.x;
		    q1 = sub2vect (p1, er);
		    q2 = sum2vect (p1, er);
		    q3 = sum2vect (p2, er);
		    q4 = sub2vect (p2, er);
		    quadralat (app->blue, q1, q2, q3, q4);
//#ifdef  GISMO
		    if (imx == M_gismo) {
			ew = wslit (r->islit);
			p1 = sum2vect (pf, ew);
			p2 = sub2vect (pf, ew);
			er.x =  ew.y;
			er.y = -ew.x;
			q1 = sub2vect (p1, er);
			q2 = sum2vect (p1, er);
			q3 = sum2vect (p2, er);
			q4 = sub2vect (p2, er);
			quadralat (app->green, q1, q2, q3, q4);
		    }
//#endif
		} else if (r->slit.shape == CIRCLE) {
    // Find the hole radius in pixels... use X scale for this
		    scircle (app->blue, pm, r->slit.width/2.0, 0);
/* Add a green ghost circle on input for gismo */
//#ifdef  GISMO
		    if (imx == M_gismo)
			scircle (app->green, pf, r->islit.width/2.0, 0);
//#endif
		}
#ifdef  OBJNAM
		if (a->Flag & OFpon) ob_label (pm, r->dat.name);
#endif
	    }	// active objects conditional
	    type = (r->type == OBJ_OBJECT);
// Plot filled circle for observ. or open circle for reference
//	    pclr = (r->flag & OBJECT_ACTIVE) ? app->green : app->brown;
// Other colors include orange, red, and some lighter ones.
// blue is being used for the slit/hole outlines.
	    pclr = (r->type == OBJ_OBJECT) ? app->green : app->orange;
// Want type to become 0 and color green for wing chip holes...
//	    if (r->slit.shape == CIRCLE && r->type == OBJ_REFERENCE) {
	    if (r->flag & OBJECT_SERVICE) {
		type = 0;
		pclr = app->green;
	    } else {
		type = 1;	// filled circle
	    }
	    fcircle (pclr, pm, d, type);
/* Add image of input object if gismo, in red dot */
//#ifdef  GISMO
	    if (imx == M_gismo && (r->flag & OBJECT_ACTIVE) &&
		(r->type != OBJ_GAP) && (r->type != OBJ_AREA) ) {
// What was meant is non-gap objects, not non-virtual objects
// Also not area objects (i.e. gismo aperture objects)
//		((r->flag & OBJECT_VIRTUAL) == 0) ) {
		fcircle (app->red, r->fpos, d, type);
#ifdef  OBJNAM
//		if (a->Flag & OFpon) ob_label (r->fpos, r->dat.name);
		if (a->Flag & OFpon) ob_label (pf, r->dat.name);
#endif
	    }
//#endif
	}

    } else if (w_active == DETECTOR) {

/* Drop the green crosses once det. bound is drawn */
//	cross (app->green, 60.0, 0.0, 8, 0);	/* Green cross at 60 */
//	cross (app->green, 0.0, 60.0, 8, 0);	/* Green cross at 60 */

// Plot the detector bounds
//	rectangle (app->red | DOTTED, 
	rectangle (app->black, 
			a->Detect->bb.x.lo, a->Detect->bb.x.hi,
			a->Detect->bb.y.lo, a->Detect->bb.y.hi);

/* Loop for objects -- plot spectral extent for each */
	C_loop (r, a->ob) {
// NOTE - want a debug level to be able to skip here any
// spectrum which is NOT ACTIVE
	    if (r->val == NULL) continue;	// for wing chip holes
	    sp = r->val->sod;
	    if (sp == NULL) continue;	// for wing chip holes
#ifdef  DEBUG
	    plot_spect (sp, app->orange, 1);	/* bounds */	// the orange box
#endif
	    pclr = (r->flag & OBJECT_ACTIVE) ? app->blue : app->brown;
/* If object is a gap, we plot it specially */
	    if (r->type == OBJ_GAP) {
//		plot_spect (sp, app->brown, 1);	/* gap */
		plot_spect (sp, app->black, 1);	/* gap (use 0 to debug) */
	    } else {
		plot_spect (sp, pclr, 0);	/* spectrum */
#ifdef  OBJNAM
		if (a->Flag & OFpon) ob_label (sp->center, r->dat.name);
#endif
	    }
/* Plot the image spectra in dotted green lines */
/* Temporarily dtop the images... debug  */
	    C_loop (sp, r->val->img) {
		if (sp == r->val->sod) continue;  // debug
		plot_spect (sp, app->green | DOTTED, 0);
	    }
	}	/* Loop for objects on detector */
    }		/* End window type decision */

// Plot an x-y axis
//	axis (black, 0.0, 5);
//	axis (black, 90.0, 5);
//	axis (black, 180.0, 2);
//	axis (black, 270.0, 2);


/* --- Scale bar plotted in corner --- */
    if (V.BX != 0.0) {
double	x, b, d, s, w;
int	i, p;
int	mx, my, mt, c;
static double	sv[6] = { 10.0, 5.0, 3.0, 2.0, 1.0, 0.5 };
char	label[16];
	b = V.BX;
	if (a->Flag & OFinc) b *= Inch;
	if (b < 0.0) b = -b;	/* The scale value mm or inch */
	w = (double)wx / 3.0;	/* Maximum scale bar size */
	x = w / b;		/* Maximal scaled value */
	d = pow (10.0, floor (log10 (x)));
	for (i=0; i<6; i++) {
	    s = sv[i] * d;
	    if (s <= x) break;
	}
/* s is now size of the scale bar (millimeters or inches) */
	if (a->Flag & OFinc) {	// inch
	    if (s >= 0.01 && s < 100.0) {
		sprintf (label, "%4.2f inch.%c", s, AN);
	    } else if (s > 99.0) {
		sprintf (label, "%8.1e feet.%c", s/12.0, AN);
	    } else {
		sprintf (label, "%8.1e m.%c", s*Inch/1000.0, AN);
	    }
	} else {		// metric
	    if (s >= 0.01 && s < 1000.0) {
		sprintf (label, "%4.2f mm.%c", s, AN);
	    } else {
		sprintf (label, "%8.1e m.%c", s/1000.0, AN);
	    }
	}
	p = (int)(b * s + 0.5);
/* p is length of the scale bar in pixels */
	mx = wx - 10 - p;	// was 10;
	my = wy - 20;
	mt = 3;
	c = strlen (label);

	lgc = gcsetup (gc, app->black);
	XDrawLine(mwin.disp,win,lgc,mx,my-mt,mx,my+mt);      // draw scale
	XDrawLine(mwin.disp,win,lgc,mx+p,my-mt,mx+p,my+mt);
	XDrawLine(mwin.disp,win,lgc,mx,my,mx+p,my);
	XDrawString(mwin.disp,win,lgc, mx + (p/2) - (c*4), my+16, label, c);
    }	/* End of scale bar section */


/* Put current center location in upper right corner */
    {
    int	mx, my;
    int	c;
    GC	tgc;
    char	cbuf[64];

    char	cra[24];
    char	cde[24];
//    vect3	v;
    double	ra, dec;

//	tgc = CBX_CreateGC (app->black, app->white, "courier", "medium", 12);

//	tgc = Set_GC (app->black, app->white, "courier", "medium", 12);
	tgc = gc;
/* tgc is used to write display positions */

//	sprintf (cbuf, "%12.5g %12.5g", V.ZX, V.ZY);
//	sprintf (cbuf, "%10.3f %10.3f", V.ZX, V.ZY);  // was 12.5g
	if (a->Flag & OFinc) {
	    sprintf (cbuf, "%10.4f %10.4f  in.", V.ZX/Inch, V.ZY/Inch);
	} else {
	    sprintf (cbuf, "%10.3f %10.3f  mm.", V.ZX, V.ZY);	// was %12.5g
	}
	c = strlen (cbuf);
	mx = cwx + 8;
	XDrawString (mwin.disp,win,tgc, mx, 64, cbuf, c);

/* Also, plot ra/dec of center location... */
/* First, do this only for mask... */
	if (w_active == MASK) {
/* Take inverse of the current center position */
	    unmaskvect (a, make2vect(V.ZX, V.ZY), &ra, &dec);
	    strcpy (cra, sexig (ra/Hour, 3, 3));
	    strcpy (cde, sexig (dec/Degree, 3, 2));
	    sprintf (cbuf, "%s  %s", cra, cde);
	    c = strlen (cbuf);
	    XDrawString (mwin.disp,win,tgc, mx, 75, cbuf, c);
	}	// conditional on mask window active

/* If requested, cursor position is in lower left corner */
	if (df & 2) {
    // when dflag has a 2 in it, print the locbuf contents...
	    c = strlen (locbuf);
	    my = wy - 4;
	    XDrawString (mwin.disp,win,tgc, 7, my, locbuf, c);
	}	/* End cursor position section */
    }	/* End position writing section */
}


	/* - - - - - - - *\
	|                 |
	|   Plot Window   |
	|                 |
	\* - - - - - - - */


void	pwin (Obs* obs, int wz)
/* Plot a standard window */
{


/* Local variables */
int	dflag = 0;	// Only set it explicitly
XEvent		event;
XSizeHints	hint;
XWindowAttributes	xwa;
char		buffer[100];
char		xserver[80];
char		title[80];
char		fontname[80];
char		I_name[40];
int		sleeptime;
int		i;
int	FontSize;
Mindx	imx;		/* Mask Index; formerly 1 if IMACS */
double	bsize;		/* Size, depends on window and instrument */
Window	xwx;		/* Scratch; used as abbreviation */

// debug things
int	sleepcount;
int	xnot;

/* Set up the time values */
    if (tsince (&previous, &current) != 0.0) previous = current;
    previous.tv_sec -= 9;

/* Make up title string */
//    sprintf (title, "(%s) %s (V%s)", I_title, P_title, P_vers);
    sprintf (I_name, "%s %.8s", I_title, obs->fname);
    sprintf (title, "(%s) %s (V%s)", I_name, P_title, P_vers);

/* Get a display */
    strcpy (xserver, genv2("DISPLAY","") );

    if ((app=CBX_InitXlib(xserver)) == NULL) {      /* connect to 'xserver' */
	fprintf(stderr,"%s: CBX_InitXlib(%s) failed\n", "Interface",xserver);
	exit(1);
    }

/* Start up a window */

    done = False;	// In case we are called again.

    wx = 550;
    wy = 600;

    win = CBX_OpenMainWindow (
	&mwin, 
//	CBX_BOTTOMRIGHT, 
	CBX_MIDRIGHT, 
	wx, wy,		/* Window size */
	&hint, 
	title, I_name,	// was I_title here
	False);		// Allow window resize, not fixed...

    gc = mwin.gfs.gc;	// Set global gc

    CBX_SelectInput (&mwin,
	ExposureMask | StructureNotifyMask | KeyPressMask |
	SubstructureNotifyMask | PropertyChangeMask |
	ButtonPressMask | ButtonReleaseMask );

// Actually we would really only need these:
//    CBX_SelectInput (&skywin, StructureNotifyMask |
//		ExposureMask  |  ButtonPressMask );


/* Graphics Contexts */
//    strcpy (fontname, "courier");
    strcpy (fontname, "lucidatypewriter");
//    strcpy (fontname, "lucida sans typewriter");
//    strcpy (fontname, "fixed");
//    gc = CBX_CreateGC (black, white, fontname, "medium", 14);
    FontSize = 14;
    (void)CBX_CreateGfC (&mwin, fontname, "bold", FontSize);
//    (void)CBX_CreateGfC (&mwin, fontname, "medium", FontSize);
// Above is medium or bold
    gc = mwin.gfs.gc;	// Set global gc (again)
    gc2 = CBX_CreateGfC_Ext (mwin.disp, fontname, "medium", FontSize-3);

// Also set hint size things here.
    hint.min_width  = 400;    hint.min_height = 450;
    hint.max_width  = 1100;   hint.max_height = 1050;
    hint.flags = PPosition | PSize | PMinSize | PMaxSize;
//    CBX_SetMainWindowProperties (&mwin, title, I_title, &hint);
    CBX_SetMainWindowProperties (&mwin, title, I_name, &hint);
// That's how we set the min/max things, manually.

    XGetWindowAttributes (mwin.disp, win, &xwa);

    set_sizes (&xwa);

/* Initialize the scale and color of each window */
//    imx = (strncmp(obs->instrument, "IMACS", 5) == 0);
//    imx = (strncmp(obs->Instr_mt->name, "IMACS", 5) == 0);
    imx = MaskIndx (obs);

/* Change this:
  use imx = MaskIndx (obs) to obtain the actual mask index,
  and then a switch statement on M_xxx index values to set
  the proper bsize for imacs, ldss, gismo and others.  */

// Need the n because can be IMACS or IMACS-short
// Set bsize to projected window size on mask or detector space.
//    bsize = imx ? 660.0 : 200.0;
// Setting the initial (default) scale for the mask plot
    switch  (imx) {
	case  M_imacs:	bsize = 660.0;		break;
	case  M_lds2:
	case  M_lds3:	bsize = 200.0;		break;
	case  M_gismo:	bsize = 360.0;		break;
	case  M_wfcd:	bsize = 180.0;		break;
	default:	bsize = 500.0;		break;
    }
    vp[MASK].BX = (float)wx / bsize;  // Was.. 450.0;
    vp[MASK].BY = -vp[MASK].BX;
    vp[MASK].ZX = 0.0;
    vp[MASK].ZY = 0.0;
//    vp[MASK].color = app->yellow;
//    vp[MASK].color = app->offwhite;
    vp[MASK].color = app->lemon;

/* Also use a switch statement here to select the proper detector
  projected space size for imacs, gismo, ldss and others.  */

//    bsize = imx ? 140.0 : 70.0;
// NOTE that meaning of imx changed; was non-zero for imacs, now mask index.
// Set the initial (default) scale for the detector plot
/*  NOTE -- we perhaps SHOULD have separate index values for the
  mask type (above) and the instrument (here) instead of using the
  mask index for both.  Do this when or if it is required.  */
    switch (imx) {
	case  M_imacs:
	case  M_gismo:	bsize = 140.0;		break;
	case  M_lds2:
	case  M_lds3:	bsize = 70.0;		break;
	default:	bsize = 170.0;		break;
    }
    vp[DETECTOR].BX = (float)wx / bsize;
    vp[DETECTOR].BY = -vp[DETECTOR].BX;
    vp[DETECTOR].ZX = 0.0;
    vp[DETECTOR].ZY = 0.0;
    vp[DETECTOR].color = app->ivory;

#ifdef  MASKONLY
    w_active = MASK;
#else
//    w_active = DETECTOR;	/* MASK; */
//    w_active = MASK;	/* DETECTOR; */
    w_active = wz;
#endif
    V = vp[w_active];
    XSetWindowBackground (mwin.disp, win, V.color);


// All the hints are set previously above.

// Some buttons

// zoom in button
    MakeMButton (&butlist, &zifield);

// zoom out button
    MakeMButton (&butlist, &zofield);

// rezero button
    MakeMButton (&butlist, &rzfield);

// Done button
// Need dbfield to describe the button, and butlist to hold addresses.
    MakeMButton (&butlist, &dbfield);

#ifndef  MASKONLY
// Mask/Detector window selection menu
//    MakePMenu (&vfield);
    MakeMenu (NULL, &vfield);
/* Add the menu items */
    CBX_AddMenuEntry (&vmenu, "Mask", 0);
    CBX_AddMenuEntry (&vmenu, "Detector", 0);
// need to make sure values are properly passed
    vmenu.sel = w_active;
//    dflag = 1;
// Experimentally remove to reduce initial re-paints
#endif


///* Create translation buttons */
//    bx += bw + 40 + zw + zd;	/* new reference location */

//    CBX_CreateSButton (&tupbut, win, gc,
//	bx, by,		/* location */
//	zw, zh, CBX_UP_TRI);	/* size and legend */

//    CBX_CreateSButton (&tlfbut, win, gc,
//	bx - zw - zd, by + zh + zd,	/* location */
//	zw, zh, CBX_LT_TRI);	/* size and legend */

//    CBX_CreateSButton (&trtbut, win, gc,
//	bx + zw + zd, by + zh + zd,	/* location */
//	zw, zh, CBX_RT_TRI);	/* size and legend */

//    CBX_CreateSButton (&tdnbut, win, gc,
//	bx, by + 2*(zh + zd),		/* location */
//	zw, zh, CBX_DN_TRI);	/* size and legend */


//    bx += 2*(zw + zd) + 20;		/* new reference location */


	printf (" Window created; title: %s \n", title);
	fflush (stdout);



/* --------------------Modify things-------------------- */



	/* - - - - - - - - - - *\
	|			|
	|      The  Loop	|
	|			|
	\* - - - - - - - - - - */

/* ----------------- The Loop ------------------ */

/* Loop for mouse and keyboard inputs */
    xnot = 0;
    while (!done) {
int	button;
#ifdef  LOOPDEBUG
int	mainwin;		//debug
#endif
	sleeptime = 100;
	sleepcount = 0;

	while (!CBX_AutoEvent(&mwin,&event,&sleeptime)) {
	    if (done) break;			/* break event loop */
/* The above break conditionals are not needed and should be removed;
  if needed to exit due to an external event, simply pass the pointer
  to sleeptime to the inner loop handler as done for IFU_handle; it
  will get reduced if an event is detected... */
/*  The warning flag is set in textback and the button callback which
  are done automatically by AutoEvent.  They don't have a handle to
  the sleeptime, so they need another way to get out.  However, dflag
  is local to this program, and is probably not needed in this loop.  */
//dpause ("In auto event loop.");
	    CBX_Sleep(&sleeptime, 20, 200);
	    sleepcount++;
	    xnot = 0;		// Clear if we have slept at all.
	}	/* Wait for event */
	if (done) break; 


//	printf (" Slept %d times at %d ms. each.\n", sleepcount, sleeptime);
//	fflush (stdout);	// debug

#ifdef  LOOPDEBUG
	mainwin = (event.xany.window == win) ? 1 : 0;	// debug
	printf (" Event %2d s=%d m=%d x=%2d",
		event.type, event.xany.send_event,
		mainwin, xnot);	// debug
	if (event.type == 12) printf (" c=%2d", event.xexpose.count);
	printf ("\n");
	fflush (stdout);
#endif

	switch (event.type) {

	case KeyPress:		/* 2 */
	    if (event.xkey.window == win) { /* main window */
		KeySym key;	/* That's a variable typedef... */
		i = XLookupString (
			(XKeyEvent*)&event, buffer, sizeof(buffer), 
			&key, 0);
		if (i == 1) { if (*buffer == 'q') done=True; }
	    }
	    break;

	case ButtonPress:	/* 4 */
// Handle the buttons
	    xwx = event.xbutton.window;	// Basically, an abbreviation...
	    button = event.xbutton.button;  // Also an abbreviation
/* Some apple computers (such as MacBook) do not have a center button
  when operating with the trackpad.  To enable center clicking, we substitute
  the Shift-click as a center click on symbol SHFTCLK.  */
#ifdef  SHFTCLK  // was just: __APPLE__
	    if (event.xbutton.state & ShiftMask) button = 2;
#endif
/* Debug print of button and state values to test mouse capability. */
	    if (obs->dlevel > 1) {
		printf ("  > Button %02X, State %.4X\n",
		  (unsigned int)event.xbutton.button, 
		  (unsigned int)event.xbutton.state);
	    }
	    if (xwx == dunbut.win) done = True;
#ifndef  MASKONLY
	    else if (xwx == vmenu.win) {
//		dflag |= handle_vmenu();
//		{ handle_vmenu(NULL); dflag = 2; }	// temporary fix
		handle_vmenu (NULL);		// Expose normally follows
		dflag |= 1;	// For MacOSX redisplay bug
	    }
#endif
	    else if (xwx == zibut.win)
		dflag |= handle_scale (1, button);
	    else if (xwx == zobut.win)
		dflag |= handle_scale (0, button);
	    else if (xwx == rzbut.win)
//		dflag |= recenter (button);
		dflag |= recenter ();
	    else if (xwx == tupbut.win)
		dflag |= handle_translate (UP, button);
	    else if (xwx == tdnbut.win)
		dflag |= handle_translate (DOWN, button);
	    else if (xwx == tlfbut.win)
		dflag |= handle_translate (LEFT, button);
	    else if (xwx == trtbut.win)
		dflag |= handle_translate (RIGHT, button);
	    else if (xwx == win)
		dflag |= handle_location (event.xbutton.x, event.xbutton.y,
				button);
	    break;

	case ButtonRelease:	/* 5 */
	    xwx = event.xbutton.window;	// Basically, an abbreviation...
	    if ((xwx == zibut.win)  ||
		(xwx == zobut.win)  ||
		(xwx == rzbut.win)  ||
		(xwx == tupbut.win) ||
		(xwx == tdnbut.win) ||
		(xwx == tlfbut.win) ||
		(xwx == trtbut.win) ||
		(xwx == win)) {
//		dflag = 1;
		break;	// We don't do it locally
	    } else  printf (" Button Release detected.\n");
	    break;

	case MotionNotify:	/* 6 */
//	    printf (" Motion notification.\n");
	    break;

	case EnterNotify:	/* 7 */
	    xwx = event.xcrossing.window;  // Basically, an abbreviation...
/* This and <8> seem not to be handled here, this code should be commented */
//	    if (xwx == win)
//		XSetWindowBackground (mwin.disp, xwx, app->white);
	    if ((xwx == dunbut.win) ||
#ifndef  MASKONLY
		(xwx == vmenu.win)  ||
#endif
		(xwx == zibut.win)  ||
		(xwx == zobut.win)  ||
		(xwx == rzbut.win)  ||
		(xwx == tupbut.win) ||
		(xwx == tdnbut.win) ||
		(xwx == tlfbut.win) ||
		(xwx == trtbut.win) )
//		(xwx == vmenu.win) )
		XSetWindowBorder (mwin.disp, xwx, app->green);
	    break;

	case LeaveNotify:	/* 8 */
	    xwx = event.xcrossing.window;  // Basically, an abbreviation...
//	    if ((xwx == win) )
////		XSetWindowBackground (mwin.disp, xwx, app->lgrey);
//		XSetWindowBackground (mwin.disp, xwx, app->white);
	    if ((xwx == dunbut.win) ||
#ifndef  MASKONLY
		(xwx == vmenu.win)  ||
#endif
		(xwx == zibut.win)  ||
		(xwx == zobut.win)  ||
		(xwx == rzbut.win)  ||
		(xwx == tupbut.win) ||
		(xwx == tdnbut.win) ||
		(xwx == tlfbut.win) ||
		(xwx == trtbut.win) )
//		(xwx == vmenu.win) )
		XSetWindowBorder (mwin.disp, xwx, app->black);
	    break;

	case Expose:		/* 12 */
	    if (event.xexpose.count > 0) { xnot=0; break; } /* ignore */
	    if (event.xexpose.window == win) {	/* that's us */
//		dflag = 1;
//		if (!xnot) dflag = 1;
		if ( (xnot == 0)  ||
			(tsince (&previous, &current) > 3.0)
			)  dflag = 1;
	    }
	    break;

	case CreateNotify:	/* 16 */
/* Ignore?  Generated by propertychangemask or substructurenotifymask */
	    break;

	case DestroyNotify:	/* 17 */
	    if (event.xdestroywindow.window == win) done = True;
/* OK, but that doesn't break the loop when quit selected */
	    break;

	case UnmapNotify:	/* 18 */
	    if (event.xmap.window == win) iconized = True;
	    break;

	case MapNotify:		/* 19 */
	    if (event.xmap.window == win) iconized = False;
	    break;

	case ReparentNotify:	/* 21 */
		/* Ignored */
	    break;

	case ConfigureNotify:	/* 22 */
	    if (event.xconfigure.window == win) {	/* main window */
/* Find new config - attributes */
		xnot |= event.xconfigure.send_event;
		XGetWindowAttributes (mwin.disp, win, &xwa);
		set_sizes (&xwa);
//		MoveButton (&dbfield);	// "Done" tracks right side
//		MoveButton (&rzfield);	// "ReZero" is centered
//		XMoveWindow (mwin.disp, dunbut.win, 
//			xwa.width - bw -10, 10);
/* The "done" button needs to move to track right side; the others don't,
  should be protected by min width value */

/* Redraw the window */
//		dflag = 1;
// Temporarily remove to prevent extra refresh at initiallize

	    } else {
//		printf (" -- Configure other window?\n");  /* debug */
		;	/* Null action */
	    }
	    break;

/* Events needed/observed:
... See /usr/include/X11/X.h for these definitions ...
..*/

	case PropertyNotify:	/* 28 */
/* Ignore?  Generated by propertychangemask or substructurenotifymask */
	    break;

	case KillNotify:	/* 52 */
/* The window has been detected to have been closed... */
	    done = True;
	    break;

	default:
	    printf (" Other event '%2d'\n", event.type);
/* See /usr/include/X11/X.h for a definition of the event. */
	    fflush (stdout);
	    break;

	} /* End event switch */

/* Do global things which the switch set up (still in loop) */

	if (iconized) dflag &= ~2;  // for lint, iconized needs to be used

/* Check if display needed */
	if (dflag) {
	    setup_window ();
	    draw_window (obs, dflag);
	    dflag = 0;
	}

    }  /* main loop */

    CBX_CloseMainWindow (&mwin);

/* Clean up lists */
    while (winlist != NULL) winlist = wlpop (winlist);
    while (butlist != NULL) butlist = blpop (butlist);

/* All done */


}

/*   ===   Revision History   ===  

-- See above near start for revisions --

2005/11-17 -- Version 2.03
	-- Revised for Mac-OS refresh bug

  ==  */
