/* maskcut == cutting of masks for most instruments from an SMDF file */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The SMDF file name is specified in the command line. */
/* Product is a disk file of mask cutting commands. */

/*  ====  STANDARD INCLUDES  ====  */
#include	<stdio.h>
#include	<stdlib.h>
#include        <string.h>
#include	<unistd.h>	/* for access */

/*  ====  GLOBAL DEFINITIONS  ====  */

const  char	PROGNAME[] = "Mask Cutting utility";
// const  char	VERSION[]  = "3.19";
// int	dodebug = 1;	/* Global debug flag */

//extern	char	CTUVERS[];	// Cut utility version string

// #include	"mgfeats.h"
// features included by maskdat.h now.

#define  LDSS_NEWC
/* Above allows cutting support for LDSS to be dropped independently
  of allowing the support of LDSS in mask generation.  Keep the
  definition of LDSS_NEWC the same as LDSS:NEW for normal operation.  */
/* The LDSS:NEW and LDSS_NEWC should always be defined, as those
  features are now standard */
/* Symbol LDSS_NEWC is now standard, conditionals removed. */

#define  maskrad  (186.6 / 2.0)
#define  suboff   200.0
//double	Offset = 0.0;	// Global offset
/* Could also use opt.subm as 0 or non-zero for this */

//#undef  USE_CTE
#define  USE_CTE
/* The USE_CTE symbol enables the thermal expansion feature, to cut
  masks at a different temperature than they are used at.
  ( See cututil.h for definitions of CTE and Tcut. )
  The CTE and Tcut are read in from maskcut.cfg.
  Global variables and constants are defined for the temperature
  of observation, cutting and the CTE of the mask material.  */

double	CTE  =  0.0;
double	Tcut = 20.0;

//#define  TESTMODE
#undef  TESTMODE
/*  Symbol  TESTMODE  is defined to allow the -t flag to turn on
  a test mode which will report interesting things.  For debug
  use only, normally disabled.  */

#include	"cututil.h"	// includes kdcutils
// cututil.h  includes mgutils.h
// mgutils.h  includes optutils.h
// optutils.h includes maskdat.h
// maskdat.h  includes ioutils.h, kdcutil.h, mgfeats.h
/* note that smdfutil.c is not included. */

const	char	VERSION[] = Xstr(MASKCTV);
char	XVERSION[16] = Xstr(MASKCTV);

// Standard Version structure
const	Vers	Program = { NULL, NULL,  MASKCTV, MASKCTR, MASKCTT };
#define  APPLICATION_NAME  "Mask Cutting utility"
//const	char	PROGNAME[] = MASKCTT;
// mgutils expect PROGNAME and VERSION strings to exist as global labels

/* The following symbols for gismo debug could be moved to mgfeats.h if
  they are actually needed globally.  */

//#undef  GISMOFFSET
#define  GISMOFFSET 1
/* Symbol  GISMOFFSET  is used to define and use a special offset value
  used for GISMO mask cutting, due to a possible offset in the cutting
  fixture used for the gismo masks.  Should not be needed.  */

//#undef  GISMONOTCH
#define  GISMONOTCH 1
/* Symbol  GISMONOTCH  is used to create a special .nc file which only
  cuts the edge notches in a gismo mask, used to test the clearence
  allowance in the cutting system.  Normally undefined. */

#define  GISMONEW
//#undef  GISMONEW
/*  Symbol  GISMONEW  is used to create the "new" GISMO mask notch
  configuration, with a small notch on the bottom, and omitting the
  two side notches at the right.  Should be defined when the new
  cutting and holding hardware is used, starting 2010/01.    */

#define  GISMONW2
//#undef  GISMONW2
/*  Symbol  GISMONW2  is used to define some newer GISMO notch dimensions
  and configuration on 2010/01-27.  Should be defined for the new cutting
  hardware.  Superseeds the GISMONEW symbol.  (Both should be defined)  */


typedef struct  {
	int	dbug;
	int	rerd;
	int	rimh;
	int	slow;
	int	test;
	int	subm;
	int	subc;
	int	xcte;
	int	subg;
	int	subw;
	int	repc;
#ifdef  GISMODECKER
	int	gsdk;
#endif
#ifdef  GISMONOTCH
	int	gnch;
#endif
	char	*base;
#if  GISMOFFSET || GISMONOTCH
	double	offset;
#endif
	Pvfs	*cp;
	Pvfs	*fp;
	}  opts;


// Parameters in global for IMACS use
double	aangle;
double	aclear;

static	time_t	lastime = 0;
static	int	serno = -1;
static	char	tstamp[12] = "00.0";
static	char	adc = ' ';

// double	SF = 1.0;	// Scale factor
/* SF will be set to 1/25.4 to produce inch output */
// int	inchout = 0;	// Logical flag for inch output
// inchout is defined in cututil.h, as is SF also.
int	mmallow = 0;

/*  ====  Local utility and include things  ====  */

#include	"../utils/charplt.h"

/*  ====  Debug routines from smdftest.c  ====  */

/* Had just dumpoptic and dumpgrating here... */

#include	"msdebug.c"

#ifdef  USE_ACON
#include	"ncrutil.h"
#endif

/*  ===  Global locations  ===  */

/* --- Extended character function (put in kdcutil.c) --- */

/*  ----  Programs for the Laser machine  ----  */

#ifdef   GISMONOTCH
// needed function prototype
//double	shole (vect2 c, double d, double h, double fr, FILE* f);
#endif

// #include	"cututil.c"

// That code is directly included, whereas it should be compiled
// into a separate object file and linked.
// Also the globals SF and inchout will need to be defined properly.



	/* - - - - - - - - - - - - - - *\
	|				|
	|      Common  Subroutines	|
	|				|
	\* - - - - - - - - - - - - - - */

#include	"maskcutsub.c"
/* Put those back here when proven right */


/* ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- */

	/* - - - - - - - - - - - - - - - - *\
	|                                   |
	|    IMACS  laser mask cutting      |
	|                                   |
	\* - - - - - - - - - - - - - - - - */

/* --- Cutting of an IMACS mask --- */

int	cutimacs (Obs* a, FILE* f, opts opt)
/* Return 1 if succeed in outputing IMACS cutting commands
  to the file f  */
{

double	tool;
vect2	c;	// location of center
static	vect2	last = {999.0, 999.0};	// Absolute last loc for slew
static	double	zlast;	// For incremental z movement...
objq	*aq;	// Start object
char	line[80];
vect2	loc, base;	// Label location, direction/size
double	rad, basl;	// Label radius, base size
double	sn, cs;		// Location sin,cos values
double	xs;		// Label size
double	h;		// Z-altitude
adjustment	ad;	// Adjustment structure...

/* NOTE:  If a is NULL, we have a request for cutting only the
  flange hole set; initialize and skip directly to that... */

//    if (a->dlevel > 0) printf ("  Cutting IMACS mask... \n");
    if (opt.dbug > 0) printf ("  Cutting IMACS mask... \n");

/* Some debug stuff */
//    if (a->dlevel > 0) {
    if (opt.dbug > 0) {
#ifdef  SIMPLECURVE
	printf ("  Inst = %d, curve = %7.5f\n", inst, Curve);
#else
vect2	cv;
#ifdef  USE_CDAT
	cv = CDhead[inst].curve;
#endif
#ifdef  USE_GSTO
	cv = *(vect2*)GS_lookup(&CDhead, inst, CD_curve, 0);
#endif
#ifdef  USE_BSTO
	cv = vect2value(getval(&CDhead,MaskCase(inst),"Curve"));
#endif
	printf ("  Inst = %d, curve = %7.5f (x), %7.5f (y)\n", inst, 
		cv.x, cv.y);
#endif	// on SIMPLECURVE
    }

/* Initializing machine */

    lasermode (Autofocus_off, f);

    Astop = (a == NULL) ? 0 : 1;
    if (Astop) lasermode (Exact_stop_on, f);

    IDent (f);
    if (inchout == 0) {		// Warning section
	fprintf (f, " ( ! DANGER!! -- Millimeter mode file! )\r\n");
	fprintf (f, " M0 ( DANGER!! Millimeter mode! )\r\n");
	fprintf (f, " ( Millimeters can cause a crash! )\r\n");
    }
    if (inchout) fprintf (f, " G20 ( Inch measurement )\r\n");
    else         fprintf (f, " G21 ( Metric measurement )\r\n");

    if (a == NULL) {
	c = make2vect (0.0, 323.0 * SF);
    } else {
/* Find start of object list; start at object following the one
  closest to the top edge of the mask where labels are put.
  Also, do optional reordering.  */
	aq = qstart (a->ob, M_imacs, opt);
// We start at point aq and end close to the label...
	c = mul2vect (aq->smpos, SF);
	c.x *= -1.0;	// Remember the x positions are reflected...
    }

    lasermode (Laser_cutting, f);
    lasermode (Laser_off, f);
    Feedrate (f, " G01", quick);
    Feol (f);

    tool = bwid / 2.0;
/* MAY LATER WANT TO SPECIFY AN ACTUAL TOOL SIZE... */

    last = c;
    h = zlast = maskz (last);
/* Work offset location, using 59 for IMACS here */
    fprintf (f, " G59  ( Work offset for IMACS )\r\n");
    IDent (f);
    fprintf (f, " G90");

    if (a == NULL) {		// flange case
/* Position vertically at a safe height, and then to xy location */
	Zoff += 3.5 * SF;	// As done for flange...
	h += 35.0*SF;	// Was maskz (last) = mask height
	Zout (f, h);  Feol (f);	// Over an inch up
	c.y -= 15.0 * SF;	// Put it over mask territory
/* NOTE Above is to get newloc to produce a z motion; there should be a
  better way, testing h values in newloc? */
	VXYout (f, c, 0);
	fprintf (f, " ( Start location )\r\n");
	IDent (f);

/* Move to the proper starting location from this known position;
  we start 4 mm right of first hole center, which should be on metal,
  and also at the correct height for flange cutting. */

	last.x += 4.5 * SF;	// displace right of hole to metal
	zlast = h = newloc (c, h, last, f);
// normally moves to height of 0.188 rather than just 0.050

/* Finally, select the gass and shutter, and go to the hole cutting */
	fprintf (f, " M10");	// select gas, 
	fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
	fprintf (f, " M14 ( Shutter open )\r\n");
	lastloc = last;
	goto Start_Holes;
    }	// flange holes feature

    VXYout (f, c, 1);
    lasermode (Autofocus_hold, f);
    Zout (f, h);
    fprintf (f, " ( Start location )\r\n");
    IDent (f);
// NOTE - this gas/shutter stuff could be a laser_mode call!!
    fprintf (f, " M10");	// select gas, 
    fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
    fprintf (f, " M14 ( Shutter open )\r\n");

//
// NOTE -- HELP!  IS H AND ZLAST IN MM OR IN?  SHOULD MULT BY SF OR NOT????
//
// Answer -- h is scaled to current measure; includes Z offset.

    lasermode (Autofocus_on, f);
    lastloc = last;

/* Here's what we do:

Find the absolute X,Y coordinate of one end of the slit,
and the differences to the other end.  Also make that
one side of the slit if it is wide, and the X,Y differences
to the other side.  These are used for the initial position,
and the incremental movement(s).  This is 3 pairs of x,y
values:  xc,yc for absolute corner, xs,ys for increment from
there along slit, and xw,yw for increment from there to the
other side of slit.  We move absolute to xc,yc, and then
incrementally by xs,ys.  If wide, move incrementally on the
remaining 3 sides: xw,yw; -xs,-ys; -xw,-yw.

These coordinates are computed from the center position, the
slit angle, and slit lengths.  Check slit width and compute
the xw,yw values, and add -0.5 * them to the first edge position.

  Make sure beam is off.
  Set G90 G00 for fast mode
  slew to one end of slit in X, Y
  switch to G01 G91 linear and incremental mode
  turn beam ON
  if offset exists, move to X, Y as an offset
  if slit is wider than tool, move to other side,
    then back to start increment, then to first side
(all done as increments, which makes it easy)
  turn beam OFF

.. */

/* One method of slowing down slit cutting so as to not overheat is to
  move to a safe z height between each slit; and at feed rate if we
  wish to really slow down a lot.  We put in slew speed z commands
  below to show where... */


	/* - - - - - - - - - - - - - - - *\
	|    Cutting Mask Object Holes    |
	\* - - - - - - - - - - - - - - - */

#ifdef  CUTREPEAT
    object_loop (f, aq, inst, 0.0, opt.repc);
#else
    object_loop (f, aq, inst, 0.0, 1);
#endif


	/* - - - - - - - - - - - - *\
	|    Engraving of Labels    |
	\* - - - - - - - - - - - - */

	lasermode (Laser_engraving, f);
	Astop = 0;
	if (!Astop) lasermode (Exact_stop_off, f);

//	sprintf (line, "%.8s  %c%c", a->fname, adc, AN);   // add ADC state
	if (serno > 0) sprintf (line, "%.8s  %d", a->fname, serno);
	else           sprintf (line, "%.8s",     a->fname);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number
// NOTE -- restore 6 mm letters when letter program has been
// revised to add Z axis control here.  Or maybe 5mm looks better
	basl = 4.5 * SF;		// 4 mm letters in field
	rad = 312.0 * SF - 2.0 * basl;	// Radial location of first line
	h = zlast;

// Start with line centered at 90 degrees
	xs = 0.5 * basl * strwid(line);	// half line length
	sincos (xs/rad, &sn, &cs);
	base = make2vect (basl*cs, basl*sn);
	loc = make2vect (-rad*sn, rad*cs);
	h = newloc (lastloc, h, loc, f);
	strplot (loc, base, -1.0/rad, 0.0, line, pltimacs, f);

// Second line is the astronomer name, centered below the file name
	sprintf (line, "%.16s%c", a->oname, AN);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number
	rad -= 1.7 * basl;
	xs = 0.5 * basl * strwid(line);
	sincos (xs/rad, &sn, &cs);
	base = make2vect (basl*cs, basl*sn);
	loc = make2vect (-rad*sn, rad*cs);
	h = newloc (lastloc, h, loc, f);
	strplot (loc, base, -1.0/rad, 0.0, line, pltimacs, f);
	lasermode (Autofocus_hold, f);

// Pull away by at least one inch (in hold mode) and turn autofocus off
	h += 30.0 * SF;
	fprintf (f, " G90 ");  Zout (f, h);  Feol (f);
// Alternate is relative motion without the G90...
	lasermode (Autofocus_off, f);

// Plot smaller label on flange, reading the opposite way,
// and both sides of the clearance hole...
// Start with things to the left of the hole...  astronomer, maskname
	basl = 4.0 * SF;			// 4 mm letters on flange
	sprintf (line, "%.16s  %.8s%c", a->oname, a->fname, AN);
	fprintf (f, "( Edge Label: %s )\r\n", line);	/* add comment */
	Gpause (f, 0.35);	// Allow autofocus to settle
	IDent (f);	// Debug line number
	rad = 323.0 * SF + 0.5 * basl;
	xs = basl * strwid(line) + 7.5 * SF;
	sincos (xs/rad, &sn, &cs);
	base = make2vect (-basl*cs, basl*sn);
	loc = make2vect (rad*sn, rad*cs);

/*  At this point, we are starting on the flange.  Since it can vary
  in height a little bit, we must increase the Z offset by about 3 mm
  and allow autofocus to pull down this much more.  */
	Zoff += 3.5 * SF;	// ????

	h = newloc (lastloc, h, loc, f);
	lasermode (Autofocus_on, f);	// Flange autofocus here...
	strplot (loc, base, 1.0/rad, 0.0, line, pltimacs, f);
// Keep autofocus on, as we move to new flange location

// Plot the astronomer name after hole, same way...
// Put in timestamp and then adc status...
//	sprintf (line, "%s  %c%c", tstamp, adc, AN);
	if (serno > 0) sprintf (line, "%s  %d", tstamp, serno);
	else           sprintf (line, "%s",     tstamp);
	fprintf (f, "( Edge Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number
	xs = 7.5 * SF;
	sincos (xs/rad, &sn, &cs);
	base = make2vect (-basl*cs, -basl*sn);
	loc = make2vect (-rad*sn, rad*cs);
	h = newloc (lastloc, h, loc, f);
	strplot (loc, base, 1.0/rad, 0.0, line, pltimacs, f);
	lasermode (Autofocus_off, f);
/* Do actual off, since we are an extra 3 mm above flange at z position */
	zlast = h;

/* An alternate way would be to use polar vector at (basl, 
  ang + 180.0*Degree) and convert to base, then for loc use
  polar vector at (rad, ang + 90.0*Degree).  */
/* Consider - specify basl, rad, angle, direction and get the
  base, loc and strplot done in a subroutine? */


	/* - - - - - - - - - - - - - - - *\
	|    Cutting Holes in Flange      |
	\* - - - - - - - - - - - - - - - */


/* Set back to cutting mode for holes in flange. */
    lasermode (Laser_cutting, f);

Start_Holes:

/* Set up the slit mask holes parameters... */
    ad.angle_adjust = 0.0;	// degrees, ccw
    ad.beam_diam = bwid / SF;	// Un-scaled beam width...
    ad.clearance = 0.020;	// mm, about .0008 inch

    Astop = 0;
    if (!Astop) lasermode (Exact_stop_off, f);

/* Change cutting of holes to make one pass around the rim, so as
  to more efficiently use the large slewing motions involved.
  This eliminates 2 loops, but interleaves their calls.  */

    IDent (f);		// Debug line number
    zlast = pvhole (323.0,  90.0, 7.0, zlast, ad, f);
    IDent (f);		// Debug line number
    a_slot         (323.0, 120.0, 6.0, 10.0,  ad, f);
    IDent (f);		// Debug line number
    zlast = pvhole (323.0, 150.0, 7.0, zlast, ad, f);
    IDent (f);		// Debug line number
    zlast = pvhole (323.0, 210.0, 7.0, zlast, ad, f);
    IDent (f);		// Debug line number
    a_slot         (323.0, 240.0, 6.0, 10.0,  ad, f);
    IDent (f);		// Debug line number
    zlast = pvhole (323.0, 270.0, 7.0, zlast, ad, f);
    IDent (f);		// Debug line number
    zlast = pvhole (323.0, 330.0, 7.0, zlast, ad, f);
    IDent (f);		// Debug line number
    a_slot         (323.0,   0.0, 6.0, 10.0,  ad, f);
    IDent (f);		// Debug line number
    zlast = pvhole (323.0,  30.0, 7.0, zlast, ad, f);

/* Calls above are in mm, and subroutines do the scaling */
/* The "pvhole" and "a_slot" calls include Autofocus on/off */

// Increased to clear clamps which can be 51 mm high
    h = maskz (make2vect(999.0, 999.0)) + 60.0 * SF;
    end_mask (f, h);

// Increment the serial number if used
    if (serno > 0) serno++;

    return  1;

}



/* ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- */


	/* - - - - - - - - - - - - - - - - *\
	|                                   |
	|    LDSS-2 laser mask cutting      |
	|                                   |
	\* - - - - - - - - - - - - - - - - */

/*  ----  Software for LDSS-2 laser cutting  ----  */
/* -- this one is for the obsolete flat masks -- */


// removing pltldss - no longer needed to plot characters on mill


int	cutldss (Obs* a, FILE* f, opts opt)
/* Do the same (similar) things for the LDSS-2 mask case */
/* Return 1 if succeed in outputing LDSS-2 cutting commands
  to the file OUT  */
{

char	line[80];
double	tool;
vect2	base;
vect2	loc;
double	xs;	// Line length
double	basl;	// Base length
double	x, y;	// Location for legend plot

/* The following are not supposed to be needed; included to remove
  the error messages on code imported from IMACS case.  All the slew
  logic should be replaced by a new program anyway. */
vect2	last;	// may not be really needed?
double	h;  // also.., nh, dh;
double	zlast;
/* All these above should not be needed; but removes the error messages. */

    if (a->dlevel > 0) printf ("  Cutting LDSS-2 mask... \n");

/* Initializing machine */

    lasermode (Autofocus_off, f);

    Astop = 1;
    if (Astop) lasermode (Exact_stop_on, f);

    IDent (f);
    if (inchout == 0) {		// Warning section
	fprintf (f, " ( ! DANGER!! -- Millimeter mode file! )\r\n");
	fprintf (f, " M0 ( DANGER!! Millimeter mode! )\r\n");
	fprintf (f, " ( Millimeters can cause a crash! )\r\n");
    }
    if (inchout) fprintf (f, " G20 ( Inch measurement )\r\n");
    else         fprintf (f, " G21 ( Metric measurement )\r\n");

    lasermode (Laser_cutting, f);
    lasermode (Laser_off, f);
    Feedrate (f, " G01", quick);
    Feol (f);

    last = make2vect (0.0, 0.0);
    h = zlast = maskz (last);

    fprintf (f, " G58  ( Work offset for LDSS )\r\n");
    IDent (f);
    fprintf (f, " G90 X0 Y0\r\n");

    lasermode (Autofocus_hold, f);
    Zout (f, h);
    fprintf (f, " ( Start location )\r\n");

// Work Offset -- currently using 59 for IMACS,
// and 58 for LDSS2 locations.  Other ones may be
// specified later.

    IDent (f);
    fprintf (f, " M10");	// select gas, 
    fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
    fprintf (f, " M14 ( Shutter open )\r\n");

    lasermode (Autofocus_on, f);
    lastloc = last;


    tool = bwid / 2.0;
/* MAY LATER WANT TO SPECIFY AN ACTUAL TOOL SIZE... */

/* Here's what we do:

Find the absolute X,Y coordinate of one end of the slit,
and the differences to the other end.  Also make that
one side of the slit if it is wide, and the X,Y differences
to the other side.  These are used for the initial position,
and the incremental movement(s).  This is 3 pairs of x,y
values:  xc,yc for absolute corner, xs,ys for increment from
there along slit, and xw,yw for increment from there to the
other side of slit.  We move absolute to xc,yc, and then
incrementally by xs,ys.  If wide, move incrementally on the
remaining 3 sides: xw,yw; -xs,-ys; -xw,-yw.

These coordinates are computed from the center position, the
slit angle, and slit lengths.  Check slit width and compute
the xw,yw values, and add -0.5 * them to the first edge position.

  Make sure beam is off.
  Set G90 G00 for fast mode
  slew to one end of slit in X, Y
  switch to G01 G91 linear and incremental mode
  turn beam ON
  if offset exists, move to X, Y as an offset
  if slit is wider than tool, move to other side,
    then back to start increment, then to first side
(all done as increments, which makes it easy)
  turn beam OFF

  (above comments same as IMACS case)

.. */


	/* - - - - - - - - - - - - - - - *\
	|    Cutting Mask Object Holes    |
	\* - - - - - - - - - - - - - - - */

// NOTE -- SINCE IMACS AND LDSS BOTH DO ESSENTIALLY THE SAME THINGS
// HERE, CONSIDER MOVING THESE COMMANDS TO A SUBROUTINE
// ALSO CONSIDER THAT FOR THE INITIALIZATION ABOVE!!

// TRANSFERRING IMACS STUFF TO HERE
// DO NOT USE THE "BACK OF MASK" ITEMS FROM IMACS...
// MAKE SURE SF IS SET HERE

#ifdef  CUTREPEAT
    object_loop (f, a->ob, inst, 0.0, opt.repc);
#else
    object_loop (f, a->ob, inst, 0.0, 1);
#endif


// LDSS ENGRAVING - LDSENG

	/* - - - - - - - - - - - - *\
	|    Engraving of Labels    |
	\* - - - - - - - - - - - - */

	lasermode (Laser_engraving, f);

// Engrave astronomer name in upper left...
// Change that to lower left...
//	sprintf (line, "%.16s%c", a->oname, AN);
	if (serno > 0) sprintf (line, "%.16s %d", a->oname, serno);
	else           sprintf (line, "%.16s",   a->oname);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number
	basl = 4.0 * SF;		// 6 mm letters in field
	y = -4.5 * 0.5 * Inch * SF;	// Bottom of mask
	y += 0.5 * basl;		// Bottom of line
	x = -6.0 * 0.5 * Inch * SF;	// Left of mask
	x += 1.5 * basl;
	loc = make2vect (x, y);
	base = make2vect (basl, 0.0);	// Base of characters
	h = newloc (lastloc, h, loc, f);
	strplot (loc, base, 0.0, 0.0, line, pltimacs, f);

// Engrave the file name in upper right...
// Similarly, lower right
	sprintf (line, "%.8s%c", a->fname, AN);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number
	x = 6.0 * 0.5 * Inch * SF;	// Right of mask
	x -= 1.5 * basl;
	xs = basl * strwid(line);	// line length
	x -= xs;			// Start of line
	loc = make2vect (x, y);
	h = newloc (lastloc, h, loc, f);
	strplot (loc, base, 0.0, 0.0, line, pltimacs, f);
//	lasermode (Autofocus_hold, f);
// Final autofocus off is real...
// autofocus hold is done in end_mask

// THE LDSS MASKS DO NOT NEED MOUNTING HOLES CUT AS WAS DONE ON
// THE IMACS MASKS AT THIS POINT.

    h = 4.0 * Inch * SF;
    end_mask (f, h);
    if (serno > 0) serno++;

    return  1;
}



/* ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- */

	/* - - - - - - - - - - - - - - *\
	|				|
	|   LDSS-3 multi-mask cutting	|
	|				|
	\* - - - - - - - - - - - - - - */

vect2	circloc (double rad, double angl) {
double	s, c;
vect2	result;
    sincos (angl, &s, &c);
    result.x = c * rad;
    result.y = s * rad;
    return result;
}

// MAKE THE ABOVE INTO A KDCUTIL ROUTINE, WITH RADIUS AND ANGLE
// ARGUMENTS WHICH RESULT IN A 2-VECTOR.  CALL IT P2RECT AND
// BASE IT ON THE RECT2 ROUTINE THERE.

/* Moved "nsides" from here to cututil.c */

static	int	cutldss3 (Obs* a, FILE* f, opts opt)
{
// cut a single mask with global offsets
// called only by cutmulti, which follows here...
int	n;
double	tool;
vect2	c;	// location of center
static	vect2	last = {999.0, 999.0};	// Absolute last loc for slew
static	double	zlast;	// For incremental z movement...
double	w;	// may not be needed??
objq	*aq;	// Start object
char	line[80];
vect2	loc, base;	// Label location, direction/size
double	rad, basl;	// Label radius, base size
double	sn, cs;		// Location sin,cos values
double	xs;		// Label size
double	h;		// Z-altitude
double	fr;

double	ba;		// base angle
double	ga;		// gap angle
double	aseg;		// angle of segment
double	ai;		// angle increment per stroke
double	swa;		// start, end wafer angle
int	stk;		// stroke count
int	js;		// stroke index
int	ns;		// segment index

// Return 0 if null inputs.
    if (a == NULL) return 0;
    if (f == NULL) return 0;

// The global offsets are already set here.

/*  Outline of what is needed:
  Do traveling salesman on the object list
  Loop over all objects; cut those objects
  Engrave the proper legend around edge
  Cut the border and locating notch
  Return 1 if done OK
.. */

/* Check for null object count... */
    n = listcount (a->ob);
    if (n < 1) {
	printf (" ** No mask objects in list, nothing written.\n");
	return  0;
    }

// COPY THE ABOVE THINGS FROM cutimacs PROGRAM ABOVE.
// BASIC INITIALIZATION HAS ALREADY BEEN DONE
// FINAL EXIT THINGS WILL BE DONE AFTER RETURN.
// ONLY THE CUTTING OF MASK FEATURES IS DONE HERE, FOLLOWED
// BY AN ELEMENTARY LABEL.
// RESTORE CUTTING FROM ENGRAVING MODE AFTER DOING THE LABEL.

    aq = qstart (a->ob, M_lds3, opt);	// reorder, find start
// We start at point aq and end close to the label...
    c = mul2vect (aq->smpos, SF);
    c.x *= -1.0;	// Remember the x positions are reflected...

/* Inform user of mask being cut.. */
//    printf (" Submask %d is \"%.8s\" for \"%.16s\".\n",
//	opt.subm,  a->fname,  a->oname );
    if (serno > 0) printf (" Mask %d.%d is \"%.8s\" for \"%.16s\".\n",
	serno, opt.subm,  a->fname,  a->oname );
    else           printf (" Submask %d is \"%.8s\" for \"%.16s\".\n",
	opt.subm,  a->fname,  a->oname );

/* Identify the sub-mask in the output file */
//    fprintf (f, "( \"%.8s\" for \"%.16s\". )\r\n", a->fname, a->oname);
    if (serno > 0) fprintf (f, "( Mask %d.%d \"%.8s\" for \"%.16s\". )\r\n", 
	serno, opt.subm, a->fname, a->oname);
    else           fprintf (f, "( Submask %d \"%.8s\" for \"%.16s\". )\r\n", 
	opt.subm, a->fname, a->oname);

    lasermode (Laser_cutting, f);
    lasermode (Laser_off, f);
    Feedrate (f, " G01", quick);
    Feol (f);

    tool = bwid / 2.0;
/* MAY LATER WANT TO SPECIFY AN ACTUAL TOOL SIZE... */

    last = c;
    h = zlast = maskz (last);
/* Work offset location, using 59 for IMACS here */
    fprintf (f, " G59  ( Work offset for Multi-mask )\r\n");
    IDent (f);
    fprintf (f, " G90");

    VXYout (f, c, 1);
    lasermode (Autofocus_hold, f);
    Zout (f, h);
    fprintf (f, " ( Start location )\r\n");
    IDent (f);
// NOTE - this gas/shutter stuff could be a laser_mode call!!
    fprintf (f, " M10");	// select gas, 
    fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
    fprintf (f, " M14 ( Shutter open )\r\n");

// Still copying stuff from cutimacs here...

    lasermode (Autofocus_on, f);
//    lasermode (Autofocus_release, f);
    lastloc = last;

    Astop = 1;
    if (Astop) lasermode (Exact_stop_on, f);

/* Loop will compute all output things, just as for IMACS */

	/* - - - - - - - - - - - - - - - *\
	|    Cutting Mask Object Holes    |
	\* - - - - - - - - - - - - - - - */

#ifdef  CUTREPEAT
    n = object_loop (f, aq, inst, 0.0, opt.repc);
#else
    n = object_loop (f, aq, inst, 0.0, 1);
#endif

/* Switch to engraving mode and cut a label */

	/* - - - - - - - - - - - - *\
	|    Engraving of Labels    |
	\* - - - - - - - - - - - - */

// NOTE -- CHANGE THESE FOR THE MULTIMASK CASE.

	lasermode (Laser_engraving, f);

	Astop = 0;
	if (!Astop) lasermode (Exact_stop_off, f);

//	sprintf (line, "%.8s -%d- %.16s", a->fname, opt.subm, a->oname);
	if (serno > 0) sprintf (line, "%.8s %d.%d %.16s", a->fname, 
			serno, opt.subm, a->oname);
	else           sprintf (line, "%.8s -%d- %.16s", a->fname, 
			opt.subm, a->oname);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number

// Write small on the small masks...
	basl = 2.6 * SF;		// 3 mm letters in field

// Write line centered at angle ALT (Angle LDSS Title) ccw from East
#define  ALT  180.0
// Line to be within radius RLT (Radius LDSS Title)
#define  RLT  (175.0/2.0)
// Line to be readable at bottom (1.0) or top (-1.0) position
#define  CVL  -1.0

// Move these definitions up before start of plotting;
// We want to start/end the objects nearest the "loc" location!

	h = zlast;
	xs = 0.5 * basl * strwid(line);	// half line length
	rad = RLT * SF - ( (CVL > 0.0) ? 0.2 : 1.1 ) * basl;
	sincos ( ALT*Degree - CVL*xs/rad, &sn, &cs);
	base = mul2vect (make2vect(-sn, cs), CVL*basl);
	loc = mul2vect (make2vect(cs, sn), rad);
	h = newloc (lastloc, h, loc, f);
	strplot (loc, base,  CVL/rad, 0.0, line, pltimacs, f);

	zlast = h;

// ABOVE IS COPY OF IMACS, NEEDS CHANGING

	/* - - - - - - - - - - *\
	|			|
	|  Cut alignment holes	|
	|			|
	\* - - - - - - - - - - */

    lasermode (Laser_cutting, f);
    fprintf (f, "( Alignment Holes. )\r\n");
    IDent (f);

// If offset is 0, use diameters of 2.45 and 1.65 here;
// For offset of 200 use 2.50 and 1.70 due to material thickness,
// and slant of drilled hole.
//
// The pin diameters are 2.38 and 1.59 mm.
// Clearance of 0.1 mm for tolerances and 0.05 for tilt angles
// is to be added:  2.48/2.53 mm and 1.69/1.74 mm will be used.
// The drawing shows 2.46 and 1.67 here; we reduce all by 0.02 mm.
// Cutting on the slope will still add 0.05 mm for material thickness.
    c = mul2vect (make2vect(91.0, 0.0), SF);
    w = ( opt.subm ? 1.73 : 1.69 ) * SF;	// Was 1.72, 1.67
    zlast = chole (c, w-bwid, zlast, slow, 1, f);

    c = mul2vect (make2vect(69.30, 58.98), SF);
    w = ( opt.subm ? 2.51 : 2.46 ) * SF;
    zlast = chole (c, w-bwid, zlast, slow, 1, f);

//    lasermode (Laser_off, f);
// Removed above, the previous chole includes laser off.

/* Cut an appropriate border, with breaks and a notch */
// Going to do ngaps (here 6) segments
// Cut them in a circle which will get automatically changed by
// the offset programs.
// Be able to put in a notch later...


	/* - - - - - - - - - - *\
	|			|
	|    Cutting of edge	|
	|			|
	\* - - - - - - - - - - */


    fprintf (f, "( Edge of sub-mask. )\r\n");

// Compute number of strokes and angle of each one
#define  nsegs  6
#define  gap    0.20
// gap was 0.3 mm, could stand to be less; material is .3 mm thick.

// reminder - inch SF=1/25.4, metric SF=1
// value-in-mm * SF = scaled-value

// bwid is the scaled beam width here.

    rad = maskrad * SF;
    rad += bwid / 2.0;
    ga = gap * SF / rad;	// gap angle
    aseg = TwoPi / nsegs;	// angle of segment
    stk = nsides (rad, 0.3*bwid, aseg - ga);	// strokes
    ai = (aseg - ga) / stk;	// angle increment
    ba = (M_PI + ga) / 2.0;	// basic starting angle
	
/* Slew to the starting point */
    c = circloc (rad, ba);
    h = newloc (lastloc, zlast, c, f);

/* Set rough cutting mode and feed rate for cutting the edge */
//    lasermode (Laser_roughcut, f);	// Cut at high speed
    lasermode (Laser_cutting, f);	// Change to regular cut.
//    fr = quick * 0.5;
    fr = slow;		// "cut" rate, should get 10.0 in/min
    IDent (f);
/* Set the feed rate for edge */
//    Feedrate (f, " G90", fr);
    Feedrate (f, "", fr);
// FEEDRATE ABOVE WILL BE SET TO A NEW VALUE AS FOUND
// FROM SOME EXPERIMENTS;  USE QUICK FOR NOW...
    Feol (f);

/* Loop to cut proper number of segments */
    for (ns=0; ns < nsegs; ns++) {
// Move to start of segment location (use newloc, track lastloc and h)
// lastloc is global and set in newloc
	SetSlew = 0;	// Do not change feed rate...
	swa = ba + aseg * ns;
	c = circloc (rad, swa);
	h = newloc (lastloc, h, c, f);

// Turn on the laser
	lasermode (Laser_on, f);
	n++;		// the rim segment is a feature!

// Loop to cut the segment, using proper number of strokes
// Gap is put at end of segment.
	for (js=1; js <= stk; js++) {
	    c = circloc (rad, swa + js*ai);
	    h = newloc (lastloc, h, c, f);
	}	// Loop for cutting segments
// Turn laser off at end of loop
	lasermode (Laser_off, f);
// Mode had been Laser_roughcut, we now drop that rough behavior
    }	// Loop for cutting segments
//    lasermode (Laser_cutting, f);	// Restore regular mode


/* THE WAY IT REALLY IS...
  Is that there are actually 2 holes for 1/16 and 3/32 SS dowel pins
  located near the rim.  These are cut before we cut the rim itself,
  in the section above.  The rim is then perfectly circular */

// End the cutting of the masks, end of the last or 7th mask
// Must stop the cut, turn laser off (done already)
// and then hold autofocus, raise up one inch, and turn autofocus off.
// ADD THIS TO MASKCUT ALSO.

    lasermode (Autofocus_hold, f);
    lasermode (Laser_cutting, f);	// Restore regular mode
// Pull away by at least one inch (in hold mode) and turn autofocus off
    h += 30.0 * SF;
    fprintf (f, " G90 ");  Zout (f, h);  Feol (f);

/* Everything else is taken care of outside this program */

    return  n;	// count of features
}


int	nexfile (char* base, char* ext) {
/* I/O utility.  Make up name of form base|xx|ext such that the
  numeric value x of a not-yet-used file name is returned. */
int	i;
char	buf[80];

    if (base == NULL) return 0;
    if (ext  == NULL) return 0;

    for (i=1; i<100; i++) {
	sprintf (buf, "%s%02d%s", base, i, ext);
	if (access (buf, F_OK)) return  i;
    }
    return  0;
/* Zero is returned if a suitable file name is not found */
}


	/* - - - - - - - - - - - - - - *\
	|				|
	|      Cut  LDSS-3  masks	|
	|				|
	\* - - - - - - - - - - - - - - */

//#define  SUBMASKS  5
#define  SUBMASKS  7
// Temporarily restricted due to laser machine problem

// int	cutmulti (sortq** mmhead, opts opt)
int	cutmul_ldss (sortq** mmhead, opts opt)
{
int	masks = 0;
// int	files = 0;
int	i, j;
int	n;
int	sm;
//int	good;
sortq	*p;	// was.. , *q;
Obs	*obs;
FILE	*out;
char	base[8];
char	extn[4];
char	filename[16];
//struct	tm	*ut;
//time_t	now;

double	sdz;
double	vert;
double	h;

// Return if nothing to be done
    if (mmhead == NULL) return 0;	// No argument

    if (opt.dbug > 0)
	printf (" cutmul_ldss for %d mask files.\n", listcount (*mmhead) );

//    if (*mmhead == NULL) return 0;	// Null queue

/* We appear to have a queue, we need to process it */
    if (opt.subc != SUBMASKS && opt.dbug > 0) 
	printf (" *Cutting %d submasks in each blank.\n", opt.subc);

/* Outline of what is to be done:

Process the queue in blocks of 7 obs pointers.
For each block we will:
  0) Set up the offset to 0, 0.
  1) Obtain a new file name
  2) Write the initial commands into that file
	>> Copy set-up from other cut programs
	>> Position cut head in center of first mask
  3) Do the individual processes for each of up to 7 pointers
	>> Pop the top obs file from the queue (break if none)
	>> Do a loop over all mask features
	>> Write the label information on periphery of mask
	>> Do commands to cut the perimeter and notch
	>> kill the obs structure and the queue member
	>> Set offsets for the next location
	>> Count another sub-mask processed
  4) Write the terminal commands into the file, close the file
	>> Retract cutting head, turn off, move to output position
	>> Close file
	>> Count another file output.
If more obs files exist, return to step 1.

The whole thing is wrapped in a for loop on *mmhead not being null.
If it becomes null within the loop, break.

.. */

    SF = inchout ? 1.0/Inch : 1.0;	// Make sure it's set.
    while (*mmhead != NULL) {

/* Set up offset to be 0,0 initially */

	setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */

/* Obtain file name and open the new file */

/* If a pointer to a valid <= 6 character base string has been provided
  in the options, use that.  Otherwise, generate a base string from
  the current date to be used here.  */
// Not using strlen here to check length...

	if (opt.base != NULL) {
	    sprintf (base, "%.6s", opt.base);  // First up to 6 char.s
	} else {
// GET THE BASE PART OF FILE NAME (SEE OUTLINE) LyyMddxx
	    sprintf (base, "L%s", curdate(0));
	}
	sprintf (extn, "%s", ".nc");

//  e.g.   n = nexfile (base, extn);
// where base is LyyMdd and extn is ".nc";
// then: sprintf (filename, "%s%02d%s", base, n, extn);
	j = nexfile (base, extn);
	if (j < 1) {
	    printf (" ** No available file name \"%s--%s\"; no output.\n",
		base, extn);
	    return masks;
	}
	sprintf (filename, "%s%02d%s", base, j, extn);
// FIND NEXT SERIALIZED NAME WHICH DOES NOT YET EXIST
// OPEN THE NEW FILE FOR WRITING
	out = iopen (filename, "w+");

// IF OUTPUT FILE IS NOT OPENED CORRECTLY, WE NEED TO WRITE
// AN ERROR MESSAGE AND EXIT.
	if (out == NULL) {
	    printf (" ** Open error; no further output.\n");
	    return  masks;
	}

	Lnum = 0;		// IDent counter
/* Initial set-up things copied from cutmask */
// set the Z offset
	inst = M_lds3;	// Special case, this is ldss-3 instrument
//	inst = I_lds3;	// Special case, this is ldss-3 instrument
#ifdef  USE_CDAT
	sdz = CDhead[inst].dz * SF;
#endif
#ifdef  USE_GSTO
	sdz = *(double*)GS_lookup(&CDhead,inst,CD_dz,0) * SF;
#endif
#ifdef  USE_BSTO
	sdz = floatvalue (getval (&CDhead,MaskCase(inst),"dz")) * SF;
#endif
	Zoff = 0.0;
/* Above was commented out, as Zoff is set in cututil.h where it is
  defined.  Actually needed dynamically when cutting multiple blanks,
  so it is not redefined to another height below.  Also, set this for
  real in the other routines which set it so multiple masks are done
  correctly with the height computed each time...    */

	vert = maskz (make2vect ( SF*(flange+1.0), 0.0) );
// Offset is same as IMACS case
	Zoff = sdz - vert;
	if (opt.dbug > 0)  printf (" --Zoff = %7.3f; Z0 = %5.3f, Vh = %6.3f\n", 
				Zoff, sdz, vert);


/* Write initial commands into the file */
// WRITE INITIAL COMMENTS AND COMMANDS TO FILE
// GET THESE FROM OTHER OUTPUT FILE INSTANCES.
	if (opt.dbug >= 0) printf ("  Cutting multi-mask %s \n", filename);
/* Some debug stuff */
#if (DEBUG_CUT > 0)
	if (opt.dbug > 0) {
#ifdef  SIMPLECURVE
	    printf ("  Inst = %d, curve = %7.5f\n", inst, Curve );
#else
vect2	cv;
//		*GS_lookup(&CDhead,inst,CD_curve) );	// this doesn't work!
#ifdef  USE_CDAT
		cv = CDhead[inst].curve;
#endif
#ifdef  USE_GSTO
		cv = *(vect2*)GS_lookup(&CDhead, inst, CD_curve, 0);
#endif
#ifdef  USE_BSTO
		cv = vect2value(getval(&CDhead,MaskCase(inst),"Curve"));
#endif
	    printf ("  Inst = %d, curve = %7.5f (x), %7.5f (y)\n", inst, 
		cv.x, cv.y);
#endif  // on SIMPLECURVE
	}
#endif  // on DEBUG_CUT

// SET UP THE CUTTING PARAMETERS AS GLOBALS

// INCLUDE VERSION NUMBER OF THIS PROGRAM
//	fprintf (out, "( Multi-mask file by %s version %s%s )\r\n",
//		PROGNAME, VERSION, CTUVERS);
	fprintf (out, "( Multi-mask file by %s version %s )\r\n",
		PROGNAME, XVERSION);
// Could split above into two lines if needed for compatability
//	fprintf (out, "( Instrument: %s )\r\n", InstName(inst) );

	if (serno > 0) fprintf (out, "( Instrument: %s  Mask: %d )\r\n", 
		MaskName(inst), serno );
	else           fprintf (out, "( Instrument: %s )\r\n", MaskName(inst) );

/* Set up global cutting parameters */
	cutparams (inst);

//	if (opt.dbug > 0) printf (" LDSS3 multi -- Acon = %.3g\n", Acon);
	if (opt.dbug > 0) printf (" LDSS3 cutpar (%s) -- Acon = %.3g\n", 
			MaskCase(inst), Acon);

/* If we have a request for slow cutting, modify the slow and vslow
  rates to slower values for fine cutting... */
	if (opt.slow) {
double	ratio;
	    ratio = vslow/slow;
	    slow = vslow;
	    vslow = ratio*slow;
	}


/* Initial cutting commands, from cutimacs */
	lasermode (Autofocus_off, out);
	lasermode (Exact_stop_on, out);
	IDent (out);
	if (inchout == 0) {		// Warning section
	    fprintf (out, " ( ! DANGER!! -- Millimeter mode file! )\r\n");
	    fprintf (out, " M0 ( DANGER!! Millimeter mode! )\r\n");
	    fprintf (out, " ( Millimeters can cause a crash! )\r\n");
	}
	if (inchout) fprintf (out, " G20 ( Inch measurement )\r\n");
	else         fprintf (out, " G21 ( Metric measurement )\r\n");
/* Further commands from imacs cutting are in cutldss3 above */

/* Process up to 7 individual sub-masks into the output file */
//	for (i=j=sm=0; i<SUBMASKS; i++) {	// loop over up to 7 sub-masks
	for (i=j=sm=0; i<opt.subc; i++) {	// loop over up to 7 sub-masks

// SET GLOBAL OFFSETS BASED ON INDEX i HERE.
// if i is 0, use 0, 0;
// Otherwise, use 200.0, 60.0 * (i-1)
	    if (i == 0) {				// Center sub-mask
		setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */
	    } else {
		setdisp (suboff, 0.0, 60.0 * (i-1), 0);  // off-center mask
	    }
	    opt.subm = i;		// sub mask designator

	    p = *mmhead;
	    if (p == NULL) break;
	    *mmhead = pop_sq (*mmhead);
	    obs = (Obs*) p->q;
/* Process the sub-mask, and count our success */
//	    j += cutldss3 (obs, out, opt);
// CHANGE THAT TO COUNT TOTAL FEATURES DONE
// SUM BOTH TOTAL FEATURES, AND NUMBER OF SUB-MASKS WITH >0 FEATURES
// REPORT NUMBER OF FEATURES ON EACH SUB-MASK

/* SHOULD SET SF HERE DUE TO THE CTE --- */
#ifdef  USE_CTE
	    if (opt.xcte) {
double	tex;
		SF = inchout ? 1.0/Inch : 1.0;
		tex = CTE * (Tcut - obs->temp);
		if (opt.dbug > 0 && tex != 0.0)
		    printf (" Correcting thermal expansion of %9.3e\n", tex);
		SF *= 1.0 + tex;
	    }
#endif

	    n = cutldss3 (obs, out, opt);
	    if (n < 1) {	// No mask cut, decrement i to try again
		i--;
	    } else {	// cut a mask, total the mask and feature count
		j += n;		// features count
		sm++;		// sub-mask count
	    }

// IF ZERO FEATURES DONE (INCLUDING LABEL) DECREMENT i TO EXTEND LOOP

/* Done with sub-mask, kill the obs structure and the sortq entry p */
	    obs = kill_obs (obs);
	    free (p);

	}	// Loop over input files indexed by i, 0-6

	setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */

/* Count masks, files output */
//	masks += j;
	masks += sm;

/* If any submasks written, fill rest of file.  Close file */
	if (j > 0) {		// Some entries
//// Increased to clear clamps which can be 51 mm high
	    h = maskz (make2vect(999.0, 999.0)) + 60.0 * SF;
	    end_mask (out, h);

#ifdef  USE_ACON
	    fill_cuttime (out);
#endif
// CLOSE FILE

/* Add checksum and signature */
	    signfile (out, '(', dodebug-1);

	    iclose (out);
	    printf (" -Processed %d features in %d sub-masks.\n", j, sm);

// increment serial number if used
	    if (serno > 0) serno++;

// COUNT FILES -- already done?
	} else {		// No entries
// CLOSE AND DELETE FILE, IT IS NULL
// rm remove rename unlink
// unlink requires unistd include; already needed for access.
	    iclose (out);
	    unlink (filename);	// Deletes the file
	}	// Null mask output test

    }	// Null ldss mask queue test

// Return number of masks processed
    setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */
    return  masks;	// ALSO NEED TO RETURN FEATURES?
}	// end cutmul_ldss




/* ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- */

	/* - - - - - - - - - - - - - - *\
	|				|
	|      Cut  GISMO  masks	|
	|				|
	\* - - - - - - - - - - - - - - */


static	int	cutgismo (Obs* a, FILE* f, opts opt)
{
// cut a single mask with global offsets
// called only by cutmulti, which follows here...

int	n=0;
double	tool;
vect2	c;	// location of center
objq	*aq;	// queue start?

static	vect2	last = {999.0, 999.0};	// Absolute last loc for slew

static	double	zlast;	// For incremental z movement...

char	line[80];
vect2	loc, base;	// Label location, direction/size
double	basl;	// base size

double	h;		// Z-altitude

// double	Slide = 11.0;	// Offset in mm per mask
// Value changed per engineering test 07/06-15
const	double	Slide = 12.0;	// Offset in mm per mask
double	dy=0.0;		// (global) offset in y for the submask
vect2	dv;		// y offset in vector form (unused)


// Return 0 if null inputs.
    if (a == NULL) return 0;
    if (f == NULL) return 0;

// The global offsets are already set here.

// Compute local y offset from opt.subm value:
#ifdef  GISMOFFSET
    dy = (opt.offset + Slide * (opt.subm - 2)) * SF;	// global y offset
#else
    dy = Slide * (opt.subm - 2) * SF;	// global y offset
#endif
// add "dy" to all y values output by this subroutine!
// Also make a 2vector displacement
    dv = make2vect (0.0, dy);

/*  Outline of what is needed:
  Do traveling salesman on the object list
  Loop over all objects; cut those objects
  Engrave the proper legend around edge
  Cut the border and locating notch
  Return 1 if done OK
.. */

/* Check for null object count... */
    n = listcount (a->ob);
    if (n < 1) {
	printf (" ** No mask objects in list, nothing written.\n");
	return  0;
    }

/* Adding debug output of basic z offset stuff.  */
    if (opt.dbug > 0) {
	printf (" *= Z offset = %.3f\n", maskz (dv) );
    }

// COPY THE ABOVE THINGS FROM cutimacs PROGRAM ABOVE.
// BASIC INITIALIZATION HAS ALREADY BEEN DONE
// FINAL EXIT THINGS WILL BE DONE AFTER RETURN.
// ONLY THE CUTTING OF MASK FEATURES IS DONE HERE, FOLLOWED
// BY AN ELEMENTARY LABEL.
// RESTORE CUTTING FROM ENGRAVING MODE AFTER DOING THE LABEL.

    aq = qstart (a->ob, M_gismo, opt);
// Optionally reorders and finds the starting point.

// NEEDS TO CHANGE FOR GISMO CASE
    c = mul2vect (aq->smpos, SF);
    c.x *= -1.0;	// Remember the x positions are reflected...
    c.y += dy;

/* Inform user of mask being cut.. */
    printf (" Submask %d is \"%.8s\" for \"%.16s\".\n",
	opt.subm+1,  a->fname,  a->oname );

/* Identify the sub-mask in the output file */
    fprintf (f, "( [%d] \"%.8s\" for \"%.16s\". )\r\n",
	opt.subm+1, a->fname, a->oname);

    lasermode (Laser_cutting, f);
    lasermode (Laser_off, f);
    Feedrate (f, " G01", quick);
    Feol (f);

    tool = bwid / 2.0;
/* MAY LATER WANT TO SPECIFY AN ACTUAL TOOL SIZE... */

    last = c;
    h = zlast = maskz (last);
/* Work offset location, using 59 for IMACS here */
    fprintf (f, " G59  ( Work offset for Multi-mask )\r\n");
// NEEDS TO CHANGE FOR GISMO CASE
    IDent (f);
    fprintf (f, " G90");

/* Count a feature for the initialization data output to f here. */
    n++;

    VXYout (f, c, 1);
    lasermode (Autofocus_hold, f);
    Zout (f, h);
    fprintf (f, " ( Start location )\r\n");
    IDent (f);
// NOTE - this gas/shutter stuff could be a laser_mode call!!
    fprintf (f, " M10");	// select gas, 
    fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
    fprintf (f, " M14 ( Shutter open )\r\n");

// Still copying stuff from cutimacs here...

    lasermode (Autofocus_on, f);
//    lasermode (Autofocus_release, f);
    lastloc = last;

    Astop = 1;
    if (Astop) lasermode (Exact_stop_on, f);

/*  We probably want something like this:

    maskinit (f, M_gismo, a, opt);
// although that above should be in the multicut calling program...

    cutstart (f, M_gismo, c);
// followed by the object loop;
// maybe put cutstart at start of object loop?

... */

/* Loop will compute all output things, just as for IMACS */

	/* - - - - - - - - - - - - - - - *\
	|    Cutting Mask Object Holes    |
	\* - - - - - - - - - - - - - - - */

#ifdef  CUTREPEAT
    n += object_loop (f, aq, inst, dy, opt.repc);
#else
    n += object_loop (f, aq, inst, dy, 1);
#endif

/* Switch to engraving mode and cut a label */

	/* - - - - - - - - - - - - *\
	|    Engraving of Labels    |
	\* - - - - - - - - - - - - */
// NEEDS TO CHANGE FOR GISMO CASE

// Be sure to add the offset dy or vector dv to origin of
// all the label plotting calls.

// We probably want to put the labels in the upper left,
// use a small font, and limit size; 2 lines of
// about 6.5 to 7 mm total height, and 50 mm total width

/*  Label should look like one of these:

    |
    |  5 Mask0030
    |    Smythehofenstein
    |
    |  4 Smythehofenstein
    |    Mask0027
    |
    |  3 Mask0024
    |  Smythehofenstein
    |
    |  Smythehofenstein
    |  2 Mask0022

.. */

// Also, the cutmul_gis outside routine should put on a mask
// label in the upper corner containing the mask file name
// and the date created:
//  - - - - - - - - - -
//  |  G123V65.nc  2006-Oct-31
//  |
//  |  Astronomername
//  |  5 Mask0021

// NOTE -- CHANGE THESE FOR THE MULTIMASK CASE.

// SWITCH TO ENGRAVE MODE!

#define  GISLABLX   8.0
#define  GISLABLY  19.0
/* The label is X to the right and Y below an arbitrary location in
  the upper left defined by the mask dimensions and the sub-mask index  */
/* The Y value had been 24.0; changed 2010/01-26 to avoid conflict between
  the maximum size of aperture mask in 5th position and the first label.  */

	lasermode (Laser_engraving, f);

	Astop = 0;
	if (!Astop) lasermode (Exact_stop_off, f);

	sprintf (line, "%d %.8s", opt.subm + 1, a->fname);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number
	basl = 2.6 * SF;	// 3 mm letters
	base = make2vect (basl, 0.0);
	loc = mul2vect ( make2vect ( -0.5 * 350.0 + GISLABLX, 
#ifdef  GISMOFFSET
	    (0.5 * 326.5 - GISLABLY) + opt.offset + (opt.subm - 4) * Slide), SF);
#else
		(0.5 * 326.5 - GISLABLY) + (opt.subm - 4) * Slide), SF);
#endif
// or add dy to a nominal location, or dv?
// Better, (0.5 * 326.5 - 25.0), SF); loc = sum2vect (loc, dv);
	strplot (loc, base, 0.0, 0.0, line, pltimacs, f);
	sprintf (line, "  %.16s", a->oname);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	loc.y += basl * 1.5;
	strplot (loc, base, 0.0, 0.0, line, pltimacs, f);

/* Hold and pull up between masks, turning autofocus off; this is similar
  to how it is done in ldss3 cutting.  */
    lasermode (Autofocus_hold, f);
//    lasermode (Laser_cutting, f);	// Restore regular mode
// Pull away by at least one inch (in hold mode) and turn autofocus off
    h += 30.0 * SF;
    fprintf (f, " G90 ");  Zout (f, h);  Feol (f);

/* Everything else is taken care of outside this program */

    return  n;	// count of features
}	// end cutgismo

#define  GISMULT  5

int	cutmul_gis (sortq** mmhead, opts opt)
/* Cut gismo masks from the mmhead queue, which should only contain
  gismo mask Obs pointers, into multiple mask .nc files.  Return
  the number of masks (subfields) successfully cut.  */
{
/* Copied from cutmul_ldss as the example... */
int	masks = 0;

int	i, j;
int	n;
int	sm;

sortq	*p;	// was.. , *q;
#ifdef  GISMONOTCH
sortq	*mhd;
int	flag;
#endif
Obs	*obs;
FILE	*out;
char	base[8];
char	extn[4];
char	filename[16];
char	titlebuf[32];

double	sdz;
double	vert;
double	h;
vect2	loc;
vect2	basv;
double	basl;


// Add ability to cut notch only file
#ifdef  GISMONOTCH
    if (mmhead == NULL && opt.gnch == 0) return 0;	// null case

    if (mmhead == NULL) mhd = NULL;
    else                mhd = *mmhead;
/* NOTE -- we are called with a sortq** which may be itself NULL, in
  which case we need opt.gnch, or it is a pointer to the sortq* to be used.
  THAT pointer may be NULL, in which case there is no list of masks and
  we need opt.gnch, or it may be real, in which case we need to set
  the local copy of opt.gnch to zero.  Above we set mhd as an actual
  pointer, which may be either NULL or valid.  Now, we test if it is
  valid (normal) or NULL (need gnch).  */
    if (mhd != NULL) {
	opt.gnch = 0; 
//	printf (" ** cutmul_gis set gnch 0 for non-null header.\n");  // debug only
    }

/* We do notch case only with mmhead null, the above sets the local copy
  of the option to zero if we have a real mmhead value.  In this case,
  we will need to call cutmul_gis with a null mmhead later if we have
  a opt.gnch actually present. */
    flag = opt.gnch;
    if (opt.dbug > 0) {
	if (opt.gnch) {
	    printf (" cutmul_gis for notch file.\n");
	} else {
	    printf (" cutmul_gis for %d mask files.\n", listcount(mhd) );
	    if (opt.subg != GISMULT) 
		printf (" *Cutting %d subfields in each mask.\n", opt.subg);
	}
    }
#else
// Return if nothing to be done
    if (mmhead == NULL) return 0;	// No argument

    if (opt.dbug > 0)
	printf (" cutmul_gis for %d mask files.\n", listcount (*mmhead) );

/* We appear to have a queue, we need to process it */
    if (opt.subg != GISMULT && opt.dbug > 0) 
	printf (" *Cutting %d subfields in each mask.\n", opt.subg);
#endif  // on GISMONOTCH

/* Outline of what is to be done:

Process the queue in blocks of 5 obs pointers.
For each block we will:
  0) Set the offset to 0 or -2 * index size	    sprintf (base, "%.6s", opt.base);  // First up to 6 char.s

  1) Obtain a new file name
  2) Write the initial commands into that file
	>> Copy set-up from other cut programs
	>> Position cut head in center of first mask
  3) Do the individual processes for each of up to 5 pointers
	>> Pop the top obs file from the queue (break if none)
	>> Do a loop over all mask features
	>> Write the label information in "proper" location
	>> kill the obs structure and the queue member
	>> Set offset for the next location
	>> Count another sub-mask processed
  4) Write the terminal commands into the file, close the file
	>> Cut the center hole
	>> Cut the notches in the sides
	>> Retract cutting head, turn off, move to output position
	>> Close file
	>> Count another file output.
If more obs files exist, return to step 1.

The whole thing is wrapped in a for loop on *mmhead not being null.
If it becomes null within the loop, break.


.. */

    SF = inchout ? 1.0/Inch : 1.0;	// Make sure it's set.
#ifdef  GISMONOTCH
    while (mhd != NULL || flag) {
	flag = 0;			// one time through
#else
    while (*mmhead != NULL) {
#endif  // on GISMONOTCH

// /* Set up offset to be 0,0 initially */
	setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */

/* Obtain file name and open the new file */

/* If a pointer to a valid <= 6 character base string has been provided
  in the options, use that.  Otherwise, generate a base string from
  the current date to be used here.  */
// Not using strlen here to check length...

	if (opt.base != NULL) {
//	    sprintf (base, "%.6s", opt.base);  // First up to 6 char.s
	    sprintf (base, "G%.6s", opt.base);  // First up to 6 char.s
	} else {
// GET THE BASE PART OF FILE NAME (SEE OUTLINE) GyyMddxx
#ifdef  GISMONOTCH
	    if (opt.gnch) sprintf (base, "Gntch");
	    else          sprintf (base, "G%s", curdate(0) );
#else
	    sprintf (base, "G%s", curdate(0) );
#endif  // on GISMONOTCH
	}
	sprintf (extn, "%s", ".nc");

//  e.g.   n = nexfile (base, extn);
// where base is LyyMdd and extn is ".nc";
// then: sprintf (filename, "%s%02d%s", base, n, extn);
	j = nexfile (base, extn);
	if (j < 1) {
	    printf (" ** No available file name \"%s--%s\"; no output.\n",
		base, extn);
	    return masks;
	}
	sprintf (filename, "%s%02d%s", base, j, extn);
// FIND NEXT SERIALIZED NAME WHICH DOES NOT YET EXIST
// OPEN THE NEW FILE FOR WRITING
	out = iopen (filename, "w+");

// IF OUTPUT FILE IS NOT OPENED CORRECTLY, WE NEED TO WRITE
// AN ERROR MESSAGE AND EXIT.
	if (out == NULL) {
	    printf (" ** Open error; no further output.\n");
	    return  masks;
	}

/*  Make up a title line, with file name and date for later. */
//	sprintf (titlebuf, "%s  %s", filename, curdate(1) );
	if (serno > 0) sprintf (titlebuf, "%s  %s  %d", filename, curdate(1), serno );	
	else           sprintf (titlebuf, "%s  %s", filename, curdate(1) );
	printf ("  Mask label:  %s\n", titlebuf);

	Lnum = 0;		// IDent counter
/* Initial set-up things copied from cutmask */
// set the Z offset
//	inst = M_lds3;	// Special case, this is ldss-3 instrument
	inst = M_gismo;	// Special case, this is GISMO
#ifdef  USE_CDAT
	sdz = CDhead[inst].dz * SF;
#endif
#ifdef  USE_GSTO
	sdz = *(double*)GS_lookup(&CDhead,inst,CD_dz,0) * SF;
#endif
#ifdef  USE_BSTO
	sdz = floatvalue (getval (&CDhead,MaskCase(inst),"dz")) * SF;
#endif
	Zoff = 0.0;
// NEEDS TO CHANGE FOR GISMO CASE
/* Above was commented out, as Zoff is set in cututil.h where it is
  defined.  Actually needed dynamically when cutting multiple blanks,
  so it is not redefined to another height below.  Also, set this for
  real in the other routines which set it so multiple masks are done
  correctly with the height computed each time...    */

//	vert = maskz (make2vect ( SF*(flange+1.0), 0.0) );
// NEEDS TO CHANGE FOR GISMO CASE
	vert = 0.0;	// for gismo cutting jig location.
// Offset is same as IMACS case
	Zoff = sdz - vert;
	if (opt.dbug > 0)  printf (" --Zoff = %7.3f; Z0 = %5.3f, Vh = %6.3f\n", 
				Zoff, sdz, vert);

/* Write initial commands into the file */
	if (opt.dbug >= 0) printf ("  Cutting GISMO mask %s \n", filename);
/* Some debug stuff */
#if (DEBUG_CUT > 0)
	if (opt.dbug > 0) {
#ifdef  SIMPLECURVE
	    printf ("  Inst = %d, curve = %7.5f\n", inst, Curve );
// NEEDS TO CHANGE FOR GISMO CASE
#else
vect2	cv;
//		*GS_lookup(&CDhead,inst,CD_curve) );	// this doesn't work!
#ifdef  USE_CDAT
		cv = CDhead[inst].curve;
#endif
#ifdef  USE_GSTO
		cv = *(vect2*)GS_lookup(&CDhead, inst, CD_curve, 0);
#endif
#ifdef  USE_BSTO
		cv = vect2value(getval(&CDhead,MaskCase(inst),"Curve"));
#endif
	    printf ("  Inst = %d, curve = %7.5f (x), %7.5f (y)\n", inst, 
		cv.x, cv.y);
#endif  // on SIMPLECURVE
	}
#endif  // on DEBUG_CUT

// SET UP THE CUTTING PARAMETERS AS GLOBALS

// INCLUDE VERSION NUMBER OF THIS PROGRAM
//	fprintf (out, "( GISMO mask file by %s version %s%s )\r\n",
//		PROGNAME, VERSION, CTUVERS);
	fprintf (out, "( GISMO mask file by %s version %s )\r\n",
		PROGNAME, XVERSION);
// Could split above into two lines if needed for compatability
	fprintf (out, "( Instrument: %s )\r\n", MaskName(inst) );
// CHANGE TO MaskName IN ALL OTHER USES TOO

/* Set up global cutting parameters */
	cutparams (inst);

//	if (opt.dbug > 0) printf (" GISMO multi -- Acon = %.3g\n", Acon);
	if (opt.dbug > 0) printf (" GISMO cutpar (%s) -- Acon = %.3g\n", 
			MaskCase(inst), Acon);

/* If we have a request for slow cutting, modify the slow and vslow
  rates to slower values for fine cutting... */
	if (opt.slow) {
double	ratio;
	    ratio = vslow/slow;
	    slow = vslow;
	    vslow = ratio*slow;
	}


/* Initial cutting commands, from cutimacs */
	lasermode (Autofocus_off, out);
	lasermode (Exact_stop_on, out);
	IDent (out);
	if (inchout == 0) {		// Warning section
	    fprintf (out, " ( ! DANGER!! -- Millimeter mode file! )\r\n");
	    fprintf (out, " M0 ( DANGER!! Millimeter mode! )\r\n");
	    fprintf (out, " ( Millimeters can cause a crash! )\r\n");
	}
	if (inchout) fprintf (out, " G20 ( Inch measurement )\r\n");
	else         fprintf (out, " G21 ( Metric measurement )\r\n");
/* Further commands from imacs cutting are in cutldss3 above */

/* Process up to 5 individual sub-masks into the output file */
#ifdef  GISMONOTCH
	for (i=j=sm=0; i<opt.subg && opt.gnch==0; i++) { // loop over up to 5 sub-masks
#else
	for (i=j=sm=0; i<opt.subg; i++) {	// loop over up to 5 sub-masks
#endif  // on GISMONOTCH
// NEEDS TO CHANGE FOR GISMO CASE

	    setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */
	    opt.subm = i;		// sub mask designator

#ifdef  GISMONOTCH
	    p = mhd;
//	    if (p == NULL && opt.gnch == 0) break;
	    if (p == NULL) break;
	    mhd = pop_sq (mhd);
	    if (mmhead != NULL) *mmhead = mhd;  /// was.. pop_sq (*mmhead);
	    if (p == NULL) obs = NULL;
	    else           obs = (Obs*) p->q;
#else
	    p = *mmhead;
	    if (p == NULL) break;
	    *mmhead = pop_sq (*mmhead);
	    obs = (Obs*) p->q;
#endif  // on GISMONOTCH
/* Process the sub-mask, and count our success */
//	    j += cutldss3 (obs, out, opt);
// CHANGE THAT TO COUNT TOTAL FEATURES DONE
// SUM BOTH TOTAL FEATURES, AND NUMBER OF SUB-MASKS WITH >0 FEATURES
// REPORT NUMBER OF FEATURES ON EACH SUB-MASK

/* SHOULD SET SF HERE DUE TO THE CTE --- */
#ifdef  USE_CTE
//	    if (opt.xcte) {
	    if (opt.xcte && obs != NULL) {
double	tex;
		SF = inchout ? 1.0/Inch : 1.0;
		tex = CTE * (Tcut - obs->temp);
		if (opt.dbug > 0 && tex != 0.0)
		    printf (" Correcting thermal expansion of %9.3e\n", tex);
		SF *= 1.0 + tex;
	    }
#endif

/* We need to write a new version of cutldss3 for gismo, to use the
  indexing of axial direction, the alternate location of the label,
  and possibly other things...  */

// NEEDS TO CHANGE FOR GISMO CASE
	    n = cutgismo (obs, out, opt);
	    if (n < 1) {	// No mask cut, decrement i to try again
		i--;
	    } else {	// cut a mask, total the mask and feature count
		j += n;		// features count
		sm++;		// sub-mask count
	    }

// IF ZERO FEATURES DONE (INCLUDING LABEL) DECREMENT i TO EXTEND LOOP

/* Done with sub-mask, kill the obs structure and the sortq entry p */
	    obs = kill_obs (obs);
	    free (p);

// /* Set the offsets to the next location (based on i) */
// // SET OFFSETS
// No, we set these above based on i...

	}	// Loop over input files indexed by i, 0-6

// NOTE  NOTE  NOTE
// While still in engrave mode, put label on top of mask!!
// already in titlebuf

#ifdef  GISMONOTCH
	if (opt.gnch) {	// Need the initialization stuff here from cutgismo
vect2	c = {0.0, 0.0};
//double	h;
	    lasermode (Laser_cutting, out);
	    lasermode (Laser_off, out);
	    Feedrate (out, " G01", quick);
	    Feol (out);

//    tool = bwid / 2.0;
/* MAY LATER WANT TO SPECIFY AN ACTUAL TOOL SIZE... */

//    last = c;
//    h = zlast = maskz (last);
	    h = maskz (c);
/* Work offset location, using 59 for IMACS here */
	    fprintf (out, " G59  ( Work offset for Multi-mask )\r\n");
// NEEDS TO CHANGE FOR GISMO CASE
	    IDent (out);
	    fprintf (out, " G90");

	    VXYout (out, c, 1);
	    lastloc = c;	// needed?  probably not, doesn't hurt
	    lasermode (Autofocus_hold, out);
	    Zout (out, h);
	    fprintf (out, " ( Start location )\r\n");
	    IDent (out);

// NOTE - this gas/shutter stuff could be a laser_mode call!!
	    fprintf (out, " M10");	// select gas, 
	    fprintf (out, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
	    fprintf (out, " M14 ( Shutter open )\r\n");

	    lasermode (Autofocus_on, out);

/// copied initialize
	} else {	// continue with label if not notch
//	if (opt.gnch == 0) {	// only allow label if not notch
#endif  // on GISMONOTCH
	fprintf (out, "( Label: %s )\r\n", titlebuf);	/* add comment */
	IDent (out);	// Debug line number
	basl = 3.0 * SF;	// 3 mm letters
	basv = make2vect (basl, 0.0);
// ADD OFFSET OPTIONS
	loc = mul2vect ( make2vect ( -0.5 * 350.0 + 44.0, 
#if GISMOFFSET || GISMONOTCH
			0.5 * 326.5 - 18.25 + opt.offset ), SF);
#else
			0.5 * 326.5 - 18.25  ), SF);
#endif
// An alternate location would be ( 0.5 * 350.0 + 8.0, -1.5 )
// Move to loc, set engraving mode, 
	lastz = maskz (loc);		// The standard global
	VXYout (out, loc, 1);
	lastloc = loc;
	Zout (out, lastz);
	Feol (out);
// or could do:  lastz = newloc (lastloc, lastz, loc, out);
//	lastz = newloc (lastloc, lastz, loc, out);
	lasermode (Autofocus_on, out);
	lasermode (Laser_engraving, out);
// Should be ready to engrave now...
	strplot (loc, basv, 0.0, 0.0, titlebuf, pltimacs, out);

#ifdef  GISMONOTCH
	}	// only allow label if not notch
	if (opt.gnch && sm==0) {
	    sm = opt.gnch;
	    if (j == 0) j = sm * 3;  // the notches count here
	}
#endif

	lasermode (Laser_cutting, out);
//	Feol (out);		// Causes a blank line.
	setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */

/* Count masks, files output */
//	masks += j;
	masks += sm;

/* Engrave the filename and the date at the top of the
  mask, as far up as we can get it.  Put it a line above
  the top mask label.  Use the filename directly, and
  translate date to yyyy-mmm-dd with alphabetic month.
  We can make up the output buffer for the label up
  above where the file name is opened, but we do the
  actual engraving step now.   */

// Engrave "titlebuf" contents at extreme upper left

/* At this point, we do these:
    1) Cut the central rectangular hole
    2) Cut single small notch on right side
    3) Cut two notches assymetrically placed on left side
    */

/* Rectangular center hole... */
	{	// new section for some local variables
vect2	pr;	// point being used
vect2	ps;	// displacement to start
vect2	pc;	// local center
double	rr = 35.0;	// rectangle fillet
double	rn = 2.5;	// notch fillet
double	xe;	// x edge
double	ye;	// y edge
double	xs;	// Step over edge
#ifdef  GISMONEW
double	ys;	// Step over edge
#ifdef  GISMONW2
double	rb = 3.0;	// Bottom notch radius
#else
double	rb = 0.5;	// Bottom notch fillet radius
#endif
double	wb = 6.0;	// Width bottom notch
#ifdef  GISMONW2
double	db = 4.8;	// Depth bottom notch
#else
double	db = 5.8;	// Depth bottom notch
#endif
#endif
double	xd;	// notch depth modified
double	yf = 0.0;	// y offset, nominally null

// Set all notch parameters before the hole code which is skipped!
	rr *= SF;
	rn *= SF;
#ifdef	GISMONEW
	rb *= SF;
	wb *= SF;
	db *= SF;
#endif
#if GISMOFFSET || GISMONOTCH
	yf += opt.offset * SF;	// Set the offset
#endif

	fprintf (out, "( Central Hole )\r\n");	/* add comment */


#ifdef  GISMONOTCH
//	if (opt.gnch == 0) {	// only allow hole if not notch
// cut a small hole at center of rectangle
    if (opt.gnch) {
	pr = make2vect (0.0, 0.0);	// on center
	h = newloc (lastloc, h, pr, out);
//	Feedrate (out, " G90", slow);
//	h = shole  (make2vect(0.0, 0.0), 020*SF - bwid, h, slow, out);

	h = chole (pr, 0.20*SF - bwid, h, slow, 1, out);
	lasermode (Autofocus_off, out);	// as done in shole

    }
#endif
	xe = 0.5 * 102.5 * SF;
	ye = 0.5 * 144.0 * SF;

// ADD OFFSET OPTIONS
	pr = make2vect (xe, ye - rr + yf);
	ps = make2vect (0.15 * SF, 0.0);
	pr = sub2vect (pr, ps);

	h = newloc (lastloc, h, pr, out);
	Feedrate (out, " G90", slow);	// or fr ?
	lasermode (Laser_on, out);
	h = cutabs (&pr, ps, h, out);
	h = cutarc (&pr, make2vect(-rr, 0.0), 0.25, out);
	h = carve_edge (&pr, make2vect(rr-xe, ye+yf), h, out);
	h = cutarc (&pr, make2vect(0.0, -rr), 0.25, out);
	h = carve_edge (&pr, make2vect(-xe, rr-ye+yf), h, out);
	h = cutarc (&pr, make2vect( rr, 0.0), 0.25, out);
	h = carve_edge (&pr, make2vect(xe-rr, -ye+yf), h, out);
	h = cutarc (&pr, make2vect(0.0,  rr), 0.25, out);
	h = carve_edge (&pr, make2vect(xe, ye-rr+yf), h, out);
	h = cutabs (&pr, neg2vect (ps), h, out);
	lasermode (Laser_off, out);
	lasermode (Autofocus_off, out);
/* We need to turn autofocus off when we cross already cut features, such
  as the center rectangle hole.  Although it is still supposed to be
  attached, there may be some stress on the tabs and we need not push
  things here.  Turn autofocus off in the next slew moves.  */
//#ifdef  GISMONOTCH
//	}	// only allow hole if not notch
//#endif

/* Cut small centered notch on right side */
	xe = 0.5 * 350 * SF;
	xs = 0.5 * SF;
#ifdef  GISMONW2
	xd = xs + 1.00*SF;
#else
	xd = xs +6.15*SF - rn;  // should be:  xs + 3.65*SF
#endif
	pr = make2vect (xe+xs, rn+yf);	// Outside of top edge (start location)
	fprintf (out, "( Centered notch on side. )\r\n");	/* add comment */
// ADD OFFSET OPTIONS
// Should add beam width here
	h = newloc (lastloc, h, pr, out);
	lasermode (Autofocus_restore, out);
	Feedrate (out, " G90", slow);
	lasermode (Laser_on, out);
	ps = make2vect (xd, 0.0);		// Straight line to right
	pc = make2vect (0.0, -rn);		// Toward center
	h = cutabs (&pr, neg2vect(ps), h, out);	// Cut inward
	h = cutarc (&pr, pc, 0.5, out);		// ccw arc
	h = cutabs (&pr, ps, h, out);		// Cut outward
	lasermode (Laser_off, out);
	lasermode (Autofocus_off, out);

#ifdef  GISMONEW
/* Cut one bottom notch, centered */
	xe = 0.0;
	ye = 0.5 * 326.5 * SF;
	ys = 0.5 * SF;
#ifdef  GISMONW2
	pr = make2vect (xe+rb, yf-(ye+ys) );
#else
	pr = make2vect (xe+0.5*wb, yf-(ye+ys) );
#endif
	fprintf (out, "( Centered notch on bottom. )\r\n");	/* add comment */
	h = newloc (lastloc, h, pr, out);
	lasermode (Autofocus_restore, out);
	Feedrate (out, " G90", slow);
	lasermode (Laser_on, out);
	ps = make2vect (0.0, ys+db-rb);		// Straight up to curve
//	pc = make2vect (-rb, 0.0);		// Toward center
	h = cutabs (&pr, ps, h, out);		// Cut to curve
#ifdef  GISMONW2
	h = cutarc (&pr, make2vect(-rb, 0.0), 0.5, out);	// ccw arc
#else
	h = cutarc (&pr, make2vect(-rb, 0.0), 0.25, out);	// ccw arc
	h = cutabs (&pr, make2vect(-(wb-2.0*rb), 0.0), h, out);	// across top
	h = cutarc (&pr, make2vect(0.0, -rb), 0.25, out);	// ccw arc
#endif
	h = cutabs (&pr, neg2vect(ps), h, out);	// back outside
	lasermode (Laser_off, out);
	lasermode (Autofocus_off, out);

/* Add hole on left side */


//    c = mul2vect (make2vect(69.30, 58.98), SF);
//    w = ( opt.subm ? 2.51 : 2.46 ) * SF;
//    zlast = chole (c, w-bwid, zlast, slow, 1, f);

	fprintf (out, "( Centered hole near side. )\r\n");	/* add comment */
	xe = -0.5 * 350.0 * SF;
	pr = make2vect (xe + 3.0*SF, 0.0);
	h = chole (pr, 4.0*SF, h, slow, 1, out);


#else  // on GISMONEW
/* Cut two notches on left side, bottom first */
	xe *= -1.0;
// ADD OFFSET OPTIONS
	fprintf (out, "( Two notches on side. )\r\n");	/* add comment */
	pr = make2vect (xe-xs, -56.25*SF + yf);
	h = newloc (lastloc, h, pr, out);
	lasermode (Autofocus_restore, out);
	Feedrate (out, " G90", slow);
	lasermode (Laser_on, out);
	pc = make2vect (0.0, rn);		// Toward center
	h = cutabs (&pr, ps, h, out);
	h = cutarc (&pr, pc, 0.25, out);
	h = cutabs (&pr, make2vect(0.0,10.0*SF-2.0*rn), h, out);
	h = cutarc (&pr, make2vect(-rn, 0.0), 0.25, out);
	h = cutabs (&pr, neg2vect(ps), h, out);
	lasermode (Laser_off, out);
	lasermode (Autofocus_off, out);

/* Move around the holding things, with autofocus off... */
	pr = make2vect (xe + 6.0*SF, -40.25*SF + yf);
	h = newloc (lastloc, h, pr, out);
	pr.y = 83.75*SF + yf;
	h = newloc (lastloc, h, pr, out);

/* Top notch */
// ADD OFFSET OPTIONS
	pr = make2vect (xe-xs, 89.75*SF + yf);
	h = newloc (lastloc, h, pr, out);
	lasermode (Autofocus_restore, out);
	Feedrate (out, " G90", slow);
	lasermode (Laser_on, out);
	h = cutabs (&pr, ps, h, out);
	h = cutarc (&pr, pc, 0.25, out);
	h = cutabs (&pr, make2vect(0.0,10.0*SF-2.0*rn), h, out);
	h = cutarc (&pr, make2vect(-rn, 0.0), 0.25, out);
	h = cutabs (&pr, neg2vect(ps), h, out);
	lasermode (Laser_off, out);
#endif	// on GISMONEW

	}	// end variable section (cutting hole and notches)


/* If any submasks written, fill rest of file.  Close file */
#ifdef  GISMONOTCH
	if (j <= 0) j = opt.gnch;
//	if (opt.gnch || j > 0) {	// Some output exists
#endif  // on GISMONOTCH
//#else
	if (j > 0) {		// Some entries
//#endif  // on GISMONOTCH
	    h = 3.0 * Inch * SF;
	    end_mask (out, h);

#ifdef  USE_ACON
	    fill_cuttime (out);
#endif
// CLOSE FILE

/* Add checksum and signature */
	    signfile (out, '(', dodebug-1);

	    iclose (out);
// NEEDS TO CHANGE FOR GISMO CASE
	    printf (" -Processed %d features in %d sub-masks.\n", j, sm);

// Increment serial number if used
	    if (serno > 0) serno++;

// COUNT FILES -- already done?
	} else {		// No entries
// CLOSE AND DELETE FILE, IT IS NULL
// rm remove rename unlink
// unlink requires unistd include; already needed for access.
	    iclose (out);
	    unlink (filename);	// Deletes the file
	}	// Null mask output test

    }	// Null ldss mask queue test

// Return number of masks processed
    setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */
// NEEDS TO CHANGE FOR GISMO CASE
    return  masks;	// ALSO NEED TO RETURN FEATURES?

}	// end cutmul_gis (gismo cutting)


	/* - - - - - - - - - - - - - - *\
	|				|
	|	Gismo Decker Mask	|
	|				|
	\* - - - - - - - - - - - - - - */

#ifdef  GISMODECKER
// Place the following in a separate file, and
// include these programs from there

double	shole (vect2 c, double d, double h, double fr, FILE* f)
/* Scale the hole parameters before calling chole */
{
vect2	s;
double	z;

    s = mul2vect (c, SF);
    z = chole (s, d, h, fr, 1, f);	// turns on autofocus if needed
    lasermode (Autofocus_off, f);	// avoid fall-through
    return  z;
}

static	double	rxrect (double x1, double x2, double y1, double y2, 
		double r, FILE* f)
// Subroutine to cut rectangular hole with radial extension
// X and Y values are at centers of fillets
// Only r is affected by beam radius compensation
// x1,x2,y1,y2 need to be multiplied by SF
// r has SF in it already
// return height at which we are left.
{
double	xl, xr;		// x centers
double	yt, yb;		// y centers
double	el, er;		// x edges
double	et, eb;		// y edges
double	h;		// start height
double	z;		// local height
vect2	c;
vect2	start;
vect2	dp={0.0, 0.15};
vect2	dv;

// Find left and right x edges (extended by r)
    if (x1 < x2)  { xl = x1*SF;  xr = x2*SF; }
    else	  { xl = x2*SF;  xr = x1*SF; }
    el = xl - r;  er = xr + r;

// Find top and bottom y edges similarly
    if (y1 < y2)  { yb = y1*SF;  yt = y2*SF; }
    else	  { yb = y2*SF;  yt = y1*SF; }
    eb = yb - r;  et = yt + r;

    dv = mul2vect (dp, SF);

    fprintf (f, "( Start rectangle cx. )\r\n");  // debug only

// Start slightly inside at upper left corner
    c = make2vect (xl, et);
    start = sub2vect (c, dv);
//    start = mul2vect ( sub2vect ( make2vect (xl, et), dp ), SF);
    h = maskz (start);	// just a guess...
//    z = newloc (start, h, c, f);	// Start location
    z = newloc (lastloc, h, start, f);	// Start location
    lasermode (Autofocus_restore, f);
    Feedrate (f, " G90", slow);
    lasermode (Laser_on, f);
//    c = sum2vect ( start, mul2vect (dp, SF) );
    z = newloc (lastloc, z, c, f);	// Go out to edge
    z = cutarc (&c, make2vect(0.0, -r), 0.25, f);	// Top left corner
    c = make2vect(el,yb);
    z = newloc (lastloc, z, c, f);	// left edge
    z = cutarc (&c, make2vect(r, 0.0), 0.25, f);	// lower left
    c = make2vect(xr,eb);
    z = newloc (lastloc, z, c, f);
    z = cutarc (&c, make2vect(0.0, r), 0.25, f);	// lower right
    c = make2vect(er,yt);
    z = newloc (lastloc, z, c, f);
    z = cutarc (&c, make2vect(-r, 0.0), 0.25, f);	// upper right
    c = make2vect(xl,et);
    z = newloc (lastloc, z, c, f);	// Completes the cut
    z = newloc (lastloc, z, start, f); 

// Carve the corners with circular interpolation, and edges with lines

// At end, move slightly back inside and turn off laser
    lasermode (Laser_off, f);
    lasermode (Autofocus_off, f);	// leave focus off for large rect.
    return  z;
}

static	int	gismodecker (opts opt)
/* Produce a special file for the GISMO decker mask, using GISMO
  cutting fixture and curvature, with dimensions programmed here.  */
{
int	rv = 0;		// return value
FILE	*f;
vect2	c;		// location
double	r;		// radius
double	d;
double	z;
double	w;

double	ga;		// gap angle
double	asg;		// segment angle
double	ai;		// angle increment
double	ba;		// starting angle
double	swa;		// sebment base angle
int	nsg;		// number of segments
int	ns;		// segment index
int	js;		// stroke index

double	sdz;
double	vert=0.0;	// To avoid uninitialized warning (KDC 2011/1-21)
double	tool=0.0;
vect2	last;
double	h;
double	zlast;
double	x, y;
char	line[80];
vect2	basl;
vect2	locn;

// Obtain an output file, return if not possible
    f = iopen ("gisdek.nc", "w+");
    if (f == NULL) return rv;

// Include here standard set-up things, or call a setup program
    SF = inchout ? 1.0/Inch: 1.0;
    inst = M_gismo;		// explicitly set the inst global
    cutparams (M_gismo);

	if (opt.dbug > 0) printf (" GismoDecker -- Acon = %.3g\n", Acon);

// Also need to get the curvature set!

#ifdef  USE_CDAT
    sdz = CDhead[inst].dz * SF;  // Scaled z offset
#endif
#ifdef  USE_GSTO
    sdz = *(double*)GS_lookup (&CDhead,inst, CD_dz,0) * SF; // Scaled z offset
#endif
#ifdef  USE_BSTO
    sdz = floatvalue (getval (&CDhead,MaskCase(inst),"dz")) * SF;
#endif

    Zoff = 0.0;

    Zoff = sdz - vert;	// vert may be uninitialized ??

    fprintf (f, "( Gismo Decker mask. )\r\n");
    fprintf (f, "( By %s version %s )\r\n", PROGNAME, XVERSION);
    fprintf (f, "( Instrument: GISMO )\r\n");

/* If we have a request for slow cutting, modify the slow and vslow
  rates to slower values for fine cutting... */
    if (opt.slow) {
double  ratio;
	ratio = vslow/slow;
	slow = vslow;
	vslow = ratio*slow;
    }


// for the type of mask blank used.
// Also set instrument type, starting location, z offset
// get globals, including bwid, slow; all scaled.

// From individual mask cutting example imacs

/* Initializing machine */

    lasermode (Autofocus_off, f);

//    Astop = (a == NULL) ? 0 : 1;
    Astop = 1;
    if (Astop) lasermode (Exact_stop_on, f);

    IDent (f);
    if (inchout == 0) {         // Warning section
	fprintf (f, " ( ! DANGER!! -- Millimeter mode file! )\r\n");
	fprintf (f, " M0 ( DANGER!! Millimeter mode! )\r\n");
	fprintf (f, " ( Millimeters can cause a crash! )\r\n");
    }
    if (inchout) fprintf (f, " G20 ( Inch measurement )\r\n");
    else         fprintf (f, " G21 ( Metric measurement )\r\n");


    lasermode (Laser_cutting, f);
    lasermode (Laser_off, f);
    Feedrate (f, " G01", quick);
    Feol (f);

    tool = bwid / 2.0;
/* MAY LATER WANT TO SPECIFY AN ACTUAL TOOL SIZE... */

/* Work offset location, using 59 for IMACS here */
    fprintf (f, " G59  ( Work offset for IMACS )\r\n");
    IDent (f);
    fprintf (f, " G90");

// actual starting of cutting onto mask
    c = mul2vect (make2vect (-43.80, 112.85), SF);
    last = c;
    h = zlast = maskz (last);

    VXYout (f, c, 1);
    lasermode (Autofocus_hold, f);
    Zout (f, h);
    fprintf (f, " ( Start location )\r\n");
    IDent (f);
// NOTE - this gas/shutter stuff could be a laser_mode call!!
    fprintf (f, " M10");        // select gas,
    fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n"); // turn gas on
    fprintf (f, " M14 ( Shutter open )\r\n");

    lasermode (Autofocus_on, f);
    lastloc = last;

// Some setup values
    r = 2.00 * SF - bwid/2.0;	// Rectangle fillet radius
    d = 3.40 * SF - bwid;	// Hole diameters
    w = 4.00 * SF - bwid;	// Locating holes

// Cut rounded rectangles at all the proper places, using subroutine
// Quadrant II
    IDent (f);
    fprintf (f, "( Quadrant II )\r\n");
    z = rxrect ( -06.50,  -43.80,  103.50, 111.00,  r, f);
    z = rxrect ( -50.50,  -86.50,   82.00,  88.00,  r, f);
    z = shole  (make2vect(-107.56,  90.00), d, z, slow, f);
    z = rxrect ( -96.00, -127.70,  104.50, 110.60,  r, f);
    z = rxrect (-137.00, -158.00,   62.90,  69.20,  r, f);
    z = shole  (make2vect(-107.56,  30.00), d, z, slow, f);

// Locating hole
    z = shole  (make2vect(-107.56,  0.00), w, z, slow, f);

// Cut any clearance holes, use autofocus and standard hole cutting
// Quadrant III
    IDent (f);
    fprintf (f, "( Quadrant III )\r\n");
    z = shole  (make2vect(-107.56, -30.00), d, z, slow, f);
    z = rxrect (-137.00, -158.00,  -62.90, -69.20,  r, f);
    z = rxrect ( -96.00, -127.70, -104.50,-110.60,  r, f);
    z = shole  (make2vect(-107.56, -90.00), d, z, slow, f);
    z = rxrect ( -50.50,  -86.50,  -82.00, -88.00,  r, f);
    z = rxrect ( -06.50,  -43.80, -103.50,-111.00,  r, f);

// Quadrant IV
    IDent (f);
    fprintf (f, "( Quadrant IV )\r\n");
    z = rxrect (  06.50,   43.80, -103.50,-111.00,  r, f);
    z = rxrect (  50.50,   86.50,  -82.00, -88.00,  r, f);
    z = rxrect (  96.00,  127.70, -104.50,-110.60,  r, f);
    z = shole  (make2vect( 107.56, -90.00), d, z, slow, f);
    z = rxrect ( 137.00,  158.00,  -62.90, -69.20,  r, f);
    z = shole  (make2vect( 107.56, -30.00), d, z, slow, f);

// Locating hole
    z = shole  (make2vect( 107.56,  0.00), w, z, slow, f);

// Quadrant  I
    IDent (f);
    fprintf (f, "( Quadrant I )\r\n");
    z = shole  (make2vect( 107.56,  30.00), d, z, slow, f);
    z = rxrect ( 137.00,  158.00,   62.90,  69.20,  r, f);
    z = shole  (make2vect( 107.56,  90.00), d, z, slow, f);
    z = rxrect (  96.00,  127.70,  104.50, 110.60,  r, f);
    z = rxrect (  50.50,   86.50,   82.00,  88.00,  r, f);
    z = rxrect (  06.50,   43.80,  103.50, 111.00,  r, f);

// Etch a label near the center, name and date
    sprintf (line, "Version %s : %s", XVERSION, curdate(1));
// lettering size is 4 mm
    basl = make2vect (4.0*SF, 0.0);
// location is centered on 0, and y of 65 mm
    locn = make2vect (-0.5 * basl.x * strwid(line), 65.0 * SF);
    fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
    IDent (f);	// Debug line number
// Move to loc, set engraving mode, 
    lastz = newloc (lastloc, lastz, locn, f);
    lasermode (Autofocus_restore, f);
    lasermode (Laser_engraving, f);
    strplot (locn, basl, 0.0, 0.0, line, pltimacs, f);
    lasermode (Autofocus_off, f);

// Cut central hole leaving 4 small knockouts
// See ldss mask cutting for the needed howto
    IDent (f);
    fprintf (f, "( Central Hole )\r\n");
    lasermode (Laser_cutting, f);
    r = 0.5 * 101.5 * SF - tool;
//    ga = 0.12 * SF / r;		// gap angle
    ga = 0.09 * SF / r;		// gap angle
    asg = M_PI_2;		// segment angle
    nsg = nsides (r, 0.3*bwid, asg - ga);	// Strokes
    ai = (asg - ga) / nsg;	// angle increment
    ba = M_PI_2 + ga / 2.0;	// starting angle

// Slew to start point
    c = circloc (r, ba);
    z = newloc (lastloc, zlast, c, f);
//    if (Autofocus & 2 == 0) lasermode (Autofocus_on, f);
//    if (Autofocus & 1) lasermode (Autofocus_release, f);
    lasermode (Autofocus_restore, f);

// Start the cutting
    lasermode (Laser_cutting, f);
    Feedrate (f, "", slow);
    Feol (f);

// Loop to cut the segments
    for (ns=0; ns < 4; ns++) {
	SetSlew = 0;			// do not slew to end of gap
	swa = ba + asg * ns;
	c = circloc (r, swa);
	z = newloc (lastloc, z, c, f);
// turn on and cut a curve
	lasermode (Laser_on, f);
	for (js=1; js <= nsg; js++) {	// loop over strokes
	    c = circloc (r, swa + js*ai);
	    z = newloc (lastloc, z, c, f);
	}	// loop cutting a segment by strokes
	lasermode (Laser_off, f);
    }	// loop cutting segments
    lasermode (Autofocus_off, f);

// Cut any edge locators

// Cut the perimeter to separate from the mask blank
// use carve edge to cut leaving gaps.

/* To cut corner chamfers, use "newloc" to cut diagonals rather than
  cutabs which would use a vector rather than a specific location.
  The new position is corner location less the notch value in x or y,
  but not both.  That is also the position of the start and end.  */

    x = 0.5 * 335.0 * SF + tool;
    y = 0.5 * 250.0 * SF + tool;
    d = 0.15 * SF;		// slight overshoot on start/end
    w = 5.0 * SF;		// chamfer displacement on coordinate
    c = make2vect (x-w+d, y);
    z = newloc (lastloc, z, c, f);
    lasermode (Autofocus_restore, f);
    lasermode (Laser_on, f);
    Feedrate (f, " G01", slow);
    Feol (f);
    last = c;
// Cut edge
    c = make2vect (-(x-w), y);
    z = carve_edge (&last, c, z, f);
// Cut chamfer
    c = make2vect (-x, y-w);
    z = newloc (last, z, c, f);
    last = c;
// Cut edge again
    c = make2vect (-x, -(y-w));
    z = carve_edge (&last, c, z, f);
// Cut chamfer
    c = make2vect (-(x-w), -y);
    z = newloc (last, z, c, f);
    last = c;
// Cut third edge
    c = make2vect (x-w, -y);
    z = carve_edge (&last, c, z, f);
// Cut chamfer
    c = make2vect (x, -(y-w));
    z = newloc (last, z, c, f);
    last = c;
// Cut final edge
    c = make2vect (x, y-w);  // was y+d);
    z = carve_edge (&last, c, z, f);
// Cut final chamfer
    c = make2vect (x-w, y);
    z = newloc (last, z, c, f);
    lasermode (Laser_off, f);

// Call to end mask, turn things off and position at finish location
    z = 3.0 * Inch * SF;
    end_mask (f, z);

#ifdef  USE_ACON
    fill_cuttime (f);
#endif

// If all is good, set the return value
    rv = 1;

/* Add checksum and signature */
    signfile (f, '(', dodebug-1);

// Close the output file
    iclose (f);
// no serial number on gismo decker mask

    return  rv;
}	// end gismodecker
#endif  // on GISMODECKER



/* ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- */

	/* - - - - - - - - - - - - - - *\
	|				|
	|   WFCCD  multi-mask cutting	|
	|				|
	\* - - - - - - - - - - - - - - */

/* Parameters defining the mask dimensions */

#define  HoleSize  (0.150 * Inch)
#define  XSize   (5.980 * Inch)
#define  YSize   (4.480 * Inch)
#define  Corner  (0.40  * Inch)
#define  VholeX  (2.875 * Inch)
#define  VholeY  (1.20  * Inch)
#define  LholeX  (1.20  * Inch)
#define  LholeY  (2.125 * Inch)
#define  Xnotch  (2.00  * Inch)
#define  Ynotch  (0.00  * Inch)
#define  OffSet  (0.250 * Inch)

/* Symbol Inch is defined in maskdat.h as 25.4
There are 4 Vhole and 4 Lhole holes, all at +/-X and +/-Y from
the nominal center.
There are two notches at -Ysize and +/-Xnotch, and one
notch at Ynotch and +Xsize.
All 4 corners are cut in by Corner in both x and y.  */

/* The 4 sub-masks will be cut with centers at X = +/- XSize/2 + OffSet
and Y = +/- YSize/2 + OffSet
The Y axis is (as in GISMO) along the cylinder axis, and the X axis
is on the curve; curvature in Y is zero.   */


/* Moved "nsides" from here to cututil.c */

static	int	cutwfccd (Obs* a, FILE* f, opts opt)
{
// cut a single mask with global offsets
// called only by cutmulti, which follows here...

int	n;
double	tool;
vect2	c;	// location of center
static	vect2	last = {999.0, 999.0};	// Absolute last loc for slew
static	double	zlast;	// For incremental z movement...
double	w;	// may not be needed??
objq	*aq;	// Start object
char	line[80];
vect2	loc, base;	// Label location, direction/size
double	basl;		// Label base size
double	xs;		// Label size
double	h;		// Z-altitude
double	fr;

// Return 0 if null inputs.
    if (a == NULL) return 0;
    if (f == NULL) return 0;

// The global offsets are already set here.

/*  Outline of what is needed:
  Do traveling salesman on the object list
  Loop over all objects; cut those objects
  Engrave the proper legend around edge
  Cut the border and locating notch
  Return 1 if done OK
.. */

/* Check for null object count... */
    n = listcount (a->ob);
    if (n < 1) {
	printf (" ** No mask objects in list, nothing written.\n");
	return  0;
    }

// COPY THE ABOVE THINGS FROM cutimacs PROGRAM ABOVE.
// BASIC INITIALIZATION HAS ALREADY BEEN DONE
// FINAL EXIT THINGS WILL BE DONE AFTER RETURN.
// ONLY THE CUTTING OF MASK FEATURES IS DONE HERE, FOLLOWED
// BY AN ELEMENTARY LABEL.
// RESTORE CUTTING FROM ENGRAVING MODE AFTER DOING THE LABEL.

    aq = qstart (a->ob, M_wfcd, opt);	// reorder, find start
// We start at point aq and end close to the label...
    c = mul2vect (aq->smpos, SF);
//    c.x *= -1.0;	// Remember the x positions are reflected...
// NOTE -- ACTUALLY X IS NOT REFLECTED FOR WFCCD

/* Inform user of mask being cut.. */
//    printf (" Submask %d is \"%.8s\" for \"%.16s\".\n",
//	opt.subm,  a->fname,  a->oname );
    if (serno > 0) printf (" Mask %d.%d is \"%.8s\" for \"%.16s\".\n",
	serno, opt.subm,  a->fname,  a->oname );
    else           printf (" Submask %d is \"%.8s\" for \"%.16s\".\n",
	opt.subm,  a->fname,  a->oname );

/* Identify the sub-mask in the output file */
//    fprintf (f, "( \"%.8s\" for \"%.16s\". )\r\n", a->fname, a->oname);
    if (serno > 0) fprintf (f, "( Mask %d.%d \"%.8s\" for \"%.16s\". )\r\n", 
	serno, opt.subm, a->fname, a->oname);
    else           fprintf (f, "( Submask %d \"%.8s\" for \"%.16s\". )\r\n", 
	opt.subm, a->fname, a->oname);

    lasermode (Laser_cutting, f);
    lasermode (Laser_off, f);
    Feedrate (f, " G01", quick);
    Feol (f);

    tool = bwid / 2.0;
/* MAY LATER WANT TO SPECIFY AN ACTUAL TOOL SIZE... */

    last = c;
    h = zlast = maskz (last);
/* Work offset location, using 59 for IMACS here */
    fprintf (f, " G59  ( Work offset for Multi-mask )\r\n");
    IDent (f);
    fprintf (f, " G90");

    VXYout (f, c, 1);
    lasermode (Autofocus_hold, f);
    Zout (f, h);
    fprintf (f, " ( Start location )\r\n");
    IDent (f);
// NOTE - this gas/shutter stuff could be a laser_mode call!!
    fprintf (f, " M10");	// select gas, 
    fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
    fprintf (f, " M14 ( Shutter open )\r\n");

// Still copying stuff from cutimacs here...

    lasermode (Autofocus_on, f);
//    lasermode (Autofocus_release, f);
    lastloc = last;

    Astop = 1;
    if (Astop) lasermode (Exact_stop_on, f);

/* Loop will compute all output things, just as for IMACS */

	/* - - - - - - - - - - - - - - - *\
	|    Cutting Mask Object Holes    |
	\* - - - - - - - - - - - - - - - */

#ifdef  CUTREPEAT
    n = object_loop (f, aq, inst, 0.0, opt.repc);
#else
    n = object_loop (f, aq, inst, 0.0, 1);
#endif

/* Switch to engraving mode and cut a label */

	/* - - - - - - - - - - - - *\
	|    Engraving of Labels    |
	\* - - - - - - - - - - - - */

// NOTE -- CHANGE THESE FOR THE MULTIMASK CASE.

	lasermode (Laser_engraving, f);

	Astop = 0;
	if (!Astop) lasermode (Exact_stop_off, f);

//	sprintf (line, "%.8s -%d- %.16s", a->fname, opt.subm, a->oname);
	if (serno > 0) sprintf (line, "%.8s %d.%d %.16s", a->fname, 
			serno, opt.subm, a->oname);
	else           sprintf (line, "%.8s -%d- %.16s", a->fname, 
			opt.subm, a->oname);
	fprintf (f, "( Label: %s )\r\n", line);	/* add comment */
	IDent (f);	// Debug line number

// Write small on the small masks...
	basl = 2.4 * SF;		// 3 mm letters in field

	h = zlast;
	xs = 0.5 * basl * strwid(line);	// half line length
//	rad = RLT * SF - ( (CVL > 0.0) ? 0.2 : 1.1 ) * basl;
//	sincos ( ALT*Degree - CVL*xs/rad, &sn, &cs);
//	base = mul2vect (make2vect(-sn, cs), CVL*basl);
	base = make2vect (basl, 0.0);
//	loc = mul2vect (make2vect(cs, sn), rad);
	loc = make2vect (-xs, 0.5*YSize*SF - 1.8*basl);
//	h = newloc (lastloc, h, loc, f);
//	strplot (loc, base,  CVL/rad, 0.0, line, pltimacs, f);
	strplot (loc, base,  0.0, 0.0, line, pltimacs, f);
// From start at loc, in direction base with curvature and slant,
// engrave line using pltimacs program onto file f.

	zlast = h;

// ABOVE IS COPY OF IMACS, NEEDS CHANGING

	/* - - - - - - - - - - *\
	|			|
	|  Cut alignment holes	|
	|			|
	\* - - - - - - - - - - */

    lasermode (Laser_cutting, f);
    fprintf (f, "( Alignment Holes. )\r\n");
    IDent (f);

/* WFCCD hole set, done with chole calls */
    w = HoleSize * SF - bwid;

/* Start with upper right, go cw and end at upper left. */
    c = mul2vect (make2vect(LholeX, LholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);

    c = mul2vect (make2vect(VholeX, VholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);

    c = mul2vect (make2vect(VholeX, -VholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);

    c = mul2vect (make2vect(LholeX, -LholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);

    c = mul2vect (make2vect(-LholeX, -LholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);

    c = mul2vect (make2vect(-VholeX, -VholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);

    c = mul2vect (make2vect(-VholeX, VholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);

    c = mul2vect (make2vect(-LholeX, LholeY), SF);
    zlast = chole (c, w, zlast, slow, 1, f);


/* Cut an appropriate border, with breaks and a notch */
// Going to do ngaps (here 6) segments
// Cut them in a circle which will get automatically changed by
// the offset programs.
// Be able to put in a notch later...


	/* - - - - - - - - - - *\
	|			|
	|    Cutting of edge	|
	|			|
	\* - - - - - - - - - - */


    fprintf (f, "( Edge of sub-mask. )\r\n");

/* Set rough cutting mode and feed rate for cutting the edge */
//    lasermode (Laser_roughcut, f);	// Cut at high speed
    lasermode (Laser_cutting, f);	// Change to regular cut.
//    fr = quick * 0.5;
    fr = slow;		// "cut" rate, should get 10.0 in/min



/* THE WAY IT REALLY IS...
  Is that there are actually 2 holes for 1/16 and 3/32 SS dowel pins
  located near the rim.  These are cut before we cut the rim itself,
  in the section above.  The rim is then perfectly circular */

// End the cutting of the masks, end of the last or 7th mask
// Must stop the cut, turn laser off (done already)
// and then hold autofocus, raise up one inch, and turn autofocus off.
// ADD THIS TO MASKCUT ALSO.

// To do this to the flat mask, start at upper left and go clockwise:
// Cut the top edge and corner
    c = mul2vect (make2vect (-0.5*(XSize+HoleSize) + Corner, 0.5*YSize), SF);
    h = newloc (lastloc, h, c, f);
// Set the feed rate here
    IDent (f);
    Feedrate (f, "", fr);
    Feol (f);
    SetSlew = 0;	// Do not change feed rate...
// Cut rest of border with fr set to slow
    lasermode (Laser_on, f);
    c = mul2vect (make2vect (0.5*XSize - Corner, 0.5*YSize), SF);
    h = carve_edge (&lastloc, c, h, f);
    c = mul2vect (make2vect (0.5*XSize, 0.5*YSize - Corner), SF);
    h = carve_edge (&lastloc, c, h, f);

// Cut the right edge, with a notch in it
    c = mul2vect (make2vect (0.5*XSize, Ynotch + 0.5*HoleSize), SF);
    h = carve_edge (&lastloc, c, h, f);
//    c = mul2vect (make2vect (0.5*XSize, Ynotch), SF);
    c = mul2vect (make2vect (0.0, -0.5*HoleSize), SF);
    h = cutarc (&lastloc, c, 0.5, f);
    c = mul2vect (make2vect (0.5*XSize, -0.5*YSize + Corner), SF);
    h = carve_edge (&lastloc, c, h, f);
    c = mul2vect (make2vect (0.5*XSize - Corner, -0.5*YSize), SF);
    h = carve_edge (&lastloc, c, h, f);

// Cut the bottom edge with two notches
    c = mul2vect (make2vect (Xnotch + 0.5*HoleSize, -0.5*YSize), SF);
    h = carve_edge (&lastloc, c, h, f);
//    c = mul2vect (make2vect (Xnotch, -0.5*YSize), SF);
    c = mul2vect (make2vect (-0.5*HoleSize, 0.0), SF);
    h = cutarc (&lastloc, c, 0.5, f);
    c = mul2vect (make2vect (-Xnotch + 0.5*HoleSize, -0.5*YSize), SF);
    h = carve_edge (&lastloc, c, h, f);
//    c = mul2vect (make2vect (-Xnotch, -0.5*YSize), SF);
    c = mul2vect (make2vect (-0.5*HoleSize, 0.0), SF);
    h = cutarc (&lastloc, c, 0.5, f);
    c = mul2vect (make2vect (-0.5*XSize + Corner, -0.5*YSize), SF);
    h = carve_edge (&lastloc, c, h, f);
    c = mul2vect (make2vect (-0.5*XSize, -0.5*YSize + Corner), SF);
    h = carve_edge (&lastloc, c, h, f);

// Cut left edge back to start point
    c = mul2vect (make2vect (-0.5*XSize, 0.5*YSize - Corner), SF);
    h = carve_edge (&lastloc, c, h, f);
    c = mul2vect (make2vect (-0.5*XSize + Corner, 0.5*YSize), SF);
    h = carve_edge (&lastloc, c, h, f);
    lasermode (Laser_off, f);
    

//double	carve_edge (vect2* next, vect2 e, double z, FILE* u)
// next is updated current location, e is absolute endpoint,
// z is height and file is output.
//double	cutarc (vect2* last, vect2 cd, double f, FILE* u)
// last is updated location, cd is center, f is ccw revolutions.

    lasermode (Autofocus_hold, f);
    lasermode (Laser_cutting, f);	// Restore regular mode
// Pull away by at least one inch (in hold mode) and turn autofocus off
    h += 30.0 * SF;
    fprintf (f, " G90 ");  Zout (f, h);  Feol (f);

/* Everything else is taken care of outside this program */

    return  n;	// count of features
}


	/* - - - - - - - - - - - - - - *\
	|				|
	|      Cut  WFCCD  masks	|
	|				|
	\* - - - - - - - - - - - - - - */

//#define  SUBMASKS  5
//#define  SUBMASKS  7
// Temporarily restricted due to laser machine problem
#define  WFCMULT  4

// int	cutmulti (sortq** mmhead, opts opt)
int	cutmul_wfccd (sortq** mmhead, opts opt)
{
int	masks = 0;
// int	files = 0;
int	i, j;
int	n;
int	sm;
//int	good;
sortq	*p;	// was.. , *q;
Obs	*obs;
FILE	*out;
char	base[8];
char	extn[4];
char	filename[16];
//struct	tm	*ut;
//time_t	now;

double	sdz;
double	vert;
double	h;
float	xs, ys;
int	mindx;	// mask index

// Return if nothing to be done
    if (mmhead == NULL) return 0;	// No argument

    if (opt.dbug > 0)
	printf (" cutmul_wfccd for %d mask files.\n", listcount (*mmhead) );

//    if (*mmhead == NULL) return 0;	// Null queue

/* We appear to have a queue, we need to process it */
    if (opt.subw != WFCMULT && opt.dbug > 0) 
	printf (" *Cutting %d submasks in each blank.\n", opt.subw);

/* Outline of what is to be done:

Process the queue in blocks of 7 obs pointers.
For each block we will:
  0) Set up the offset to 0, 0.
  1) Obtain a new file name
  2) Write the initial commands into that file
	>> Copy set-up from other cut programs
	>> Position cut head in center of first mask
  3) Do the individual processes for each of up to 7 pointers
	>> Pop the top obs file from the queue (break if none)
	>> Do a loop over all mask features
	>> Write the label information on periphery of mask
	>> Do commands to cut the perimeter and notch
	>> kill the obs structure and the queue member
	>> Set offsets for the next location
	>> Count another sub-mask processed
  4) Write the terminal commands into the file, close the file
	>> Retract cutting head, turn off, move to output position
	>> Close file
	>> Count another file output.
If more obs files exist, return to step 1.

The whole thing is wrapped in a for loop on *mmhead not being null.
If it becomes null within the loop, break.

.. */

    SF = inchout ? 1.0/Inch : 1.0;	// Make sure it's set.
    while (*mmhead != NULL) {

/* Set up offset to be 0,0 initially */
	setdisp (0.0, 0.0, 0.0, 0);	/* Set initial offset and angle */

/* Obtain file name and open the new file */

/* If a pointer to a valid <= 6 character base string has been provided
  in the options, use that.  Otherwise, generate a base string from
  the current date to be used here.  */
// Not using strlen here to check length...

	if (opt.base != NULL) {
	    sprintf (base, "%.6s", opt.base);  // First up to 6 char.s
	} else {
// GET THE BASE PART OF FILE NAME (SEE OUTLINE) GyyMddxx
	    sprintf (base, "G%s", curdate(0));
	}
	sprintf (extn, "%s", ".nc");

//  e.g.   n = nexfile (base, extn);
// where base is GyyMdd and extn is ".nc";
// then: sprintf (filename, "%s%02d%s", base, n, extn);
	j = nexfile (base, extn);
	if (j < 1) {
	    printf (" ** No available file name \"%s--%s\"; no output.\n",
		base, extn);
	    return masks;
	}
	sprintf (filename, "%s%02d%s", base, j, extn);
// FIND NEXT SERIALIZED NAME WHICH DOES NOT YET EXIST
// OPEN THE NEW FILE FOR WRITING
	out = iopen (filename, "w+");

// IF OUTPUT FILE IS NOT OPENED CORRECTLY, WE NEED TO WRITE
// AN ERROR MESSAGE AND EXIT.
	if (out == NULL) {
	    printf (" ** Open error; no further output.\n");
	    return  masks;
	}

	Lnum = 0;		// IDent counter
/* Initial set-up things copied from cutmask */
// set the Z offset
	inst = I_wfcd;	// Special case, this is wfccd instrument
	mindx = MaskNamIndx(InstCase(inst));
#ifdef  USE_CDAT
	sdz = CDhead[inst].dz * SF;
#endif
#ifdef  USE_GSTO
	sdz = *(double*)GS_lookup(&CDhead,inst,CD_dz,0) * SF;
#endif
#ifdef  USE_BSTO
	sdz = floatvalue (getval (&CDhead,MaskCase(inst),"dz")) * SF;
#endif
// CHANGE THE OFFSET COMPUTATION TO CYLINDER BLANK VALUES SIMILAR TO GISMO
	Zoff = 0.0;
// NEEDS TO CHANGE FOR GISMO CASE
/* Above was commented out, as Zoff is set in cututil.h where it is
  defined.  Actually needed dynamically when cutting multiple blanks,
  so it is not redefined to another height below.  Also, set this for
  real in the other routines which set it so multiple masks are done
  correctly with the height computed each time...    */

//	vert = maskz (make2vect ( SF*(flange+1.0), 0.0) );
// NEEDS TO CHANGE FOR GISMO CASE
	vert = 0.0;	// for gismo cutting jig location.  (used for wfccd)
// Offset is same as IMACS case
	Zoff = sdz - vert;
	if (opt.dbug > 0)  printf (" --Zoff = %7.3f; Z0 = %5.3f, Vh = %6.3f\n", 
				Zoff, sdz, vert);

/* Write initial commands into the file */
// WRITE INITIAL COMMENTS AND COMMANDS TO FILE
// GET THESE FROM OTHER OUTPUT FILE INSTANCES.
	if (opt.dbug >= 0) printf ("  Cutting multi-mask %s \n", filename);
/* Some debug stuff */
#if (DEBUG_CUT > 0)
	if (opt.dbug > 0) {
#ifdef  SIMPLECURVE
	    printf ("  Inst = %d, curve = %7.5f\n", inst, Curve );
#else
vect2	cv;
//		*GS_lookup(&CDhead,inst,CD_curve) );	// this doesn't work!
#ifdef  USE_CDAT
		cv = CDhead[inst].curve;
#endif
#ifdef  USE_GSTO
		cv = *(vect2*)GS_lookup(&CDhead, inst, CD_curve, 0);
#endif
#ifdef  USE_BSTO
		cv = vect2value(getval(&CDhead,MaskCase(inst),"Curve"));
#endif
	    printf ("  Inst = %d, curve = %7.5f (x), %7.5f (y)\n", inst, 
		cv.x, cv.y);
#endif  // on SIMPLECURVE
	}
#endif  // on DEBUG_CUT

// SET UP THE CUTTING PARAMETERS AS GLOBALS

// INCLUDE VERSION NUMBER OF THIS PROGRAM
//	fprintf (out, "( Multi-mask file by %s version %s%s )\r\n",
//		PROGNAME, VERSION, CTUVERS);
	fprintf (out, "( WFCCD mask file by %s version %s )\r\n",
		PROGNAME, XVERSION);
// Could split above into two lines if needed for compatability
//	fprintf (out, "( Instrument: %s )\r\n", InstName(inst) );

//	if (serno > 0) fprintf (out, "( Mask: %d  Instrument: %s )\r\n", 
//		serno, InstName(inst) );
//	else           fprintf (out, "( Instrument: %s )\r\n", InstName(inst) );

	if (serno > 0) fprintf (out, "( Instrument: %s  Mask: %d )\r\n", 
		MaskName(mindx), serno );
	else           fprintf (out, "( Instrument: %s )\r\n", MaskName(mindx) );

//	if (serno > 0) fprintf (stdout, "( Instrument: %s  Mask: %d )\r\n", 
//		MaskName(mindx), serno );
//	else           fprintf (stdout, "( Instrument: %s )\r\n", MaskName(mindx) );

/* Set up global cutting parameters */
	cutparams (mindx);

//	if (opt.dbug > 0) printf (" WFCCD multi -- Acon = %.3g\n", Acon);
	if (opt.dbug > 0) printf (" WFCCD cutpar (%s) -- Acon = %.3g\n", 
			MaskCase(mindx), Acon);

/* If we have a request for slow cutting, modify the slow and vslow
  rates to slower values for fine cutting... */
	if (opt.slow) {
double	ratio;
	    ratio = vslow/slow;
	    slow = vslow;
	    vslow = ratio*slow;
	}


/* Initial cutting commands, from cutimacs */
	lasermode (Autofocus_off, out);
	lasermode (Exact_stop_on, out);
	IDent (out);
	if (inchout == 0) {		// Warning section
	    fprintf (out, " ( ! DANGER!! -- Millimeter mode file! )\r\n");
	    fprintf (out, " M0 ( DANGER!! Millimeter mode! )\r\n");
	    fprintf (out, " ( Millimeters can cause a crash! )\r\n");
	}
	if (inchout) fprintf (out, " G20 ( Inch measurement )\r\n");
	else         fprintf (out, " G21 ( Metric measurement )\r\n");
/* Further commands from imacs cutting are in cutldss3 above */

/* Process up to 4 individual sub-masks into the output file */
	for (i=j=sm=0; i<opt.subw; i++) {	// loop over up to 4 sub-masks

	    opt.subm = i + 1;		// sub mask designator

	    ys = (i < 2) ? 1.0 : -1.0;
	    xs = (i%2 == 0) ? 1.0 : -1.0;
	    if (inchout) { xs /= Inch; ys /= Inch; }	// scaling offset!
	    // setdisp x: xs * (OffSet + 0.5*XSize), y: ys * (OffSet + 0.5*YSize)
	    // angle: 0.0, cf: 1
// ADD A NEW CALL TO setdisp HERE
	    setdisp (xs*(OffSet + 0.5*XSize), ys*(OffSet + 0.5*YSize), 0.0, 1);

	    p = *mmhead;
	    if (p == NULL) break;
	    *mmhead = pop_sq (*mmhead);
	    obs = (Obs*) p->q;
/* Process the sub-mask, and count our success */

/* SHOULD SET SF HERE DUE TO THE CTE --- */
#ifdef  USE_CTE
	    if (opt.xcte) {
double	tex;
		SF = inchout ? 1.0/Inch : 1.0;
		tex = CTE * (Tcut - obs->temp);
		if (opt.dbug > 0 && tex != 0.0)
		    printf (" Correcting thermal expansion of %9.3e\n", tex);
		SF *= 1.0 + tex;
	    }
#endif

	    n = cutwfccd (obs, out, opt);
	    if (n < 1) {	// No mask cut, decrement i to try again
		i--;
	    } else {	// cut a mask, total the mask and feature count
		j += n;		// features count
		sm++;		// sub-mask count
	    }

// IF ZERO FEATURES DONE (INCLUDING LABEL) DECREMENT i TO EXTEND LOOP

/* Done with sub-mask, kill the obs structure and the sortq entry p */
	    obs = kill_obs (obs);
	    free (p);

// /* Set the offsets to the next location (based on i) */
// // SET OFFSETS
// No, we set these above based on i...

	}	// Loop over input files indexed by i, 0-6

	setdisp (0.0, 0.0, 0.0, 0);	// Reset displacement for end commands

/* Count masks, files output */
//	masks += j;
	masks += sm;

/* If any submasks written, fill rest of file.  Close file */
	if (j > 0) {		// Some entries
//// Increased to clear clamps which can be 51 mm high
	    h = maskz (make2vect(999.0, 999.0)) + 60.0 * SF;
	    end_mask (out, h);

#ifdef  USE_ACON
	    fill_cuttime (out);
#endif
// CLOSE FILE

/*  Put in signature */
	    signfile (out, '(', dodebug-1);

	    iclose (out);
	    printf (" -Processed %d features in %d sub-masks.\n", j, sm);

// increment serial number if used
	    if (serno > 0) serno++;

// COUNT FILES -- already done?
	} else {		// No entries
// CLOSE AND DELETE FILE, IT IS NULL
// rm remove rename unlink
// unlink requires unistd include; already needed for access.
	    iclose (out);
	    unlink (filename);	// Deletes the file
	}	// Null mask output test

    }	// Null ldss mask queue test

// Return number of masks processed
    setdisp (0.0, 0.0, 0.0, 0);	/* Set default offset and angle (good form) */
//    Offset = 0.0;
    return  masks;	// ALSO NEED TO RETURN FEATURES?
}	// end cutmul_wfccd




/* ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- */

	/* - - - - - - - - - - - - - - *\
	|				|
	|     Cut  Multiple  Masks	|
	|				|
	\* - - - - - - - - - - - - - - */

int	cutmulti (sortq** mmhead, opts opt)
/* Process multi-mask queue, separating different mask types into
  separate queues, and calling those processing programs separately.  */
{
sortq	*ldhead = NULL;
sortq	*gshead = NULL;
sortq	*wfhead = NULL;
sortq	*hp;
Obs	*obs;
Mindx	mt;
int	k=0;	// the return value
#ifdef  GISMONOTCH
int	kn=0;	// count of gismo masks made
#endif

// Loop to run through the given queue.
    while (*mmhead != NULL) {

// Pop the head mask from the queue
	hp = *mmhead;
	*mmhead = pop_sq (*mmhead);
	obs = (Obs*) hp->q;

// See what kind it is
	mt = MaskIndx (obs);
	if (opt.dbug > 0) {
	    printf (" Mask %s (%s) of type %s found.\n", obs->fname,
		obs->oname, MaskName(mt) );
	}

// Push it into the appropriate queue
	if      (mt == M_lds3 ) ldhead = push_sq (hp, ldhead);
	else if (mt == M_gismo) gshead = push_sq (hp, gshead);
	else if (mt == M_wfcd ) wfhead = push_sq (hp, wfhead);
	else {
	    printf (" ** Mask %.8s unsupported type %s killed.\n",
		obs->fname, MaskName(mt) );
	    obs = kill_obs (obs);
	    free (hp);
	}
    }
    kn = (gshead != NULL);	// Have actual gismo masks.

    if (opt.rerd) opt.rerd = 1;	// reset for multi-mask case

// Call separate cutting tasks for each known type (some may be null)
    k += cutmul_ldss (&ldhead, opt);
    k += cutmul_gis  (&gshead, opt);
    k += cutmul_wfccd (&wfhead, opt);
/* Add special mask features for multi masks here */
#ifdef  GISMODECKER
    if (opt.gsdk) k += gismodecker (opt);
#endif

#ifdef  GISMONOTCH
    if (opt.gnch && kn) k += cutmul_gis (NULL, opt);
/* In cutmul_gis, we insure that if the gshead pointer is not null, we
  cut actual masks and not a notch feature mask.  However, if we did
  that, the notch mask did not get cut.  Here we check for that case,
  and call again for the notch case.  However, if there were no actual
  gismo masks to be cut, the notch mask has already been done, so we
  need to test for that and skip doing it twice.  */
//printf (" ** cutmulti location 2 done cutmulgis second call\n");
#endif

    return  k;
}


	/* - - - - - - - - - - - - - - *\
	|				|
	|     Cut individual mask	|
	|				|
	\* - - - - - - - - - - - - - - */


int	cutmask (char* infile, opts opt, sortq** mmhead)
/* Read the input SMDF file, produce an output file for it.  Return
  number of files correctly processed (1 if done, 0 if none). 
  Returns -1 if a mask data file is rejected.  */
{
#define	outext  ".nc"
/* outext is extension forced onto output file */
Obs	*obs;	/* Observation structure read */
FILE	*OUT;
char	outfile [80];
char	buffer [80];	/* Scratch */
char	line [120];
char	fullinfile[128];
char	fc;
char	prefix [4];
double	sdz;	// Scaled z offset
double	vert;	// Vertex correction
int	i;
int	r = 0;
int	k;
int	nj;
time_t	now;
sortq	*px;

#ifdef  TESTMODE
/* Test mode -- is a special debug feature.  We do the test here, and
  then return without processing.  */
    if (opt.test) {
char	bufr[32];
time_t	now, new;

// Here, we test the time stamp things...

// Get a time stamp, report it.
	strcpy (bufr, timestamp (&now));
	printf ("  Time Stamp of %d = %s\n", (int)now, bufr);

// Find the unix time from the time stamp
	new = unstamp (bufr);

// Report the unix time translated by the system
	printf ("  Translated:   %d -- %d\n", (int)new, (int)(new-now) );
	printf ("  UT time: %s",
		asctime(gmtime(&now)) );
// Note that asctime returns string terminated by \n ...
	printf ("  UT tran: %s",
		asctime(gmtime(&new)) );

// Also show the current local time (from translated, to verify which is right)
	printf ("   Local:  %s",
		asctime(localtime(&new)) );

// End test.
	return  r;
    }
#endif	// on TESTMODE


/* If opt.rimh is on, we name the output file, null the obs pointer,
  and call the imacs program.  That takes null to mean rim hole stuff. */

    if (opt.rimh) {
//	inst = 0;	// It's imacs
	inst = M_imacs;	// It's imacs
	obs = NULL;
	strcpy (prefix, "I");
	sprintf (outfile, "%s%s", prefix, "rim");
	nj = 0;
	if (opt.dbug > 0) printf (" *Rim file \"%s\" starting.\n", outfile);
	strcpy (fullinfile, "Maskcut Data");
/* Another special case, the gismo notch mask test... */
//#ifdef  GISMONOTCH
/* THIS SECTION SHOULD BE UNUSED */
//    else if (opt.gnch) {
//	inst = M_gismo;
//	obs = NULL;
//	strcpy (prefix, "G");
//    }
//#endif  // on GISMONOTCH
    } else {
#if (DEBUG_CUT > 1)
	printf (" Processing file \"%s\".\n", infile);
// The above is redundant, since read_smdf (below) normally reports this.
#endif
	strcpy (outfile, infile);
    
	obs = read_smdf (infile, NULL, 0);
	if (obs == NULL) {
	    printf (" Verry sorry, could not read \"%s\".\n", infile);
/* Use return rather than goto finish; there is no obs data to clean. */
	    return  0;
	}
// The above section was moved up a few lines to avoid a crash and dump
// in the case that the file was not available.  KDC 08/02-22.

/* re-set the reorder count when reading a new file */
	if (opt.rerd) opt.rerd = 1;
/* To eliminate duplicate reordering operations, the rerd option is
  incremented each time reorder is called.  However, if multiple files
  are being processed, it should be reset for each file so as to not
  inadvertently omit reordering of the second and subsequent files */
// This may not help ldss3 or other multiple files...

	obs->dlevel = opt.dbug;
// That's a null optical element queue, and no special handling.
	strcpy (fullinfile, LastXopen);
	if (opt.dbug > 0) printf (" Source file opened: \"%s\".\n", fullinfile);
    }

    r = 0;		/* Return code */

/* Check for, and possibly reject, any data created by an obsolete
  program, as found in the obs file.  */
/* compare to current program version rather than FTP, since that's the
  processing features which we wish to protect.  - It would do no good
  to object to or mention a beta version data when running a beta version
  of the processing program.  If this program is obsolete with respect
  to the FTP data, that got stopped much earlier.  */
    if (obs == NULL) {
	k = 0;		// For the rim file case
    } else {
	k = comp_vers ((Vers*)obs->source, opt.cp);   // was-- opt.fp);
    }
    if (k < -2) {
	printf (" ** Warning!  File %s made by UNKNOWN software!\n", fullinfile);
	if (opt.test) {
	    printf ("  (We let it slide in test mode)\n");
	    k = 3;
	}
    } else if (k != 0) {
	printf (" ** Warning!  File %s made by %s version %s\n",
		fullinfile,
		((Vers*)obs->source)->tag, fvstr(((Vers*)obs->source)->v) );
	printf ("   was produced by ");
   	switch (k) {
	    case -2: printf ("Obsolete");		break;
	    case -1: printf ("Older");			break;
	    case  0: printf ("Current");		break;
	    case  1: printf ("advanced (beta)");	break;
	    default: printf ("unknown");		break;
	}
	printf (" software.\n");
    }
    if (k < -1 && opt.test != 7) {  // special dispensation for -t7 here
	printf ("  Operation on %s data rejected.\n",
		(k == -2) ? "Obsolete" : "Unknown" );
	r = -1;
	goto finish;
/* At finish, we clean up the obs structure, and return r */
    }


/* Inch/metric logic set or reset here */
    SF = inchout ? 1.0/Inch : 1.0;

/* NOTE -- Will later use SF to implement the temperature coeficient
  of expansion; SF *= 1.0 + cte*(tcut - tobs); to cut at a greater
  dimension if cutting is warmer than observing...  */

/* Move scale factor set on CTE to below */
/* NOTE - WARNING - if CTE used to set SF here, it can 
  give message in wrong place for the ldss3 multiple mask case */

/* Try to find instrument name... */
    if (! opt.rimh) {	// exclude rim case
    /* Dump a few things for debug... */
	if (opt.dbug > 0) {
	    printf (" Observing with instrument \"%s\".\n",
		    obs->Instr_mt->name);
	}
    
/* Try to find instrument name... */
//	inst = InstIndx (obs->Instr_mt->name);
	inst = MaskIndx (obs);
  
	if (inst < 0) {	/* Unknown instrument message */
	    printf (" Sorry, I don't know how to make a mask for");
/* If no instrument, or front end exists, write front end;
  otherwise write instrument  */
	    if (obs->Frontend != NULL)
		printf (" (instrument) \"%s\".\n", obs->Frontend->name);
	    else
		printf (" instrument \"%s\".\n", obs->Instr_mt->name);
	    printf ("  ( I only know: ");
	    for (i=0; i < Masks; i++) {
		if (i > 0) printf (", ");
		printf ("\"%s\"", MaskName(i));
		fflush (stdout);
	    }
	    printf (" ).\n");
	    goto  finish;	// allows for cleanup
	}

/* Take this section out, use cutparms in individual cutting programs */
/* Fill in the older globals from the cfig structure */
#ifdef  USE_CDAT
	flange   = CDhead[inst].flange;
	aangle   = CDhead[inst].aangle;
	aclear   = CDhead[inst].aclear;
	LongSlit = CDhead[inst].LongSlit;
	SlitGap  = CDhead[inst].SlitGap;
	avtool   = CDhead[inst].tool;
#ifdef  SIMPLECURVE
	Curve    = CDhead[inst].curve.x;
#endif
#ifdef  USE_CTE
	CTE      = CDhead[inst].cte;
	Tcut     = CDhead[inst].tcut;
#endif
#ifdef  USE_ACON
	Acon     = CDhead[inst].acon * 1.e-6;
#endif
#endif	// on USE_CDAT

#ifdef  USE_GSTO
	flange = *(double*)GS_lookup (&CDhead, inst, CD_flange, 0);
	aangle = *(double*)GS_lookup (&CDhead, inst, CD_aangle, 0);
	aclear = *(double*)GS_lookup (&CDhead, inst, CD_aclear, 0);
	LongSlit = *(double*)GS_lookup (&CDhead, inst, CD_LongSlit, 0);
	SlitGap  = *(double*)GS_lookup (&CDhead, inst, CD_SlitGap, 0);
	avtool   = *(double*)GS_lookup (&CDhead, inst, CD_tool, 0);
#ifdef  SIMPLECURVE
	Curve    = *(vect2*)GS_lookup (&CDhead, inst, CD_curve, 0).x;
#endif
#ifdef  USE_CTE
	CTE      = *(double*)GS_lookup (&CDhead, inst, CD_cte, 0);
	Tcut     = *(double*)GS_lookup (&CDhead, inst, CD_tcut, 0);
#endif
#ifdef  USE_ACON
	Acon  = *(double*)GS_lookup (&CDhead, inst, CD_Acon, 0) * 1.e-6;
#endif
#endif	// on USE_GSTO
#ifdef  USE_BSTO
	flange   = floatvalue(getval(&CDhead,MaskCase(inst),"Flange"));
	aangle   = floatvalue(getval(&CDhead,MaskCase(inst),"Aangle"));
	aclear   = floatvalue(getval(&CDhead,MaskCase(inst),"Aclear"));
	LongSlit = floatvalue(getval(&CDhead,MaskCase(inst),"LongSlit"));
	SlitGap  = floatvalue(getval(&CDhead,MaskCase(inst),"SlitGap"));
	avtool   = floatvalue(getval(&CDhead,MaskCase(inst),"Rtool"));
#ifdef  SIMPLECURVE
	Curve    = floatvalue(getval(&CDhead,MaskCase(inst),"Curve"));
#endif
#ifdef  USE_CTE
	CTE      = floatvalue(getval(&CDhead,MaskCase(inst),"CTE"));
	Tcut     = floatvalue(getval(&CDhead,MaskCase(inst),"Tcut"));
#endif
#ifdef  USE_ACON
	Acon     = floatvalue(getval(&CDhead,MaskCase(inst),"Acon")) * 1.e-6;
#endif
#endif	// on USE_BSTO
/* Take above section out, use cutparams in individual masks */

	if (opt.dbug > 0) printf (" cutmask -- Acon = %.3g\n", Acon);

	switch (inst) {
	    case M_imacs:	// IMACS
		strcpy (prefix, "I");
		break;
	    case M_cfg:		// Centerfield
	    case M_lds2:	// LDSS-2
		strcpy (prefix, "L");
		break;
// NOTE THAT CASE 3 IS NOT YET DEFINED AND WILL NEED TO BE IMPLEMENTED
// IN INTGUI/MASKGEN AND UTILITY PROGRAMS BEFORE THIS WILL WORK.
	    case M_lds3:	// New LDSS-2
	    case M_gismo:	// Also GISMO is multi-cut
	    case M_wfcd:	// Also WFCCD masks
// Put the obs structure found onto a queue provided
// and then return; the queue is processed later.
		px = (sortq*) malloc (sizeof(sortq));
		px->next = px->last = NULL;
		px->q = obs;
		*mmhead = push_sq (px, *mmhead);
		return  0;
/*  === MULTI-CUT MASKS EXIT HERE === */
	    default:
		*prefix = '\0';
		break;
	}
    }	// End section excluding rim holes feature

/* Move scale factor set to here from above */
#ifdef  USE_CTE
    if (opt.xcte && ! opt.rimh) {
double	tex;
//	SF *= 1.0 + CTE * (Tcut - obs->temp);
	tex = CTE * (Tcut - obs->temp);
	if (opt.dbug > 0 && tex != 0.0)
	    printf (" Correcting thermal expansion of %9.3e\n", tex);
	SF *= 1.0 + tex;
    }
#endif

/* Set the Z offset.  It depends on instrument... */
//    Zoff = SF * Afocus;

#ifdef  USE_CDAT
    sdz = CDhead[inst].dz * SF;  // Scaled z offset
#endif
#ifdef  USE_GSTO
    sdz = *(double*)GS_lookup (&CDhead,inst, CD_dz,0) * SF; // Scaled z offset
#endif
#ifdef  USE_BSTO
    sdz = floatvalue (getval (&CDhead,MaskCase(inst),"dz")) * SF;
#endif

    Zoff = 0.0;
    switch (inst) {
	case M_imacs:		// IMACS
	case M_lds3:		// LDSS-3 (and we don't get here)
	    vert = maskz (make2vect ( SF*(flange+1.0), 0.0) );
// Note, done with Zoff of zero here; the current vertex height.
//	    Zoff = -39.018;
	    break;
// NOTE:  Need gismo case computation here?  NO - gismo are multimask,
// and are processed by cutmulti above; they will need a similar section.
	case M_lds2:		// LDSS-2
	case M_cfg:		// Centerfield
	default:
	    vert = 0.0;
// flat plates have no vertex correction.
    }
    Zoff = sdz - vert;
    if (opt.dbug > 0)  printf (" --Zoff = %7.3f; Z0 = %5.3f, Vh = %6.3f\n", 
				Zoff, sdz, vert);

    if (! opt.rimh) {	// exclude rim case
/* Find number of objects to be plotted/cut; if insufficient, send a
  message and exit before creating output. */
	nj = listcount (obs->ob);
	if (nj < 1) {
	    printf (" ** No mask objects in the list; nothing written.\n");
	    goto finish;	// allow cleanup
	}
    
/* Set up the output file.  The file name is in obs->fname, and does not
  have an extension with it.  Add the "outext" and open this file.  If it
  already exists, ask for permission first. */
/* Put a prefix on the start of the output file... */
    
	sprintf (outfile, "%s%.8s%c", prefix, obs->fname, AN);
	outfile[9] = '\0';
	trims (outfile, "\t ");
    }	// section excluding rim hole feature
//    if (strlen(outfile) < 8) strcat (outfile, inchout ? "i" : "m");
    if (inchout == 0) strcat (outfile, "m");  // Only for metric...
    strcat (outfile, outext);
/* Result is output file name as specified in the file. */

/* Now, see if outfile already exists... */
    if (access (outfile, F_OK) == 0) {	/* F_OK should be 0 */
	sprintf (line, " File %s exists.  Overwrite? [no]", outfile);
	askmestr (buffer, 40, line);
	fc = *skipover (buffer, WHITESPACE);  // First real character
	if (fc == AN) goto  finish;   // Null reply
	if (!(fc == 'y' || fc == 'Y')) {
	    printf ("  Nothing Written. \n");
	    goto  finish;	// allows for cleanup
	}
    }
    OUT = iopen (outfile, "w+");
    if (OUT == NULL) {
	printf (" Error opening %s.\n", outfile);
/* FIX UP THAT MESSAGE; ADD REASON, ETC. */
    } else {			// Valid OUT file, process the input	

	Lnum = 0;		// IDent counter
// Allow for optional reordering to optimize path, if requested
// Set up number of interchanges...
	if (! opt.rimh) {
	    i = (2 * nj + 1) / 3;
	    if (opt.rerd == 1) {
		obs->ob = reorder (obs->ob, i, opt.dbug);
		opt.rerd++;
	    }
// Change above to 1+DEBUG_UTL to suppress messages.
	}

/* Come up with a time stamp field.  If time is same as lastime, it
  would not be unique, so we wait until it is unique... */

/* NOTE -- no standard C structure can put something in a loop and
  before the test.  So, we construct our own loop structure using
  the despised goto statement as follows:  */

Getit:	strcpy (tstamp, timestamp (&now));
	if (now == lastime) {
	    usleep (30000);
	    goto Getit;
	}
	lastime = now;
// The do-it-yourself loop above seems optimally efficient.
// See older versions (maskcut.c39 and before) for other methods.

/* NOTE -- An alternate method would be to increment "now" if it
  happened to be same as lastime, and use tmstamp to get the
  appropriate time stamp.
  HOWEVER, the problem exists that several time stamps could then
  be generated for successive seconds very quickly if this program
  is used in a script, and a subsequent instance of this program
  could find a system time which gives a duplicate time stamp.
  What we are doing is to delay to use the system time as a unique
  global value to insure unique time stamps.    ... */

/* Also, obtain the ADC status letter... */
	if (inst == M_imacs && !opt.rimh)
	    adc = (strcmp (obs->Tel_scope->name, "MagNoADC") ) ? 'Y' : 'N';
	else adc = 'N';
/* inst is 0 for imacs, 1 for ldss;  name is the telescope name, and
  will be MagNoADC for the no adc case.  */

/* Call proper processing routine based on "inst" index */
	if (opt.dbug > 0) printf ("Process as instrument type %d (%s).\n",
	    inst+1, MaskName(inst) );  // was.. inames[inst] );

/* Let user know the output file name */
	if (opt.dbug > 0) printf (" --Writing output file \"%s\"\n", outfile);

/* Print the principle title comment first in file */
	if (opt.rimh) {
	    fprintf (OUT, "( IMACS flange holes. )\r\n");
	} else {
	    fprintf (OUT, "( \"%.8s\" for \"%.16s\". )\r\n", obs->fname, obs->oname);
/* COULD ALSO COPY GOLBAL COMMENTS IN OBS STRUCTURE... */
	}

/* Insert the time stamp and adc status in a comment here */
//	fprintf (OUT, "( \"%s\" ADC \"%c\" )\r\n", tstamp, adc);
	if (serno > 0) fprintf (OUT, "( \"%s\"  Mask %d )\r\n", tstamp, serno);
	else           fprintf (OUT, "( \"%s\" )\r\n",         tstamp);

/* Add the name of the source file here */
/*  ... all we know in this routine is the "infile" name which was
  passed to read_smdf earlier.  That program creates a list of possible
  suffex values (.SMF, .smf) which is passed to iopenx and only in there
  is an actual file name generated and used.  So someone, either iopenx
  or read_smdf will need to set a global buffer to the input file name
  which is actually used to get an accurate output here...  */
	fprintf (OUT, "( From \"%s\" )\r\n", fullinfile);


/* Put in the program and version producing it... */
//	fprintf (OUT, "( By %s version %s%S )\r\n", PROGNAME, VERSION, CTUVERS);
	fprintf (OUT, "( By %s version %s )\r\n", 
#ifdef  SOLARIS
		PROGNAME, XVERSION);
// Solaris has an annoying lint message if the cast is used.  Both work.
#else
		PROGNAME, XVERSION);
//		PROGNAME, VERSION, (wchar_t*)CTUVERS);
// Changed to eliminate Linux error message, causes lint message ?!
#endif

/* Enter the instrument name in the file as a comment */
	fprintf (OUT, "( Instrument: %s )\r\n", MaskName(inst) );
// That comment may be used later to identify the instrument

/* Set up the cutting parameters */
	cutparams (inst);

//	if (opt.dbug > 0) printf (" Normal maskcut -- Acon = %.3g\n", Acon);
	if (opt.dbug > 0) printf (" Normal maskcut (%s) -- Acon = %.3g\n", 
			MaskCase(inst), Acon);
/* If we have a request for slow cutting, modify the slow and vslow
  rates to slower values for fine cutting... */
	if (opt.slow) {
double	ratio;
	    ratio = vslow/slow;
	    slow = vslow;
	    vslow = ratio*slow;
	}
    
	switch (inst) {
	    case M_imacs:	// IMACS
/* Call the routine here with obs as an argument, also outfile. */
/* Return code is 1 (good) or 0 (fail) */
		r = cutimacs (obs, OUT, opt);
		break;	/* omit for debug, and fall through */
	    case M_lds2:	// LDSS-2
		r = cutldss (obs, OUT, opt);
		break;
	    default:
		printf (" No output routine yet for %s.\n", MaskName(inst));
	}

	if (opt.dbug >= 0) printf (" --Cut %d features in mask.\n", nj);
#ifdef  USE_ACON
	fill_cuttime (OUT);
#endif

/* Add checksum and signature */
	signfile (OUT, '(', dodebug-1);

	iclose (OUT);	// clean up the act.
    }	// End of OUT file writing section

/* Which box looks better? */

	/* * * * * * * **
	**    finish   **
	** * * * * * * */

	/* - - - - - - - *\
	|     finish      |
	\* - - - - - - - */
finish:
/* NOTE -- We use a goto here to make sure obs structure gets
  cleared before exiting the routine.  */

/* Clear the obs structure (done in mgutils) */
    kill_obs (obs);

/* All done, now we may return */
    return  r;
}



/* ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- 
   ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- */

        /* - - - - - - - - *\
        |                   |
        |    Main Program   |
        |                   |
        \* - - - - - - - - */

#define  GETOPTUSE
//#undef  GETOPTUSE
/*  Symbol  GETOPTUSE  is used to turn on (or off) the use of the
  standard "getopt" utility in parsing the command line.  Use this
  local define to enable the "old" method quickly if needed.  */


/*  ====  MAIN PROGRAM  ====  */

#define  CUTFILE  "maskcut.cfg"

int  main (int argc, char** argv)
{
#ifdef  GETOPTUSE
extern char *optarg;
extern int optind,opterr,optopt;
int c;
#else
char	*pp, px;
int	i;
#endif
int	n=0;
int	j, k;
int	r=0;	// rejected masks
int	dohelp = 0;
opts	opt;
char	infile [80];
char	*filename;
namlist	*head = NULL;
sortq	*mmhead = NULL;
namlist	*comps = NULL;
#ifdef  GUIDATE
int	xcte = 1;	// Initial xcte default
#else
int	xcte = 0;	// Initial xcte default (could be 1)
#endif  // on GUIDATE
#ifdef  GISMONOTCH
double	yy;		// scratch space
#endif
char	*vstr;

/* Obtain file name for input, from parameter or user */

/* Mark I program.  */
    opt.dbug = 0;	/* Debug level */
    opt.rerd = 1;	/* Reorder */
    opt.rimh = 0;	/* Flange hole file feature */
#ifdef  GISMODECKER
    opt.gsdk = 0;	/* Decker mask feature */
#endif
    opt.test = 0;	/* Test mode */
    opt.slow = 0;	/* Slow cutting option */
    opt.subc = SUBMASKS;  /* Sub mask count feature */
    opt.subg = GISMULT;
    opt.subw = WFCMULT;
    opt.repc = 1;	/* Repeat count */
#ifndef  USE_CTE
    xcte = 0;		/* Drop use of CTE -- always zero here. */
#endif
    opt.xcte = xcte;	/* Use CTE default */
#if  GISMOFFSET || GISMONOTCH
    opt.offset = 0.0;	/* Basic offset set here */
#endif
#ifdef  GISMONOTCH
    opt.gnch = 0;	/* default is no notch special mask */
#endif
    opt.base = NULL;	/* Base file name string */
    opt.cp = NULL;
    opt.fp = NULL;
    *infile = '\0';	/* used for single execution */

    mmhead = NULL;
    CDhead = NULL;	/* Initialize */

/* If no names are given, we do the help output.  If help is requested,
  we break the loop with help flag set. */

// Create the combined version dynamically for titles
    vstr = combinev (MASKCTV, CUTUTLV);
    strcpy (XVERSION, vstr);
    free (vstr);

/* Because flags are decoded with files in a loop here, we must check
  the version outside the loop, and before any processing is done.
  Therefore it is done here before we even print the version.  */
/* See if this version is any good, or if we continue.  */
    comps = pushtxt (MASKCTT, comps);
    comps = pushtxt (CUTUTLT, comps);
    k = check_ver (comps, &(opt.cp), &(opt.fp));
    comps = kill_name (comps);
// if feature exists, we could exit on k < -2 here.
    if (k < -1) {
	printf (" ** Operation of Obsolete software rejected.\n");
//	return  1;
	k = 1;
	goto main_finish;
    }



	/* - - - - - - - - - - - - - - *\
	|				|
	|	Scan Command Line	|
	|				|
	\* - - - - - - - - - - - - - - */

/* We have 2 ways of scanning, either the "getopt" utility, or a direct
  method.  These should be equivalent.  Newer use should be "getopt" as
  it is easier to modify the flag meanings.  The one which is used is
  selected by a symbol definition just before the main program.    */

#ifdef  GETOPTUSE
/* Decode all the options of the command line */
/* :: follows letter with optional argument, : follows one with required
  argument, otherwise there are no arguments to the options.  */
// flags: d::qt::imrsf   g (GISMODECKER)  k (GISMONOTCH)  n:ec::b::
    while((c = getopt(argc, argv, "b:c::d::efgikmn:qrst::xy::h")) != -1) switch(c) {
	case 'b':				// base name string
	    opt.base = optarg;		break;
	case 'c':				// limit count submasks 1-7
	    k = optarg == NULL ? 0 : atoi (optarg);
	    if (k > 0 && k <= 7) opt.subc = k;	break;
	case 'd':				// debug level
	    opt.dbug = dodebug = optarg == NULL ? 1 : atoi (optarg);	break;
#ifdef  USE_CTE
//	case 'e':	opt.xcte = 1 - xcte;	break;	// toggle expansion
	case 'e':	opt.xcte = 0;	break;	// turn off
#else
	case 'e':	opt.xcte = 0;	break;	// not allowed to be on
#endif
// Alternate is opt.xcte = 0;  Can turn off if was on, but
// cannot turn it on if initial default is off (ifndef USE_CTE)
	case 'f':	opt.rimh = 1;	break;	// flange holes
#ifdef  GISMODECKER
	case 'g':	opt.gsdk = 1;	break;	// decker
#endif
	case 'i':	inchout  = 1;	break;	// inch output (default)
#ifdef  GISMONOTCH
	case 'k':	opt.gnch = 1;	break;	// gismo notch special
#endif
	case 'm':	if (mmallow) inchout = 0;	break;
	case 'n':	k = optarg == NULL ? 0 : atoi (optarg);
		if (k > 0) serno = k;	break;	// serial number
	case 'q':	opt.dbug = 0;	break;	// quiet
	case 'r':	opt.rerd = 0;	break;	// omit reordering
	case 's':	opt.slow = 1;	break;	// slow cutting
//#ifdef  TESTMODE
//	case 't':	k = optarg == NULL ? -1 : atoi (optarg);
//	    if (k >= 0) opt.test = k;	break;	// test mode setting
	case 't':
	    opt.test = optarg == NULL ? 1 : atoi (optarg);	break;
//#endif
#ifdef  CUTREPEAT
	case 'x':	opt.repc = 2;	break;	// repeat allowed
#endif
#if GISMOFFSET || GISMONOTCH
	case 'y':
	    yy = optarg == NULL ? 0.0 : atof (optarg);
	    if (fabs(yy) < 3.0) opt.offset += yy;	break;
/* Execution of gismo notch and gismo decker if specified is in cutmulti */
#endif
	case 'h':	dohelp = 1;		break;	// parameter help
	default:
	    printf (" unused %c option\n", c);	break;
    }  /* end of switch, only statement in the getopt while loop */


/* This code is the standard thing to put all the arguments in a list,
  which will then be processed later after any help or other things have
  been processed.  */
    for (c = optind; c < argc; c++) {
	if (dohelp) break;	// do not do anything if help requested.
	if (opt.dbug > 0) printf (" Adding %s to argument list.\n", argv[c]);
	head = pushtxt (argv[c], head);  /* Collect arguments */
    }


#else  // on GETOPTUSE
/* Scan for switches and arguments */
/* Loop to translate switches and process files */
    for (n=0,i=1; i<argc; i++) {    /* argv[0] is command name */
//printf (" ** in main location 1, dohelp=%d\n", dohelp);	// debug only
	pp = argv[i];
	if (*pp == '-') {       /* Scan normal switch */
	    while ((px = *++pp) != '\0') {
		if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
//		if      (px == 'd')  opt.dbug = 1;  /* Debug output */
		if      (px == 'd') {	/* Set debug value [1] */
		    if (pp[1] == AN) k = 0, opt.dbug = 1;
		    else k = sscanf (++pp, "%d", &j);
		    if (k == 1) opt.dbug = j;
		    break;
		}
		else if (px == 'q')  opt.dbug = 0;  /* Debug output */
//		else if (px == 't')  opt.test = 0;  /* No test mode */
//#ifdef  TESTMODE
		else if (px == 't') {	/* Set test mode flag value */
		    if (pp[1] == AN) k = 0, opt.test = 1;
		    else k = sscanf (++pp, "%d", &j);
		    if (k == 1) opt.test = j;
		    break;
		}
//#endif
		else if (px == 'i')  inchout = 1;  /* Inch output */
		else if (px == 'm' && mmallow) inchout=0;/* Metric output */
		else if (px == 'r')  opt.rerd = 0;  /* No Reorder */
/* Above is for special masks; confusion can result as this option
  formerly was implemented to set reordering true... */
/* Above is an accomodation for mistakes in putting in -r vs. +r */
		else if (px == 's')  opt.slow = 1;
		else if (px == 'f')  opt.rimh = 1;  /* flange holes */
#ifdef  GISMODECKER
		else if (px == 'g')  opt.gsdk = 1;  /* decker */
#endif
#ifdef  GISMONOTCH
		else if (px == 'k')  opt.gnch = 1;  /* notch special */
#endif
		else if (px == 'n') {		    /* Serial number */
		    if (pp[1] == AN) k = 0;
		    else k = sscanf (++pp, "%d", &j);
		    if (k==1 && j>0) serno = j;
		    break;
		}
		else if (px == 'e')  opt.xcte = 0;  /* Expansion comp. */
		else if (px == 'c') {	// limit to count of submasks
		    if (pp[1] == AN) k = 0;
		    else k = sscanf (++pp, "%d", &j);
		    if (k==1 && j>0 && j<=7) opt.subc = j;
		    break;
		}
		else if (px == 'b') {	// Capture base name string
		    if (strlen(++pp) > 0) {	// Any non-null string
			opt.base = pp;	// Store pointer
		    }
		    break;
		}
#ifdef  CUTREPEAT
		else if (px == 'x')  opt.repc = 2;  /* repeat features */
#endif
#if GISMOFFSET || GISMONOTCH
		else if (px == 'y') {  // offset parameter
		    if (pp[1] == AN) k = 0;
		    else k = sscanf (++pp, "%lf", &yy);
		    if (k == 1 && fabs(yy) < 3.0) {	// limit offset size
			opt.offset += yy;	// parameter increments offset
		    }
		    break;
/* Execution of gismo notch and gismo decker if specified is in cutmulti */
//printf (" ** in main location 2, dohelp=%d\n", dohelp);	// debug only
		}
#endif
		else if (px == 'h')  dohelp = 1;  /* show help */
	    }
	} else if (*pp == '+') {        /* Scan positive switch */
	    while ((px = *++pp) != '\0') {
		if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
		if      (px == 'd') {
			opt.dbug = 2;      /* More Debug */
			printf ("\n PLEASE REPLACE +d FLAG WITH -d2 FLAG!!\n\n");
		}
		else if (px == 'q') {
			opt.dbug = 0;  /* Debug output */
			printf ("\n PLEASE REPLACE +q FLAG WITH -q FLAG!!\n\n");
		}
		else if (px == 'r') {
			opt.rerd = 1;  /* Reorder */
			printf ("\n DO NOT USE +r FLAG, IT IS REDUNDANT!!\n\n");
		}
		else if (px == 'e') {
			opt.xcte = 1;  /* Expansion comp. */
			printf ("\n DO NOT USE +e FLAG.  SEE DOCUMENTATION!\n\n");
		}
		else if (px == 't') {
			opt.test = 1;  /* Test mode */
			printf ("\n PLEASE REPLACE +t FLAG WITH -t1 FLAG!!\n\n");
		}
#ifdef  CUTREPEAT
		else if (px == 'x') {
			opt.repc = 2;  /* repeat features */
			printf ("\n PLEASE REPLACE +x FLAG WITH -x FLAG!!\n\n");
		}
#endif
		else if (px == 'n') {		    /* Serial number */
		    printf ("\n\n PLEASE REPLACE +n.. FLAG WITH -n.. FLAG!!\n");
		    printf ("   USE OF FLAGS WITH + INTRODUCER WILL\n");
		    printf ("   NOT BE SUPPORTED IN THE FUTURE!!\n\n");
		    if (pp[1] == AN) k = 0;
		    else k = sscanf (++pp, "%d", &j);
		    if (k==1 && j>0) serno = j;
		    break;
		}
		else if (px == 'h') {
			dohelp = 1;  /* show help */
			printf ("\n USE -h RATHER THAN +h FOR HELP!!\n\n");
		}
	    }
	} else if (*pp == '?') {
		dohelp = 1;
	} else {	/* File name, do some work? no save the name... */

/* Put the argument at pp onto the list to be processed */
	    if (opt.dbug > 0) printf (" Adding %s to argument list.\n", pp);
	    head = pushtxt (pp, head);  /* Collect arguments */

//	    j = cutmask (pp, opt, &mmhead);
//	    if (j >= 0) n += j;	// count masks cut
//	    else r++;		// count rejected masks
////	    n++;

	}	// end of scan of argument token
//printf (" ** in main location 3, dohelp=%d\n", dohelp);	// debug only
	if (dohelp) break;

    }	/* Argument translation */
#endif  // on GETOPTUSE

/*  End of command line scan operations.  */

/* At this point, we have a list of arguments at head, and all the basic
  flag information has been stored where it will be used.  Now, do the work.  */

/* Initialization section and reading of maskcut.cfg is moved here past the
  command line scan, so that the debug flag may be used to debug the reading
  of the maskcut.cfg data when necessary.  */

/* The initialization steps must include reading data from the maskcut.cfg
  file, as that is where the metric mode allow is set, and that is needed
  when scanning the command line flags.  */

/* To allow metric mode, we may put a special flag in the file maskcut.cfg
  and get that data here.  If the data is not present or is set to 0
  the mmallow value stays zero and we don't use metric mode at all.  */

/* Get data if we do not yet have it, for help and debug output */
    if (CDhead == NULL) {
	CDhead = get_cut_data (CUTFILE, opt.dbug+dohelp);
// NOTE -- check for NULL return and exit if so!
	if (CDhead == NULL) {
	    printf (" *Could not open data file \"%s\".\n", CUTFILE);
	    printf (" (Should be in local directory or /usr/local/etc/maskgen/)\n");
	    return 0;
	}
#ifdef  USE_CDAT
	mmallow = CDhead[M_imacs].mma;
#endif
#ifdef  USE_GSTO
	mmallow = *(int*)GS_lookup (&CDhead, M_imacs, CD_mma, 0);
#endif
#ifdef  USE_BSTO
	mmallow = intvalue(getval(&CDhead,"IMACS","Mmode"));
#endif
// set default for mmallow
	inst = 0;	// better default.
	setdisp (0.0, 0.0, 0.0, 0);  /* Set default offset and angle */
    }


    dodebug = opt.dbug;   // For library programs

    if (opt.dbug > 1) {
#ifdef USE_CTE
	printf (" ** USE_CTE is defined.\n");
#else
	printf (" ** USE_CTE is NOT defined.\n");
#endif
	printf ("  xcte = %d, opt.xcte = %d\n", xcte, opt.xcte);
    }

#if GISMOFFSET || GISMONOTCH
    if (opt.offset != 0.0) printf (" -- GISMO offset %.2f used.\n", opt.offset);
#endif


/* If we want metric mode, we must pass the test; ask a riddle... */
/* Query and response section for metric mode users... */
    if (inchout == 0) {
	if (mmallow) {
char	buf[24];
char	ans[] = "Yes";
	    askme (buf, 20, " ? Metric output?  REALLY? ");
	    if (strcmp(buf, upcase(ans))) inchout = 1;
	    if (inchout == 0) 
	printf (" OK, then.  But that's DANGEROUS!!\n");
	    else printf (" Sorry, wrong.  You get inches.\n");
	} else {
	    inchout = 1;	/* Sorry, Charlie... */
	}
    }

/* Debug for scan items... */
    if (opt.dbug | opt.test) {
	printf (" Debug = %d, Test = %d\n", opt.dbug, opt.test);
	printf (" Curently n = %d\n", n);
    }

/*  Basic loop; process all the things in the list here.  */
    while (head) {	// List not NULL

	filename = pop_name (&head);
	if ( filename == NULL) continue;
	if (*filename == AN) continue;

	j = cutmask (filename, opt, &mmhead);
	if (j >= 0) n += j;	// count masks cut
	else r++;		// count rejected masks
//	n++;
	free (filename);
// NO, the below is not how we do it...
//	if (head) free (filename);
/* Leave the last filename in there; it might be the name of the special
  mask being cut in the next section... */
    }


/* Special masks have a null file name and are cut here; they can be
  ordered only with the proper flag and no other mask requests.  */

/* Implement the IMACS flange hole feature if requested */
    if (n == 0 && opt.rimh) n += cutmask (filename, opt, &mmhead);

#ifdef  TESTMODE
/* Allow test mode with no file specified. */
    if (n == 0 && opt.test) n = 1 + cutmask (filename, opt, &mmhead);
#endif

//    free (filename);	// may be redundant

//printf (" ** main dohelp=%d, gnch=%d\n", dohelp, opt.gnch);  // debug only
/*  Process the multi-mask files, if any (also gismo decker and special) */
//printf (" ** in main before cutmulti, dohelp=%d\n", dohelp);	// debug only
    j = cutmulti (&mmhead, opt);
    if (j >= 0) n += j;
    else  r += (-j);

/* Write what was done... */
    if (n > 0) printf ("\n Processed %d files.\n", n);
    else if (r > 0) printf ("\n Rejected %d files.\n", r);
    else  dohelp = 1;

    if (dohelp) {
//	printf ("\n %s -- Version %s%s", PROGNAME, VERSION, CTUVERS);
	printf ("\n %s -- Version %s", PROGNAME, XVERSION);
	printf ("\n  Parameters are: ");
	printf ("\n    SMDF file name(s)");
#ifdef  GETOPTUSE
	printf ("\n   Switches");
#else
	printf ("\n   Switches (preceede file names)");
#endif
//	printf ("\n     -d = Some debug information");
	printf ("\n     -d* = Set debug level to * [default 1]");
	printf ("\n     -q  = Omit debug information (level 0)");
//	printf ("\n     -i = Inch mode, ---i.nc output [default]");
if (mmallow) printf ("\n     -m  = Metric mode, ---m.nc output");
	printf ("\n     -f  = IMACS flange hole file");
#ifdef  GISMODECKER
	printf ("\n     -g  = GISMO decker mask file");
#endif
//	printf ("\n     +r = reorder nodes");
	printf ("\n     -r  = omit reordering of nodes");
	printf ("\n     -b* = base string (*) of multicut file");
	printf ("\n     -c* = Limit LDSS to *(1-7) submasks");
#ifdef  USE_CTE
	if (!xcte)	// Based on default value.
#ifdef  GETOPTUSE
	printf ("\n     -e  = Use thermal expansion correction");
#else
	printf ("\n     +e  = Use thermal expansion correction");
#endif
	else
	printf ("\n     -e  = Don't use thermal expansion correction");
// No use of -e if USE_CTE is not defined.
#endif
	printf ("\n     -n* = Set mask number to (*)");
#ifdef  GISMONOTCH
	printf ("\n     -k  = Gismo notch test.");
#endif
#if GISMONOTCH || GISMOFFSET
	printf ("\n     -y* = Gismo offset(*) mm.");
#endif
#ifdef  CUTREPEAT
	printf ("\n     -x  = double cut mask features");
#endif
	printf ("\n     -s  = cut at slow speed");
//#ifdef  TESTMODE
//	printf ("\n     -t = No test mode [default]");
	printf ("\n     -t* = Set test mode to (*)");
//#endif
//	printf ("\n     +d = More debug information");
//	printf ("\n     +t = Test mode ");
	printf ("\n     -h  = Display this help");
//	printf ("\n   When debug is on, limit to 10 data records.");
#ifdef  GETOPTUSE
	if (dodebug) printf ("\n   Using getopt command line scanning.");
#else
	if (dodebug) printf ("\n   Using discrete command line scanning.");
#endif
	printf ("\n \n");
//	return (1);
	k = 1;
// No longer skip return of storage, it has probably been obtained anyway.
//	goto main_finish;
//    } else {
//	printf ("\n %s -- Version %s\n", PROGNAME, VERSION);
    }

/* Return dynamic storage */
#ifdef  USE_CDAT
    CDhead = kill_cdat (CDhead);
#endif
#ifdef  USE_GSTO
    while (CDhead != NULL) CDhead = kill_gsto (CDhead);
#endif

#ifdef  USE_BSTO
    CDhead = btree_kill (CDhead, kill_kvm);
#endif

    k = r;	// was 0, but we return rejected mask count.
main_finish:
    opt.cp = kill_pvfs (opt.cp);
    opt.fp = kill_pvfs (opt.fp);
    return  k;		/* for lint */
}

/* Change History:

    2011/03-17 -- Version 3.36
	-- Replaced fclose with iclose for file name support.
	-- New version number to distinguish changes.

    2010/09-15 -- Version 3.35
	-- Moved reading of maskcut.cfg below the command line decode to
	allow debug.
	-- Fixed parsing in mgutils.c to properly decode Z offset in the
	dynamic parsing feature case.  Needed in GISMO case.
	-- Fixed count of mask features in cutgismo to include initialization
	outputs, allowing dummy mask to work properly.
	(2011/01-21) Initialize vert in gismodecker subroutine.
	--> Minimum version due to GISMO mask cutting changes

    2010/01-27 -- Version 3.34
	-- Add  GISMONW2  symbol to extensively modify the GISMO notch
	dimensions yet again.

    2010/01-15 -- Version 3.33
	-- Introduce  GISMONEW  symbol and new GISMO cutting section to
	provide the new notch configuration.
	-- Raise labels of GISMO mask slightly.

    2009/07-22 -- Version 3.32
	-- Add debug/test feature for -t7 which will allow operation on
	an otherwise rejected .SMF file.  To allow processing of an older
	.SMF or one created by the normalize utility.
	-- Add output of inline comments if <= 68 characters and has no
	parenthesis imbeded.

    2009/01-30 -- Version 3.31
	-- Avoid cutting slit for Gismo Long Slit which is in the .SMF
	file (for data reduction).  Done in maskcutsub.c by using a
	new feature of write/read smdf file, in which a mixed case
	keyword indicates the virtual flag.
	--> Becomes minimum version when GISMO is released.

    2009/01-08 -- Version 3.30
	-- Properly initialize file count "n" in main.

    2008/12-18 -- Version 3.29
	-- Changed to using getopt command line parsing.  The +n flag is
	no longer acceptable.
	--> Minimum version in release 2.08 due to -e, +n flag changes.

    2008/12-03 -- Version 3.28
	-- Fix double operation of reorder for IMACS masks by incrementing
	opt.rerd when doing reorder; except when doing it in multi-cut case.
	-- The -e flag only turns off opt.xcte rather than toggeling.

    2008/11-25 -- Version 3.27
	-- Add GETOPTUSE flag and code to use getopt command line
	scanning method.  Also, reorganize loops to have only one
	entry of configuration file reading and simplify the structure
	of how the arguments are used.
	-- Restructure order of operation, make a list of command line
	arguments and process it in a simple loop after all flags.  Do
	the help display last.
	-- Remove a lot of commented code.
	--> Actual minimum version of release 2.07

    2008/11-13 -- Version 3.26
	-- Add messages when +() flag detected to discourage use.
	--> Becomes minimum version at release 2.07

    2008/10-03 -- Version 3.25
	-- Redefine input of test mode parameter (currently unused)
	-- Change the +() flags to -() flags as much as possible
	to prepare for future use of "getopt" function.
	--> Becomes minimum version at release 2.06

    2008/04-28 -- Version 3.24
	-- Add checksum file security support.
	--> Becomes minimum version at release 2.05

    2008/03-06 -- Version 3.23
	-- Add test on REJOLDATA to optionally reject older data input
	(That test is not done, symbol is unused).
	-- Add Program (Vers) structure
	-- Reject obsolete version operation.
	--> This version will be minimum acceptable at release 2.04

    2008/02-06 -- Version 3.22
	-- Add support for WFCCD multi-cut masks
	-- Change format of Instrument: line to be compatable with the
	decoding in ncplot (ncrutil.c).
	-- Fixed bug if input file could not be read.
	--> This version will be minimum acceptable at release 2.03
	--> This version will be minimum acceptable at release 2.04

    2007/11-01 -- Version 3.21
	-- Add the +n feature to the help display

    2007/10-18 -- Version 3.20
	-- Support for version control system.
	-- New version text and XVERSION introduced.
	-- Set as Minimum Acceptable Version for release 2.00

    2007/10-03 -- Version 3.19
	-- Support for repeat count under feature CUTREPEAT

    2007-10-03 -- Version 3.18
	-- Version number actually updated when new output format
	for chole using repeat count introduced.  Next step will be
	to introduce actual repeat count in object_loop.

    2007-06-15 -- Version 3.18
	-- Change GISMO slide movement constant

    2007-03-20 -- Version 3.17
	-- Add corner chamfer in decker mask
	-- Add locator holes in decker mask
	-- Add offset feature for gismo masks
	-- Add special gismo notch only .nc file feature

    2007-03-19 -- Version 3.16
	-- Add optional serial number serno; remove adc character.

    2006-12-13 -- Version 3.15
	Correct cutting of GISMO decker mask

    2006-11-22 -- Version 3.14
	Move GISMO labels down slightly, and fix offset in config file.

    2006-11-15 -- Version 3.13
	GISMO support debugged, labeling GISMO decker mask

    2006-11-02 -- Version 3.12
	Move duplicated code into local subroutines

    2006-10-24 -- Version 3.10
	Add support for GISMO mask cutting

    2006-10-16 -- Version 3.09
	Add support for "area" feature type (non-published)

    2006-06-13 -- Version 3.08
	Add support for cdata table, with b-tree or gsto reading

    2006-05-26 -- Version 3.07
	Add support for b-tree dat storage subroutines.

    2006-05-18 -- Version 3.06
	Removed ncrutil.c into a separate .o file, refering to it
	only using ncrutil.h here.

    2006-05-17 -- Version 3.05
	Use ncrutil.c to read file for cutting time.
	Add time estimate comment to end of .nc file produced.

    2006-05-16 -- Version 3.04
	Add code to predict cutting time for file.

    2006-02-14 -- Version 3.03
	Now using improved reordering subroutine (in mgutils.c)
	And, pass lint complaints on version number writing.

    2005-12-15 -- Version 3.02
	Default to use CTE when date field in intgui is done.

    2005-12-14 -- Version 3.01
	Fix format error on version comment in output.

    2005-12-08 -- Version 3.00
	Better default of debug level and output tests.
	Removed the LDSS_NEWC symbol conditionals.
	Use of CTE now standard.

    2005-11-18 -- Version 2.98
	Support new method of configuration data storage
	Use new storage for CTE/Tcut values
	The configstore program is removed

    2005-11-11 -- Version 2.97
	Moving configuration reading to mgutils

    2005-10-10 -- Version 2.96
	General clean-up.
	Also, added cut util version string

    2005-09-20 -- Version 2.95
	Add the USE_CTE parameter to introduce cutting based on
	the temperature at observation.  Also add a flag to turn
	the computation on/off.

    2005-09-12 -- Version 2.94
	Allow input of base string for constructing file name of the
	multicut command file.

    2005-09-07 -- Version 2.93
	Modify cututil to turn autofocus on and hold off unconditionally
	when the autofocus on command is done.  Previously, the existing
	autofocus and hold status was used to only make changes.

    2005-07-13 -- Version 2.92
	Cleanup of commented code and unused things; no change
	to version as no changes to compiled code.

    2005-06-14 -- Version 2.92
	Removed "grating" and "instrument" elements from Obs
	structure, use order, Disperser, Tel_scope, Instru_mt
	elements instead.
	Name LDSS3 replaces LDSS:NEW in .cfg file.

    2005-05-20 -- Version 2.91
	New feature to interactively limit LDSS submask count
	Put reorder reports on a debug level count
	Put displacement output on debug level
	Add report of label information when each submask is cut.

    2005-03-28 -- Version 2.90
	Support of extended read_smdf changes call sequence to
	include element queue and a flag value.

    2005-03-28 -- Version 2.90
	Add LDSS_NEWC symbol defined parallel to LDSS:NEW, and the
	ability to undefine it as an experiment to deny the mask
	cutting feature for LDSS while allowing support for mask
	generation parts of the software.
	Mistakenly still labeled as 2.89

    2005-03-23 -- Version 2.89
	Added inclusion of mgfeats.h for global features operation

    2005-03-11 -- Version 2.88
	Testing new start location for slit queue...
	Changed x sign 3-15

    2005-03-09 -- Version 2.87
	Removed spurious warning of void maskcut.par;
	Making release version without LDSS support.

    2005-03-07 -- Version 2.86
	Fixed bug in setting of Z height when cutting multiple blanks
	for the ldss instrument.

    2005-03-02 -- Version 2.85
	Reducing LDSS multi cut to 5 masks to bypass laser machine
	alignment problem temporarily.

    2005-02-22 -- Version 2.84
	Increase diameter of alignment hole in LDSS by .02 mm.
	Change size of LDSS mask lettering to 2.6 mm for readability test.

    2004-12-17 -- Version 2.83
	Adding code for LDSS multi-mask making compatable with the
	current mask cutting.  Add parameter to cutmask call to make
	a queue of multi-mask Obs structures; add a program to cut
	multi-masks; add recognition of LDSS-2 instrument.  Put all
	this temporarily behind the LDSS:NEW symbol.

    2004-12-07 -- Version 2.82
	Adding compatable support for LDSS multi-mask making; add a
	call to "setdisp" to set null values for compatability.

    2004-08-16 -- Version 2.81
	Slight code change to avoid a warning message in the loop to
	find a unique time stamp.

    2004-08-13 -- Version 2.80
	Metric mode requires permission in config. file which is not
	normally present; as well as the command line flag which is
	not the default.  Then, it prompts for a special reply when
	given -- that should discourage its use.
	Also, some things cleaned up.

    2004-03-08 -- Version 2.75
	Add "From" comment giving input file name.  All the new engraving
	and comment things.

    2004-03-05 -- Version 2.72
	Non-released version, adding new output of timestamp to .nc file
	and in engraving on rim.  Add adc state in field and rim.

    2004-03-04 -- Version 2.70
	Extended characters moved to kdcutil.c
	Test of time-stamp put behind TESTMODE which is undef

    2004-03-03 -- Version 2.68
	Adding extend character function in preparation for formatting
	the time-stamp HMS in extended single characters...
	Also, test the time-stamp stuff in test mode.

    2004-01-15 -- Version 2.65
	Put program name and version into output file

    2004-01-09 -- Version 2.60
	Change default to reorder nodes

    2003-09-24 -- Version 2.50
	Add parameters to configuration file for long slit cutting

    2003-09-12 -- Version 2.40
	Add switch for slower cutting
	Revise crect to enable cutting of long slits

    2003-07-10 -- Version 2.30
	Inch output is now the default
	Use mode for exact stop on/off

    2003-06-30 -- Version 2.25
	For long motions which are split into pieces, temporarily
	disable exact stop mode for smoother movement.

    2003-06-28 -- Version 2.20
	Trying to modify pltimacs to add any needed Z motions to
	the character plotting code.

    2003-06-28 -- Version 2.10
	Add flange hole only file feature

    2003-06-28 -- Version 2.00
	Debug the inch mode output
	Add pause at autofocus on, and moving up from hold.
	Smaller labels in imacs mask
	Add double pass in flange holes
	Add faster engraving rate
	Test mode does not actually do anything
	Add imacs rim hole feature

    2003-03-25 -- Version 1.90
	Change "dodebug" to library global (new library requirement)
	Debug engraving pen up problem - link with charplt.o

    2003-02-26 -- Version 1.80
	Add line numbers at each major feature

    2002-12-17 -- Version 1.70
	Add count of objects; compute re-order number to be 2/3
	of total objects in reorder computation.

    2002-12-16 -- Version 1.60
	Add leading character to file name I=imacs, L=ldss

    2002-12-13 -- Version 1.55
	Added newline in character plot when setting incremental mode
	Changed format of some initial lines

    2002-12-06 -- Version 1.50
	Change directory search to new find file program

    2002-12-04 -- Version 1.45
	Using file finding subroutine to obtain "maskcut.cfg" from 
	several possible directories.

    2002-12-02 -- Version 1.40
	Add autofocus release, automatically
	Change sotufocus height logic at flange

    2002-11-27 -- Version 1.30
	Add "hold" option to autofocus control
	Change autofocus offset logic

    2002-11-07 -- Version 1.20
	Regularized formatting of x,y,z values with subroutines
	Changed circle plotting to cut much less radius at start/end
	Fixed the autofocus logic for the IMACS flange points

    2002-11-04 -- Version 1.10
	Change labels in ldss to lower edge from upper, to avoid
	the holding fixtures...
	Also change order of label plotting to cut time slightly.

    2002-10-31 -- Version 1.05
	Similar re-write to ldss-2 cutting program
	Removed commented code from older versions

    2002-10-30 -- Version 1.00
	Slit/feature cutting logic totally rewritten
	Add complex hole shapes for test mask construction

    2002-09-05 -- Version 0.70
	Some things moved to cututil.c
	Different instrument name decoding

    2002-09-03 -- Version 0.65
	Change clearance circles to fast feed rate

    2002-08-28 -- Version 0.62
	Fix avoidance code and ldss motions

    2002-08-28 -- Version 0.60
	Add configuration file
	Introduce new Z motion logic
	Start avoidance code also

    2002-08-26 -- Version 0.57
	Change order of alignment hole plotting for reduced slew

    2002-08-23 -- Version 0.55
	Allow re-ordering nodes to minimize travel

    2002-08-22 -- Version 0.50
	Add parameters from mail of 30 May
	Add engraving of file and astronomer names on LDSS-2 mask

    2002-08-15 -- Version 0.45
	Changed the order in which IMACS clearance holes are plotted

    2002-08-09 -- Version 0.42
	Debugging LDSS code to drop the IMACS Z-height code

    2002-06-24 -- Version 0.40
	Changing LDSS machining to laser system, the milling machine
	code is in a version called ldsscut.c now.

    2002-03-21 -- Version 0.32
	Metric output standard, the -i flag will put inch
	information in output, and append "i" to file name.
	Beam width changed to 0.003 inch or 75 microns; ~0.2 arcsec.
	More cleanup of old comments

    2002-03-20 -- Version 0.26
	Fixed bug in slit length computation
	Cleanup of some old comments
	Modified back of mask cutting direction
	-- Version 0.27
	Cleaned up circle plotting
	Changed extension to .nc from .MSK
	-- Version 0.28
	Added displacement of start/stop laser within holes
	Fixed small mode bug in hole drilling
	-- Version 0.30
	Added "newloc" subroutine to position to new feature location
	and do proper z movements to get there.
	Using this for all features, including labels and flange holes

    2002-03-19 -- Version 0.23
	Add \r\n for IMACS output line end stuff.  If it needs to be
	changed, use a #define constant for laser EOL.
	-- Version 0.25
	Numerous changes suggested by Bigelow for laser cutting...
	Better formats for output; eliminate some redundancy in
	commands for characters.

    2002-03-18 -- Veraion 0.22
	Invert sign of x for IMACS mask cutting, since it is being
	cut on the "back" of the focal plane.  See "back of mask".

    2002-03-14 -- Version 0.20
	Ability to cut circular reference holes added (for IMACS)
	Using bwid for beam width in mask and alignment holes

    2002-03-11 -- Version 0.17
	Cleaning up scale factor use, old comments

    2002-03-08 -- Version 0.16
	Drop obs structure after finishing conversion (use goto)
	Changed inch/metric scaling to be more clear

    2002-03-07 -- Version 0.15
	Add lasermode subroutine
	Add plotting of two labels

    2002-03-04 -- Version 0.10
	Remove debug restriction on slits plotted
	Add cutting of clearance and alignment holes for IMACS

    2002-02-06 -- Version 0.06
	Clean-up comments
	Slightly better structure for G-codes

    2002-01-31 -- Version 0.05
	Produces G-code file for LDSS2 masks

    2002-01-30 -- Version 0.02
	Name changed.  Not instrument specific.
	Introduced instrument name to system.
	Multiple instrument structure built.
	Ready for individual cutting programs to be inserted.
	Sample code for ldss2 cutting included

    2002-01-29 -- Version 0.01
	Initial version.

 */


