/*  mgutils.h == Mask Generation Utility Definitions */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*  Utility (library) subroutines for mask generation and related
  application programs.  */

#ifndef  INCLUDE_MGUTILS_H
#define  INCLUDE_MGUTILS_H

/* Take care of all standard system includes here */
#include        <stdlib.h>
#include        <stdio.h>
#include        <string.h>
#include	<math.h>
#include	<ctype.h>	/* For "toupper" */
#include	<time.h>	/* For clock - debug use */
#include	<unistd.h>	// for access, stat
#include	<sys/types.h>	// for stat
#include	<sys/stat.h>	// for stat

#include	<openssl/evp.h>
#include	<openssl/rsa.h>
#include	<openssl/pem.h>


#include	OPTUTILH	/* general optics structures */
/* optutils.h includes maskdat.h */
/* maskdat.h  includes ioutil.h, kdcutil.h */
/* We need ioutil.h, kdcutil.h, maskdat.h also */

// #ifdef  USE_BSTO
#ifdef  READ_BD
// #include	"../utils/btree.h"
#include	BTREEH
#endif
// Replace above by BTREEH and modify makefile appropriately!

/*  ===  Globals  ===  */
extern int  dodebug;	/* Defined in main program */
extern	FILE	*dbfile;

extern	int	Nfaroff;	/* Far off mask messages */
extern	int	Noutdet;	/* Outside detector messages */
extern	int	Nduppos;	/* Duplicate mask positions */
extern	int	Nconflx;	/* Conflicts reported */


/*  ====  Definitions  ====  */
#define  OBSLIST_ENTRY  257	/* Length of record */
#define  OBJECTLIST_ENTRY  255	/* Length of record */
#define  SMDF_ENTRY  255	/* Length of record */
#define  NEWLINES  "\n\r"

#define  MESCOUNT  6		/* Message count limit */


// #define  DefTemp  20.00390625
/* Default temperature is 20 + 1/256 degrees, an exact value */
// #define  DefTemp  11.99609375
#define  DefTemp  12.0390625
/* Better suggestion is 12 - 1/256 degrees, also exact */
/* Actual average is 12.28437 (weighted) or 12.03818 -- use nearest
  1/256 fraction to make it exact.  Use 12 + 73/256 or 12 + 10/256 
  i.e. 12.28515625 and 12.0390625 */
// MOVE THIS TO MASKDAT, then use it in maskcut to turn off the
// CTE use if it exactly matches.  Also, initialize in defobs
// and in the gui reading of the obs data.
// Best would be to use the maskcut flag to turn CTE on/off.


/*  ====  Some definitions to remove error messages...  ====  */
typedef  int  bool;
typedef  int  boolean;

/*  --NOTE-- We have to fix the above to real things!!! */


/*  ====  Typedefs  ====  */
/* namlist -- structure; members are:  next, last, name (char*) */
/* The above is an older version - see maskdat.h for new... */

/* Object list -- Master structure containing all data from the
  object list file.  */
typedef  struct  objectlist {
	char*	title;		/* Observation title */
// Only title (not needed) and comment and objq headers are active!
	namlist*  comments;	/* Comments from input file */
	objq*	head;		/* Per-object data listed */
	} objectlist;

typedef  struct  sort_tag {
	struct  sort_tag  *next, *last;
	void	*q;
//	objq	*q;
	}  sortq;

/*  -- Re-globalization of instrument index --  */

typedef	enum	Indx {
// Names for Instrument Index values (found by InstIndx)
	I_imacs,	// Any IMACS mask
// Experimentally switching these next two
	I_lds2,		// LDSS-2 masks
	I_cfg,		// Special centerfield mask
	I_lds3,		// LDSS-3 masks
	I_wfcd,		// WFCCD camera on 100"
	I_other
	} Indx;
#define  Instruments  5
// The Instruments should be same as I_other count.

typedef	enum	Mindx {
// Names for Mask Index values, found by MaskIndx
	M_imacs,	// Any IMACS mask
	M_lds2,		// LDSS-2 masks (obsolete)
	M_cfg,		// Special centerfield mask
	M_lds3,		// LDSS-3 masks, subfielded into imacs blank
	M_gismo,	// GISMO masks
	M_wfcd,		// WFCCD masks, subfielded on GISMO blank
	M_other
	} Mindx;
#define  Masks  M_other
// The Masks should be same as M_other count.

//typedef	enum	IXndx {
//// Names for Instrument extended index (found by ExtIndx)
//	IX_ldss,
//	IX_imx2,
//	IX_imx4
//	} IXndx;

typedef	enum	IXndx {
// Names for Instrument extended index, found by IxIndx
	IX_norm,	// Normal, no extended index
	IX_imx2,	// IMACS f/2
	IX_imx4,	// IMACS f/4
	IX_other
	} IXndx;

typedef	enum	Dindx {
// Names for various disperser types
	D_grat,		// grating
	D_gris,		// grism
	D_ech,		// echelle
	D_dirx,		// direct, no disperser
	D_other
	} Dindx;

typedef	enum	Findx {
// Names for special features, found by FeatIndx
	F_normal,	// No special feature, normal mask
	F_direct,	// A direct instrument, no disperser
	F_moe,		// The MOE feature is present
	F_gisap,	// Gismo aperture mask feature
	F_other
	} Findx;

/* General storage structure; indx is application dependent index,
 type is a GS_types value, and data is a pointer to the data. */
//#ifdef  USE_GSTO
#ifdef  READ_GD
typedef  struct  gsto {
	struct	gsto	*next, *last;
	int	indx;
	int	type;
	void	*data;
	} gsto;
#endif

/* General storage types, indicating the data; gsto is pointer to
  another gsto structure, int to a single int, double to a double,
  and strx to a general linked structure with no dynamic storage.
  -- This enum is extensible, when the supporting code is also
  extended to support all types. -- */
//#ifdef  USE_GSTO
#ifdef  READ_GD
typedef	enum	GS_types {
	GS_gsto,	// Secondary queue
	GS_int,		// Integer value
	GS_double,	// Double value
	GS_vect2,	// Vector
	GS_strx		// Queue pointer
	}GS_types;
#endif

/* General storage for cut data; the enum subscripts the data
  which is stored in the base level gsto structure.  A function
  returns the design type given the index here. */
#ifdef  READ_GD
typedef	enum	CD_indx {
	CD_refr,	// any reference values
	CD_curve,	// curvature in diopters
	CD_beam,	// beam diameter in mm.
	CD_slew,	// slew rate mm/min
	CD_cut,		// cut rate mm/min
	CD_fine,	// fine motion mm/min
	CD_afoc,	// autofocus range, mm
	CD_dx,		// offset in x, mm
	CD_dy,		// offset in y, mm
	CD_dz,		// offset in z, mm
	CD_flange,	// flange radius, mm (IMACS masks)
	CD_aangle,	// adjustment angle, degrees (IMACS)
	CD_aclear,	// alignment clearance, mm (IMACS)
	CD_LongSlit,	// longest unsupported slit, mm
	CD_SlitGap,	// gap inserted, mm
	CD_tool,	// tool clearance, mm
	CD_cte,		// thermal expansion
	CD_tcut,	// Cutting temperature, Celsius
	CD_dmx,		// Maximum cut distance, mm
	CD_av,		// avoidance queue, pointer
	CD_mma,		// millimeter mode allowed flag, int
	CD_Acon		// Acceleration Constant
	}CD_indx;
#endif

//  #ifdef  USE_BSTO
#ifdef  READ_BD
typedef	struct	keyval {
	char	*key;
	void	*val;
	} keyval;

typedef	struct	listval {
	namlist	*text;
	void	*queue;
	} listval;
#endif

#ifdef  USE_CDAT
cdata*	kill_cdat (cdata* head);
#endif

/*  ====  Default Value coordination  ====  */
// #define  D_refsiz    5.803  // Old size, 2.0 mm
#define  D_refsiz    5.0
#define  D_minsep    2.50
#define  D_maxradi 309.44
// maxradi is 15 arc minutes with nominal focus
#define  D_maxradl  88.0

#define  D_plotradi  313.2
#define  D_plotradl   93.3
/* Added plotting radius for masks 2007/2-21 KDC */

/* Rotator offset angles here */
#define CHUECO (-46.150)
#define LDSSROA (-60.3)

#ifdef  WFCCD
const bbox WFC_mask = { {-75.0, 75.0}, {-40.0, 40.0} };
// NOTE -- that's just a guess; find out the real thing!
// If we can't have a const here, define the dimensions and
// use them to initialize a bbox where we can...
#endif

/*  ====  Security Functions Globals  ====  */

//const unsigned char  HLIMIT = '\x0';
#define  HLIMIT  '\x0'

#define  BS  320
/* Buffer Size of all binary and text buffers used here, should be able
  to accomodate at least (number of key bits) / 2 bytes, although the
  longest line will be about (key bits) / 4 + 64.    */

/* Definitions of status line check bits.  For "decode_last"  */

#define  FS_exists  (1 << 0)	// Status line exists
#define  FS_cksum   (1 << 1)	// Checksum exists
#define  FS_cksok   (1 << 2)	// Checksum matches data
#define  FS_sig     (1 << 3)	// Signature exists

#define  FS_sigok   (1 << 4)	// Signature is valid
#define  FS_cks6    (1 << 5)	// Checksum is sixbit encoded
#define  FS_sig6    (1 << 6)	// Signature is sixbit encoded
#define  FS_fmtok   (1 << 7)	// Status line format is good

#define  FS_verx    (1 << 8)	// Status line version exists
#define  FS_verok   (1 << 9)	// Status line version is in range
#define  FS_vercur  (1 << 10)	// Status line version is current



/*  ====  Function Prototypes  ====  */

/*  ===  Special %#%#% genlist prototypes  ===  */

void*	G_pop_c (void* x);
void*	G_pop_l (void* x);
void*	G_pop (void* x, int c);

void*	G_head (void* a);
void*	G_tail (void* a);
void	G_circ (void* a);

void*	G_push_l (void* a, void* b);
void*	G_push_cl (void* y, void* x);
void*	G_push_c (void* y, void* x);

int	listcount (void* x);

/*  ===  Interval and Bounding Box support  ===  */

intval	interval (double a, double b);
int	inintv (double a, intval v);
int	ovr_lap (intval a, intval b);
int	bbover (bbox a, bbox b);
bbox	boundbox (intval x, intval y);
bbox	boundbv2 (vect2 a, vect2 b);
intval	intex (intval v, double p);
bbox	boundex (bbox b, vect2 p);
bbox	boundor (bbox a, bbox b);
bbox	bbedge (sp_edge e);
//bbox	bbspec (spect* s);
bbox	bbslit (vect2 p, slit t);
double	intlen (intval x);

/*  ===  Maskgen specific I/O  ===  */
void	data_file (char** fspec, char* filename, char* ev, int mf);

/*  ===  Subroutine to read optical data  ===  */
element*  get_optic_data (char* filename, int dl);

/*  ===  Globalization  ===  */
Indx	InstIndx (char* name);
char*	ObInstN (Obs* obs);
//#ifdef  GISCAM
char*	ObFrontN (Obs* obs);
//#endif
char*	InstName (Indx i);
#ifdef  READ_BD
char*	InstCase (Indx i);
#endif

char*	ObDisN (Obs* obs);
Dindx	DispIndx (Obs* obs);
char*	DispName (Dindx i);

Mindx	MaskNamIndx (char* name);
Mindx	MaskIndx (Obs* obs);
char*	MaskName (Mindx i);
#ifdef  READ_BD
char*	MaskCase (Mindx i);
#endif

Indx	InstNdx (Obs* obs);
Findx	FeatIndx (Obs* obs);
char*	FeatName (Findx i);

/*  ===  Object Queue manipulation utilities  ===  */
objq    *pop_obj (objq*);
spect	*kill_spect (spect*);
objval	*kill_objval (objval*);
objq	*kill_obj (objq*);
objq	*kill_objq (objq*);
objq	*push_obj (objq*, objq*);

/*  ===  Other dynamic storage  ===  */
void	kill_ptr (void**);
void	clean_obs (Obs*);
Obs*	kill_obs (Obs*);

/*  ===  Decoding data  ===  */
/*  ===  Encoding data  ===  */
// Those routines have been moved to ioutils.c

/*  --  Debug temporary  --  */
void	dbroq (char* title, Obs* ob);

/*  ====  Generalized Storage Support  ====  */
#ifdef  USE_GSTO
//gsto	*new_gsto (void);
gsto	*kill_gsto (gsto* a);
//int	get_CDtype (int item);
void*	GS_lookup (gsto* *head, int index, int item, int flag);
#endif

#ifdef  READ_BD
#ifdef  USE_BSTO
//int	kvcomp (void* a, void* b);
int	keycomp (char* a, keyval* b);
//keyval*	newkv (char* key, void* val);
//void	kventer (BNODE** root, char* Case, char* key, char* val);
#endif
int	kill_kvm (keyval* d);
char*	getval (BNODE* *root, char* Case, char* key);
void*	getlsq (BNODE* *root, char* Case, char* key);

#endif  // on READ_BD


/* === Cutting parameters and avoidance support === */
#ifdef  USE_CDAT
cdata*	get_cut_data (char* filename, int dl);
#endif
#ifdef  USE_GSTO
gsto*	get_cut_data (char* filename, int dl);
#endif
#ifdef  USE_BSTO
BNODE*	get_cut_data (char* filename, int dl);
#endif

/* -- Geometry of slit:  Find largest radius to slit edge */
int	Xcircle (objq* q, vect2 p, double r);
int	avoidanc (objq* ob, avoids* av, double tool);
// double	Rmax (objq*  p);

/* -- Grating angle computation */
// all obsolete or local (static) in mgutils

/*  ---  Optical utility (local to mgutils)  --  */
// double	foclen (Obs* ob);

/*  ====  Checking the object list warning flags  ====  */
#ifdef  OBJCHECK
int	object_check (objq* ob);
#endif

/*  =====  Filling in some structures  =====  */
objq*  get_object (char* data, int type, namlist* preq, Obs* ob);
//grating*	get_grating (char* name);
// detector*	get_detector (element* inst, element* gs, int dl);
// Above 3 routines may be local to mgutils and should not appear here.

/*  ===  Object queue manipulation  ===  */
// objq*	fill_object (char*, int, namlist*);
// Above is well and truly obsolete and is removed.

/*  ===  Reading files  ===  */
Obs*	defobs (element*);
void	get_mpos (Obs*);
void	normalize_obs (Obs* a);
#ifdef  USE_CDAT
void	check_temp (Obs* obs, cdata* Chead);
#endif
#ifdef  USE_GSTO
void	check_temp (Obs* obs, gsto* Chead);
#endif
#ifdef  USE_BSTO
void	check_temp (Obs* obs, BNODE* Chead);
#endif

Obs*	read_obsfile (FILE*, element*);
objq*	read_objectlist (FILE*, Obs*);
objq*	read_objectqueue (Obs* ob);
void	fill_objects (Obs* ob);

/*  ===  Fill in object data from optics data  ===  */

/*  ===  Conflict geometry  ===  */
int	within (double, double, double);
bool	overlap (double, double, double, double);
#ifdef  DIFREF
vect2	adifref (vect2 fp, vect2 zp);
//static	vect3	drefractn (vect3 pos, vect2 zp, double rc);
//static	vect3	urefractn (vect3 pos, vect2 zp, double rc);
#endif  // on DIFREF
/*  ===  Transform Geometry  ===  */
vect2	maskvect (Obs*, double, double);
void    unmaskvect (Obs*, vect2, double*, double*);
void	pbnormal (passband*);
vect2	det_pos (Obs*, double, vect3, int);
/*  ===  Spectrum Geometry  ===  */
sp_edge	getedge (Obs*, objq*, vect2, double, double, int, passband);
vect2	lslit (slit);
vect2	rslit (slit);
vect2	wslit (slit);
spect*	getspect (Obs*, objq*, int, passband);

/*  ===  Multi-Object Echelle support  ===  */
// static	double	moecw (int n);
// static	int	moeorder (double  wavl);
// static	passband	moepb (int n);
// (above declared static)
spect*	moespect (Obs *obs, objq *ob);

// spect*	killspect (spect*);	// see kill_spect

/*  ===  Conflict geometry  (continued)  ===  */
// int	spect_edge_intersect (vect2, vect2, double, sp_edge, double, int);
/* Add debug int to end of all these intersection programs... */
/* boolean	spect_intersect (vect2, vect2, vect2, spect*); */
// boolean	spect_intersect (vect2, vect2, vect2, spect*, int); /* debug only */
/* boolean	spect_overlap (spect*, spect*); */
// boolean	spect_overlap (spect*, spect*, int);  /* debug only */
// (above declared static)

/*  ===  Fill in object data from optics data  ===  */
// int	fillobject (objq*, Obs*);
// used only locally, fillobject is now static...
int	object_setup (Obs*);

/*  ====  Generating virtual objects for gaps  ====  */
#ifdef  GAPVO
// void	Gap_count (int, objq*);
// static	objq*	gap_queue (Obs* ob);
void	add_gaps (Obs* ob);
#endif

/*  ====  Sorting the object queue  ====  */

int	sort_decide (objq*, objq*);
int	sort_check (objq*, int(decision)(objq*,objq*)); // debug routine
objq*	sort_merge (objq*, objq*, int(decision)(objq*,objq*));
/* Need to typedef sortq too */
sortq*	push_sq (sortq*, sortq*);
sortq*	pop_sq (sortq*);
sortq*	kill_sq (sortq*);
objq*	sort_obj (objq*, int(decision)(objq*,objq*));


/*  ===  Conflict queue management  === */
cfl*	push_cfl (cfl*, cfl*);
cfl*	pop_cfl (cfl*);
cfl*	spop_cfl (cfl*);
cfl*	find_cfl (objq*);

/*  ===  Add and drop conflicts  ===  */
cfl*	add_conflict (objq*, objq*, int, bool);
cfl*	drop_conflict (cfl*, bool);

/*  ===  Find conflicts  ===  */
int	find_conflict (objq*, objq*, Obs*);
int	seek_conflict (objq*, bool, Obs*);
int	scan_conflicts (Obs*);

/* -- Debug only routine for maskgen -- */
void	check_cfcount (objq* ob, char* title);

/*  ===  Manage conflicts - drop, count, resolve  ===  */
double	hcpriority (objq*, double);  // use only by conflictedest
int	drop_single_confs (Obs*);
//int	count_conobj (objq*);	// no outside use
int	repconf (Obs*);

/*  ===  Object queue status -- set active/inactive  === */
int	set_active (objq*, Obs*);
int	set_inactive (objq*);
objq*	max_conflict (objq** loq, objq** hiq);
int	order_decide (objq* a, objq* b);
int	priority_decide (objq *a, objq *b);
int	deconflict (Obs *obs);

objq*	inact_obj (Obs *obs);

/*  ===  Writing SMDF file from object queue  ===  */
#ifdef  GSDATA
void	gpout (gpos g, FILE* f1, FILE* f2, int wl);
#endif
void	write_smdf (Obs* obs);
void	write_obf (Obs* obs, int fx);

/*  ===  Object queue utilities  ===  */
objq*	find_object (objq* a, char* name);
int	active_objs (objq*);
int	RefCount (objq* b);
void	repobj (objq* ob);
#ifdef  NSROT
void	rotobj (objq* ob, double angl, int sr);
#endif

/*  ===  Object queue ordering  ===  */
double	smdist (objq*, objq*);
double	totdist (objq*);
/* static void revert (objq* ob, int n);  */
objq*	reorder (objq*, int, int);

/*  =====  SMDF Reading subroutines  =====  */
/* static namlist*  kill_litlist (namlist*  n); */
Obs*	read_smdf (char* filename, element *gs, int special);

/*  ====  Wing Chip Support programs  ====  */
objq*	wing_holes (double x, double y, double d, int f);
objq*	wing_set (Obs* obs);
void	add_wings (Obs* obs);

/*  ====  Slit Extension subroutines  ====  */
sleg	*head_sleg (sleg* a);
sleg	*tail_sleg (sleg* a);
//sleg	*push_sleg (sleg*, sleg*);
//sleg	*pop_sleg (sleg*);
sleg	*kill_sleg (sleg*);
//void	augment_side (Obs* obs, objq* ob, int e, double a);
//objq	*conflict_x (Obs* ob, objq* b, int e, double a, objq* f);
// Above routines declared static
sleg	*scan_free (Obs* ob);
sleg	*scan_sle (sleg* a, Obs* obs);
sleg	*sort_sle (sleg* a);
sleg	*fix_sle (sleg* a, Obs* obs);
void	extend_slits (Obs* ob);

/*  ====  Generalized File Caching  ====  */
char*  locread (char* filename);

/*  ====  Key File Utilities  ====  */
char*	pubfs (char* rl);
RSA*	get_ftpkey (char* rl);

/*  ====  File Signature Utilities  ====  */
long  int	llof (char* buf, int bufl, FILE* f);
int	decode_last (char* ll, char* x, char* v, 
	unsigned char* hash, int* lh, unsigned char* sig, int* ls);

EVP_MD_CTX*  cxfile (char* buf, int bufl, FILE* f);
int	verifile (FILE* f, RSA* key, int dlvl, char** fver);
int	verifide (FILE* f, int dlvl);
int	signfile (FILE* f, char c, int dlvl);

/*  ===  System Dependent file name return program  ===  */
//char*	fhname (FILE* f);

/*  ===  System dependent part of version support  ===  */
float	vschecks (Vers* q);
Pvfs*	makevfs (void);

// Pvfs*	getlocv (char* cfn, int* ls);
// Pvfs*	getcurver (char* url, char* file);
Pvfs*	getcurver (char* file);
//int	check_ver (namlist* comps);
int	check_ver (namlist* comps, Pvfs** cpp, Pvfs** fpp);

/*  ===  Test of new parsing and file finding  ===  */

#ifdef  LOCATE
void	xcat (char** a, char* b);
char*	locfile	(char* mode, char* names, char* sufx, char* dirs);
#endif  // on LOCATE

/*  ====  */
#endif     /*  INCLUDE_MGUTILS_H  */
/*  ====  */

