/*  NCPLOT == plot the moves in a .nc file */
/* Contains draw_window for motion list; pwin for the nc plot window,
  and a main program to obtain and decode the nc file. */
/* Includes the decoding from ncputil.c, and other plotting subroutines
  from cwplot.c  */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                  Copyright (C) 2011 by                            **
**     Ken Clardy,  Carnegie Observatories,  Pasadena California.    **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// General things - standard includes, global data
/* Standard includes */
#include	<stdio.h>
#include	<stdlib.h>
// #define		_GNU_SOURCE	// for strcasestr in ncputil.c
#include	<string.h>
#include	<unistd.h>	/* For getopt */


/* ioutils.h has the definition of namlist, and some subroutines */
// #define   IOUTILH  "../utils/ioutils.h"
// #include        IOUTILH
// included in maskdat.h

// #include	KDCUTILH
// The kdcutil.h is needed by utilities
// included in maskdat.h

//#include	MASKDATH
//// need maskdat for the avoids structure

#include	MGUTILH
// mgutils.h includes maskdat.h and needed subroutines, and features

// LOOPDEBUG specifies the output of event information to find any
// problems with the loop
//#define  LOOPDEBUG
#undef  LOOPDEBUG


/*  ====  GLOBAL DEFINITIONS  ====  */

const   char    PROGNAME[] = "Plot of NC file";
// const   char    VERSION[]  =  "1.15";
const	char	VERSION[] = Xstr(NCPLOTV);
//int     dodebug = 1;    /* Global debug flag */
// dodebug is defined in optutils, included through mgutils...
extern int dodebug;


#ifdef  sparc
FILE*	stdbugf = stderr;
#else
FILE*	stdbugf;
#endif

avoids	*iav = NULL;
avoids	*xav = NULL;
avoids	*pav = NULL;

#define	I_title	"NC Plot"

/* Global Data and structures here */

/* For non-solaris, rather than including kdcutil.h... */
#ifndef __sun
#define  nint(a)  (int)round(a)
#endif

// #include	"ccbxlib.c"	/* Just put whole thing in... */
#include	"cxt.h"		/* From library directory; new plotting */

#include	"utils.h"	/* No longer included by ccb lib */

#include	"intgui.h"	/* Definition of interface */


// Standard Version structure
const	Vers	Program = { NULL, NULL,  NCPLOTV, NCPLOTR, NCPLOTT };
#define  APPLICATION_NAME  "Plot of NC file"
//const	char	PROGNAME[] = NCPLOTT;
// mgutils expect PROGNAME and VERSION strings to exist as global labels


// #include	"cwplot.h"	/* Plot tools */
// this was only used here, we put those in directly...
/*  ===  Globals  ===  */
// extern int  dodebug;	/* Defined in main program */
enum	directions {
	UP,
	DOWN,
	LEFT,
	RIGHT };
#define  DOTTED  0
// End of things from cwplot.h included directly.
// directions and others are in dbplot.h

#include	<math.h>

// Above previously included from tcwplot.c...

#include	<sys/time.h>

static struct timeval	previous;
static struct timeval	current;
static	int	bpressed=0;

/*   ----  Graphics Globals  ----  */

static Application  *app;
static MainWindow  mwin;

static Button	dunbut;
static Button	zibut;
static Button	zobut;
static Button	rzbut;

/* And, a list of windows... */
wble*	winlist = NULL;

//  --button location--

#define  kby  3
#define  kbh  7

/* Button blocks for button widgets */
/* butblock:  xloc,yloc, width,height, label, *data,
	*button, *mwin, aligntype, laboff, callback  */

butblock	dbfield = {
	150, kby,  2, kbh,  "Done", NULL,
	&dunbut, &mwin, Rightside, 0, NULL };

butblock	zifield = {
	10,  kby, 1, kbh, "In", NULL,
	&zibut, &mwin, Leftside, 0, NULL };
// An approximation to a button block, needs lotsa work...

butblock	zofield = {
	30,  kby, 1, kbh, "Out", NULL,
	&zobut, &mwin, Leftside, 0, NULL };

butblock	rzfield = {
	100,  kby, 2, kbh, "ReZero", NULL,
	&rzbut, &mwin, Leftside, 0, NULL };
// callback field in buttons above was ToggleButton, has been
// removed as we are using manual rather than auto buttons here.


/* The list of buttons */
bble*	butlist = NULL;



#include  "ncputil.c"
// Defines the mlist pointer, and utilities to make list

// calls for plotting will use this mlist pointer, so this needs
// to be defined before the plot stuff...

/* -- Global things needed for plotting the window  -- */
/* These things were in cwplot.c which was therefore an application
 specific plotting program.  They are removed to here, so common code
 can be used between this app and dbplot.c routines.  */

// static Display  *display;
static Window	win;
static Button	dunbut;
// static Menu	vmenu;	// unused
static Button	zibut;
static Button	zobut;
static Button	rzbut;


//GC	gc;
//GC	lgc;
Bool	done=False;
char	locbuf[64];

/* We have the following properties for each view/window
	> Scaling parameters BX, BY, ZX, ZY
	> Color of background
  These are indexed by MASK, DETECTOR and kept in a
  structure.  */

typedef	struct  vprops {
	double	BX, BY, ZX, ZY;
	u_long	color;
	} viewprop;

// Set initial size for ncplot window here
static	int	wx = 600;
static	int	wy = 630;
static	int	cwx, cwy;

static	viewprop  V;

// there is only one view, not vp[2] views here...


#define	bw	60
#define	bh	20

//#define	kw	36
//#define	kh	18
#define	kw	3
#define	kh	8

#define zw	18
#define zh	18
#define zd	4


/* --- Include the common plotting programs here --- */

// #include  "complot.c"
#include  "cwplot.c"

// #include  "cwplot.c"
// Includes all the common plot stuff we can re-use from smplot,
// twplot and dbplot routines.

//static	Bool	iconized=False;
// not defined until tcwplot is included...

/* Other plot stuff here is specific to the job of plotting the
  nc motion file onto a window */



	/* - - - - - - - *\
	|                 |
	|   draw_window   |
	|                 |
	\* - - - - - - - */

void	draw_window (mlist* top, int df, opts* opt)
/* Actually do the plotting */
{
mlist	*p;
double	x, y;
double	px, py;
u_long	vclr;
vect2	cen;
vect2	rad;
vect2	pv, nv;
pvec2	pp, pn;
double	a1, a2;
avoids	*atop;
avoids	*av;
//static int  kk;  //debug only
GC	lgc;

    if (df == 0) return;

// show dflag identification
//    k = df/4;		// debug only
// 1 - initial; 2 = button; 4 = ? ; 8 = expose; 16 = configure.
//    fprintf (stderr, " DW-%d ", k);	// debug only
//    fflush (stderr);			// debug
// Remove debug print temporarily

/* We will plot all the motion vectors into the window.  Also, we
  may plot background fields. */

/* A loop scanning the mlist will plot lines of the appropriate
  color for each motion vector.  Some motions may be circular.
  Also, we may need to project the actual coordinates given in
  the list to a local viewplane using two projection vectors. 
  These will be the viewport x axis unit vector and the viewport
  y axis unit vector.  X axis unit vector is 1,0,0 rotated by the
  (view_azimuth - 180) about the z axis.  it is always in the
  original xy plane.  The Y axis unit vector is the 0,1,0 vector
  rotated first by (view_azimuth - 180) about z axis, and then
  by (view_altitude - 90) about the x axis.
*/

/* Put fiducal marks at center on edges for zoom reference */
    lgc = mwin.gfs.gc;
    XSetForeground (mwin.disp, lgc, app->black);
    XDrawLine (mwin.disp, win, lgc,  0, cwy, 20,    cwy);
    XDrawLine (mwin.disp, win, lgc, wx, cwy, wx-20, cwy);
    XDrawLine (mwin.disp, win, lgc, cwx, 50, cwx,    70);
    XDrawLine (mwin.disp, win, lgc, cwx, wy, cwx, wy-20);

/* We may want to put some fiducial marks on the plot? 
  Maybe a cross at the origin?  Tooling locations? */

    cross (app->black, 0.0, 0.0, 8, 0);	/* black cross at origin */

/*  Loop to scan mlist and plot all vectors, properly scaled... */
    px = py = UNSET;
    C_loop (p, top) {

	if (p->interp < 0) continue;	// Omit special blocks
// Keep location of current entry for start of next vector. 
// Thus, first vector is not plotted yet.
	x = p->v.x;
	y = p->v.y;

// Decide color of plotted vector based on status values in entry
	if (p->laser) {		// cutting
	    if (p->par > 3) vclr = app->brown;	// roughcut
	    else if (p->par == 3) vclr = app->green;	// engraving
	    else  vclr = (p->feed < 520.0) ? app->black : app->red;  // cutting
	} else {		// not cutting
	    vclr = (p->feed > 600.0) ? app->grey : app->yellow;
	}
// NOTE THAT FEED RATE SELECTION DEPENDS ON MM UNITS; SCALE ALL
// THE DIMENSIONS IN THE UTILITY WHICH SCANS G-CODE

// Plot the vector properly scaled, if possible
	if ( !( px < VSET || py < VSET || x < VSET || y < VSET) ) {
// Decide to plot vect or arc based on interp
	    if (p->interp == 1) {		// Line
		vect (vclr, px, py, x, y);
	    } else if (p->interp == 2 || p->interp == 3) {  // Circular arc
// This is tricky.  We define start and end based on cw/ccw and
// on start, end points vs. center.
// Get the center and previous/next locations:
		cen = get2de3v (p->ctr);
		pv = make2vect (px, py);
		nv = get2de3v (p->v);
// Convert pv/nv to be relative to the center:
		pv = sub2vect (pv, cen);
		nv = sub2vect (nv, cen);
// Obtain polar versions of pv, nv:
		pp = polar2 (pv);
		pn = polar2 (nv);
// Angles are those of these vectors..
		a1 = pp.t / Degree;
		a2 = pn.t / Degree - a1;	// ccw distance to end
		rad.x = rad.y = pn.r;
// later all this changes with look angle
// Fix up angles for both range and sign...
		if (a1 < 0.0) a1 += 360.0;
		if (p->interp == 2) {	// go cw to a2
		    if (a2 > 0.0) a2 -= 360.0;
		} else {		// go ccw
		    if (a2 < 0.0) a2 += 360.0;
		}

		if (fabs(a2) < 0.001) a2 = 360.0;	// full circle

		sarc (vclr, cen, rad, a1, a2);

	    }
	}

// Allow for plot of circular arc based on interp value

// Update coordinates for next loop
	px = x;
	py = y;

    }

/* If IMACS, plot the mask flange in blue */
    if (IMACS) {
	cen = make2vect (0.0, 0.0);
	scircle (app->blue, cen, 333.0, 0);
	scircle (app->blue, cen, 313.0, 0);
    }
// WHEN PUTTING IN THE VIEW ANGLES, CHANGE THIS TO THE SCALED ARC
// SUBROUTINE, AND COMPUTE USING THE ACTUAL Z-HEIGHT OF THE FLANGE,
// WHICH IS ABOUT -12.7 MM PROBABLY; OR USE 0.0 FOR Z.

/* If not IMACS and pav is not null, we plot the rectangular
  bounds of the ldss mask, dimensions hard-wired here... */
//    if (LDSS && pav != NULL) {
/* Always plot LDSS mask bound in blue */
    if (LDSS) {
	rectangle (app->blue, -76.2, 76.2, -57.15, 57.15);
    }

/* If plotting a gismo mask, we plot the outer rectangle here.  */
/* Or, if a wfccd multi-mask, show the mask blank size */
    if (MaskI == M_gismo || MaskI == M_wfcd) {
	rectangle (app->blue, -175.0, 175.0,  -163.25, 163.25);
    }

/* Plot all the current avoidances, in color below */
//    if      (IMACS & 1) atop = iav;
//    else if (IMACS & 2) atop = pav;	// Set optionally from flag
// Was if IMACS atop = iav always...  (restored)
    if      (IMACS) atop = iav;
    else if (LDSS)  atop = pav;   // Set optionally based on flag
    else            atop = NULL;
    C_loop (av, atop) {
/* plot circle at x,y radius r+rtool */
	cen = make2vect (av->x, av->y);
	scircle (app->orange, cen, av->r + rtool, 0);
    }


/* --- Scale bar plotted in corner --- */
    if (V.BX != 0.0) {
double	x, b, d, s, w;
int	i, p;
int	mx, my, mt, c;
static double	sv[6] = { 10.0, 5.0, 3.0, 2.0, 1.0, 0.5 };
char	label[16];
GC	lgc;
	b = V.BX;
	if (opt->inch) b *= Inch;
	if (b < 0.0) b = -b;	/* The scale value mm or inch */
	w = (double)wx / 3.0;	/* Maximum scale bar size, units */
	x = w / b;		/* Maximal scaled value */
	d = pow (10.0, floor (log10 (x)));
	for (i=0; i<6; i++) {
	    s = sv[i] * d;
	    if (s <= x) break;
	}
/* s is now size of the scale bar (millimeters or inches) */
	if (opt->inch) {	// inch
	    if (s >= 0.01 && s < 100.0) {
		sprintf (label, "%4.2f inch.%c", s, AN);
	    } else if (s > 99.0) {
		sprintf (label, "%8.1e feet.%c", s/12.0, AN);
	    } else {
		sprintf (label, "%8.1e m.%c", s*Inch/1000.0, AN);
	    }
	} else {		// metric
	    if (s >= 0.01 && s < 1000.0) {
		sprintf (label, "%4.2f mm.%c", s, AN);
	    } else {
		sprintf (label, "%8.1e m.%c", s/1000.0, AN);
	    }
	}
	if (opt->inch) {
	    p = (int)(b * s + 0.5);
	} else {
	    p = (int)(b * s + 0.5);
	}
/* p is length of the scale bar in pixels */
	mx = wx - 10 - p;	// was 10;
	my = wy - 20;
	mt = 3;
	c = strlen (label);

	lgc = gcsetup (mwin.gfs.gc, app->black);
	XDrawLine(mwin.disp,win,lgc,mx,my-mt,mx,my+mt);      // draw scale
	XDrawLine(mwin.disp,win,lgc,mx+p,my-mt,mx+p,my+mt);
	XDrawLine(mwin.disp,win,lgc,mx,my,mx+p,my);
	XDrawString(mwin.disp,win,lgc, mx + (p/2) - (c*4), my+16, label, c);
    }	/* End of scale bar section */

/* Put current center location in upper right corner */
    {
    int	mx, my;
    int	c;
    GC	tgc;
    char	cbuf[64];
//	tgc = CBX_CreateGC (app->black, app->white, "courier", "medium", 12);
/* tgc is used to write display positions */
	tgc = mwin.gfs.gc;
	XSetForeground (mwin.disp, tgc, app->black);

//	sprintf (cbuf, "%11.4g %11.4g", V.ZX, V.ZY);	// was %10.3g

	if (opt->inch) {
	    sprintf (cbuf, "%10.4f %10.4f  in.", V.ZX/Inch, V.ZY/Inch);
	} else {
	    sprintf (cbuf, "%10.3f %10.3f  mm.", V.ZX, V.ZY);	// was %11.4g
	}

	c = strlen (cbuf);
	mx = cwx + 8;
	XDrawString (mwin.disp,win,tgc, mx, 64, cbuf, c);

/* If requested, cursor position is in lower left corner */
	if (df & 2) {
    // when dflag has a 2 in it, print the locbuf contents...
	    c = strlen (locbuf);
	    my = wy - 4;
	    XDrawString (mwin.disp,win,tgc, 7, my, locbuf, c);
	}	/* End cursor position section */
    }	/* End position writing section */

//	printf ("  -- Draw Window exiting.\n");  //debug only
//	fflush (stdout);

}


	/* - - - - - - - *\
	|                 |
	|   Plot Window   |
	|                 |
	\* - - - - - - - */

void	pwin (mlist* top, opts opt)
// was (Obs* obs)
/* Plot a standard window */
{


/* Local variables */
// int	dflag = 1;
int	dflag = 0;	// Only set it explicitly
XEvent		event;
XSizeHints	hint;
XWindowAttributes	xwa;
//char		buffer[100];
char		xserver[80];
char		title[80];
char		fontname[80];
int		sleeptime;
//int		i;
int	FontSize;
Window	xwx;

// debug things
int	sleepcount;
int	xnot;

/* Set up the time values */
    if (tsince (&previous, &current) != 0.0) previous = current;
    previous.tv_sec -= 9;

/* Get the imacs status first */
//    IMACS = isimacs (top);
//    if (IMACS) printf ("  -- IMACS mask detected. --\n");

/* Make up title string */
    sprintf (title, "%s (V%s)", PROGNAME, VERSION);

    if (opt.dbug) printf (" Starting %s\n", title);
// above satisfies lint for use of opt argument

/* Get a display */
    strcpy (xserver, genv2("DISPLAY","") );

    if ((app=CBX_InitXlib(xserver)) == NULL) {      /* connect to 'xserver' */
	fprintf(stderr,"%s: CBX_InitXlib(%s) failed\n", "Interface",xserver);
	exit(1);
    }
// copied from intgui...


/* Start up a window */
 
// We probably need these few things in the hint structure...
//    hint.min_width  = 400;    hint.min_height = 450;
//    hint.max_width  = 1000;   hint.max_height = 1050;
//    hint.flags = PPosition | PSize | PMinSize | PMaxSize;
// That does not seem to make any difference...


// 	Xloc(Hsize), Yloc(Vsize),		/* Window size */

    win = CBX_OpenMainWindow (
	&mwin, 
//	CBX_BOTTOMRIGHT, 
	CBX_MIDRIGHT, 
	wx, wy,		/* Window size */
	&hint, 
	title, I_title, 
	False);		// Allow window resize, not fixed...

    CBX_SelectInput (&mwin,
	ExposureMask | StructureNotifyMask | KeyPressMask |
	SubstructureNotifyMask | PropertyChangeMask |
	ButtonPressMask | ButtonReleaseMask );

// Actually we would really only need these:
//    CBX_SelectInput (&skywin, StructureNotifyMask |
//		ExposureMask  |  ButtonPressMask );

/* Graphics Contexts */
//    strcpy (fontname, "courier");
//    strcpy (fontname, "lucida sans typewriter");
    strcpy (fontname, "lucidatypewriter");
//    gc = CBX_CreateGC (black, white, fontname, "medium", 14);
    FontSize = 14;
//    (void)CBX_CreateGfC (&mwin, fontname, "medium", FontSize);
    (void)CBX_CreateGfC (&mwin, fontname, "bold", FontSize);

// Also set hint size things here.
    hint.min_width  = 400;    hint.min_height = 450;
    hint.max_width  = 1100;   hint.max_height = 1050;
    hint.flags = PPosition | PSize | PMinSize | PMaxSize;
    CBX_SetMainWindowProperties (&mwin, title, I_title, &hint);
// That's how we set the min/max things, manually.

    XGetWindowAttributes (mwin.disp, mwin.win, &xwa);

    set_sizes (&xwa);

/* Initialize scale and color of window */
    V.BX = (float)wx / mlextent(top);
    V.BY = -V.BX;
    V.ZX = 0.0;
    V.ZY = 0.0;
    V.color = app->ivory;
    XSetWindowBackground (mwin.disp, win, V.color);
// LATER, WE NEED DYNAMIC COLOR SET, DECIDED BY A MENU...


/* Create the buttons: in/out, rezero, done */

// Some xloc are set in "set_sizes" in cwplot...

// zoom in button
    MakeMButton (&butlist, &zifield);

// zoom out button
    MakeMButton (&butlist, &zofield);

// rezero button
    MakeMButton (&butlist, &rzfield);

// Done button
// Need dbfield to describe the button, and butlist to hold addresses.
    MakeMButton (&butlist, &dbfield);

// Debug print of button sizes...
#ifdef  DEBUG
    printf (" Zoom In  %4d %4d - %4d %4d\n",
	zifield.xloc, zifield.yloc,  zifield.width, zifield.height );
    printf (" Zoom Out %4d %4d - %4d %4d\n",
	zofield.xloc, zofield.yloc,  zofield.width, zofield.height );
    printf ("  Rezero  %4d %4d - %4d %4d\n",
	rzfield.xloc, rzfield.yloc,  rzfield.width, rzfield.height );
    printf ("   Done   %4d %4d - %4d %4d\n",
	dbfield.xloc, dbfield.yloc,  dbfield.width, dbfield.height );
#endif

/* Create the look angle entry fields */


// --- Separator old/new graphics ---

//    printf (" Window created; title: %s \n", title);
//    fflush (stdout);



/* --------------------Modify things-------------------- */



	/* - - - - - - - - - - *\
	|			|
	|      The  Loop	|
	|			|
	\* - - - - - - - - - - */

/* Loop for mouse and keyboard inputs */
    xnot = 0;
    while (!done) {
int	button;
#ifdef  LOOPDEBUG
int	mainwin;		//debug
#endif
	sleeptime = 50;
	sleepcount = 0;

	while (!CBX_AutoEvent(&mwin,&event,&sleeptime)) {
	    if (done) break;			/* break event loop */
//	    if (warnflag) break;   /* Get out if work to be done */
/* The above break conditionals are not needed and should be removed;
  if needed to exit due to an external event, simply pass the pointer
  to sleeptime to the inner loop handler as done for IFU_handle; it
  will get reduced if an event is detected... */
/*  The warning flag is set in textback and the button callback which
  are done automatically by AutoEvent.  They don't have a handle to
  the sleeptime, so they need another way to get out.  However, dflag
  is local to this program, and is probably not needed in this loop.  */
//dpause ("In auto event loop.");
	    CBX_Sleep(&sleeptime,50,300);
	    sleepcount++;
	    xnot = 0;		// Clear if we have slept at all.
	}	/* Wait for event */
	if (done) break; 


//	printf (" Slept %d times at %d ms. each.\n", sleepcount, sleeptime);
//	fflush (stdout);	// debug

#ifdef  LOOPDEBUG
	mainwin = (event.xany.window == win) ? 1 : 0;	// debug
	printf (" Event %2d s=%d m=%d x=%2d",
		event.type, event.xany.send_event,
		mainwin, xnot);	// debug
	if (event.type == 12) printf (" c=%2d", event.xexpose.count);
	printf ("\n");
	fflush (stdout);
#endif

	switch (event.type) {

	case KeyPress:		/* 2 */
// Actually, no key press events are defined for this window
	    break;

	case ButtonPress:	/* 4 */
// Handle all the manual buttons
	    button = event.xbutton.button;  // Basically, an abbreviation
#ifdef  SHFTCLK  // Was just: __APPLE__
/* For Apple computers, shift-click should be center click */
	if (event.xbutton.state & ShiftMask) button = 2;
#endif
	    xwx = event.xbutton.window;	// Basically, an abbreviation...
	    if (xwx == dunbut.win) done = True;
	    else if (xwx == zibut.win)
		dflag |= handle_scale (1, button);
	    else if (xwx == zobut.win)
		dflag |= handle_scale (0, button);
	    else if (xwx == rzbut.win)
//		dflag |= recenter (button);
		dflag |= recenter ();
	    else if (xwx == win)
		dflag |= handle_location (event.xbutton.x, event.xbutton.y,
				button);
	    else {	// debug
		printf (" Other button in window %ld\n", event.xbutton.window);
		fflush (stdout);
	    }		// debug
	if (dflag) bpressed++;
	if (dflag) dflag += 4 * 2;	// button press event 2
	    break;

	case ButtonRelease:	/* 5 */
	    xwx = event.xbutton.window;	// Basically, an abbreviation...
	    if ((xwx == zibut.win)  ||
		(xwx == zobut.win)  ||
		(xwx == rzbut.win)  ||
		(xwx == win)) {
//		dflag = 1;
		break;	// We don't do it locally
	    } else  printf (" Button Release detected.\n");
	    break;

	case MotionNotify:	/* 6 */
	    printf (" Motion notification.\n");
	    break;

	case EnterNotify:	/* 7 */
	    xwx = event.xcrossing.window;  // Basically, an abbreviation...
/* This and <8> seem not to be handled here, this code should be commented */
//	    if (xwx == win)
//		XSetWindowBackground (mwin.disp, xwx, app->white);
	    if ((xwx == dunbut.win) ||
		(xwx == zibut.win)  ||
		(xwx == zobut.win)  ||
		(xwx == rzbut.win)
		)
		XSetWindowBorder (mwin.disp, xwx, app->red);
	    break;

	case LeaveNotify:	/* 8 */
	    xwx = event.xcrossing.window;  // Basically, an abbreviation...
	    if ((xwx == dunbut.win) ||
		(xwx == zibut.win)  ||
		(xwx == zobut.win)  ||
		(xwx == rzbut.win) 
		)
		XSetWindowBorder (mwin.disp, xwx, app->black);
	    break;

	case Expose:		/* 12 */
//	    if (event.xexpose.count > 0) break;  /* ignore */
/* Debug - clear the xnot suppression if an expose is ignored */
	    if (event.xexpose.count > 0) { xnot=0; break; }  /* ignore */
	    if (event.xexpose.window == win) {	/* that's us */
//		dflag = 1;
/* Debug - only update screen when xnot is clear */
//		if (!xnot) dflag = 1;
		if ( (xnot == 0)  ||
			(tsince (&previous, &current) > 3.0)
			)  dflag = 1;
//		dflag += 4*8;	// debug expose event 8
	    }
	    break;

	case CreateNotify:	/* 16 */
/* Ignore?  Generated by propertychangemask or substructurenotifymask */
	    break;

	case DestroyNotify:	/* 17 */
	    if (event.xdestroywindow.window == win) done = True;
/* OK, but that doesn't break the loop when quit selected */
	    break;

	case UnmapNotify:	/* 18 */
//	    if (event.xmap.window == win) iconized = True;
	    break;

	case MapNotify:		/* 19 */
//	    if (event.xmap.window == win) iconized = False;
	    break;

	case ReparentNotify:	/* 21 */
		/* Ignored */
	    break;

	case ConfigureNotify:	/* 22 */
	    if (event.xconfigure.window == win) {	/* main window */
/* Find new config - attributes */
		xnot |= event.xconfigure.send_event;
		XGetWindowAttributes (mwin.disp, win, &xwa);
		set_sizes (&xwa);	// Sets button positions, center
//		MoveButton (&dbfield);	// "Done" tracks right side
//		MoveButton (&rzfield);	// "ReZero" is centered
// move is done in set_sizes now

/* Redraw the window */
// No, don't - this always occurs in combination with others,
// especially including an expose event.
//		dflag = 1;
//		dflag += 4*16;	// Configure notify debug 16

	    } else {
//		printf (" -- Configure other window?\n");  /* debug */
		;	/* Null action */
	    }
	    break;

/* Events needed/observed:
... See /usr/include/X11/X.h for these definitions ...
..*/

	case PropertyNotify:	/* 28 */
/* Ignore?  Generated by propertychangemask or substructurenotifymask */
	    break;

	case KillNotify:	/* 52 */
/* The window has been detected to have been closed... */
	    done = True;
	    break;

	default:
	    printf (" Other event '%2d'\n", event.type);
/* See /usr/include/X11/X.h for a definition of the event. */
	    fflush (stdout);
	    break;

	} /* End event switch */

/* Do global things which the switch set up (still in loop) */

/* Check if display needed */
	if (dflag) {
//	    XClearWindow (mwin.disp,win);	// Done in setup_window
	    setup_window ();	// Clear and draw buttons
//	    meta_setup ();	// Clear and draw buttons
	    draw_window (top, dflag, &opt);
	    dflag = 0;
	}

    }  /* main loop */

    CBX_CloseMainWindow (&mwin);

/* Clean up lists */
    while (winlist != NULL) winlist = wlpop (winlist);
    while (butlist != NULL) butlist = blpop (butlist);

/* All done */


}


/* main is copied from nctest.c... */

// Main program to get the options and file spec. for data file

        /* - - - - - - - - *\
        |                   |
        |    Main Program   |
        |                   |
        \* - - - - - - - - */

/* Example main from ~/imacs/maskgen/smdfps.c */

#define  GETOPTUSE
//#undef  GETOPTUSE
/*  Symbol  GETOPTUSE  is used to turn on (or off) the use of the
  standard "getopt" utility in parsing the command line.  Use this
  local define to enable the "old" method quickly if needed.  */


/*  ====  MAIN PROGRAM  ====  */

#undef  PROMPTLOOP
// #define  COMFILE  "Test01.nc"  // A sample file

int  main (int argc, char** argv) {
#ifdef  GETOPTUSE
extern char *optarg;
extern int optind,opterr,optopt;
int c;
#else
char    *pp;    /* Scan variable */
char    px;     /* Scan character */
int     i;
#endif
int     dohelp = 0;
char	infile [120];	/* For prompting */
opts	opt;
namlist	*head = NULL;
char	*filename;
int	k, m, n;	/* Process counts */
int	kr;
mlist	*mltop;
//char    configfile[] = "maskcut.cfg";
// Temproarily we get config data from a plot edition...
//char    configfile[]  = "ncplot.cfg";
//char    configfile2[] = "maskcut.cfg";
int	avplot = 0;
// namlist	*locs = NULL;
namlist	*comps=NULL;

#ifndef  sparc
/* In lieu of initialization */
    stdbugf = stderr;
#endif


/* Parameter is the command file; a default is provided */
/* Usual flags, debug and help */

    opt.dbug = 0;	/* Or - put data in define above */
    opt.rptm = 1;	/* Time feature */
    opt.rptz = 1;	/* Z excursion feature */
    opt.inch = 0;	/* Inch output feature */
    opt.cp = NULL;
    dodebug = 0;	/* NOTE = temporary for debugging */
    avplot = 0;		/* No avoidance plotting for ldss */

#ifdef  GETOPTUSE
/* Decode all the options of the command line */
/* :: follows letter with optional argument, : follows one with required
  argument, otherwise there are no arguments to the options.  */
    while((c = getopt(argc, argv, "ad::itzh")) != -1) switch(c) {
	case 'a':	avplot = 1;	break;  // Plot ldss avoids
	case 'd':				// debug level
	    opt.dbug = dodebug = optarg == NULL ? 1 : atoi (optarg);	break;
	case 'i':	opt.inch = 1;	break;  // Inch displays
	case 't':	opt.rptm = 0;	break;  // Omit report times
	case 'z':	opt.rptz = 0;	break;	// Report Z
	case 'h':	dohelp = 1;		break;	// parameter help
	default:
	    printf (" unused %c option\n", c);	break;
    }  /* end of switch, only statement in the getopt while loop */

    for (c = optind; c < argc; c++) {
	if (opt.dbug > 0) printf (" Adding %s to argument list.\n", argv[c]);
	head = pushtxt (argv[c], head);  /* Collect arguments */
    }

#else  // on GETOPTUSE
/* Scan for switches and arguments */
    for (i=1; i<argc; i++) {    /* argv[0] is command name */
	pp = argv[i];
	if (*pp == '-') {       /* Scan normal switch */
	    while ((px = *++pp) != AN) {
		if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
		if      (px == 'd')  dodebug = 0;  /* Debug output */
/* May want -d to cause debugging later? */
		else if (px == 'h')  dohelp = 1;  /* show help */
		else if (px == 'a')  avplot = 1;  /* Plot ldss avoids */
		else if (px == 't')  opt.rptm = 0;  /* Report times */
		else if (px == 'z')  opt.rptz = 0;  /* Report Z */
		else if (px == 'i')  opt.inch = 1;  /* Inch displays */
//		else if (px == 'w')  opt.pwin = 1;  /* window plot */
	    }
	} else if (*pp == '+') {        /* Scan positive switch */
	    while ((px = *++pp) != AN) {
		if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
		if      (px == 'd')  opt.dbug = dodebug = 1;      /* Debug output */
		else if (px == 'h')  dohelp = 1;  /* show help */
	    }
	} else if (*pp == '?') {
		dohelp = 1;
	} else {
#ifdef debug
	    printf ("Adding %s to list.\n", pp);  /* debug */
#endif
	    head = pushtxt (pp, head);	/* Collect arguments */
	}
    }
#endif  // on GETOPTUSE

// Following moved here from above the command line scan for debug to be set.
    if (get_configdata (M_imacs, opt.dbug)) return -1;

    if (dodebug && !dohelp)
	printf ("\n %s -- Version %s\n", PROGNAME, VERSION);

/* See if this version is any good, or if we continue.  */
//    comps = pushtxt (MASKCTT, comps);
    comps = pushtxt (NCPLOTT, comps);
    k = check_ver (comps, &(opt.cp), NULL);
/* Pointer values are NULL since we don't yet check for the source program
  and its version for our data file, xxx.nc.  */
    comps = kill_name (comps);
// if feature exists, we could exit on k < -2 here.
    if (k < -1) {
	printf ("  ** Operation of %s software rejected.\n",
		(k == -2) ? "Obsolete" : "Unknown" );
	kr = 1;
	goto main_finish;
    }

    if (dohelp) {
	printf ("\n %s -- Version %s", PROGNAME, VERSION);
	printf ("\n  Parameters are: ");
	printf ("\n    NC File name(s) ");
#ifdef  PROMPTLOOP
	printf ("\n    Inputs are prompted for if not given.");
#else
//	printf ("\n    Use \"%s\" if not given.", COMFILE);
	printf ("\n    Input is prompted for if not given.");
#endif
	printf ("\n   Switches");
#ifdef  GETOPTUSE
	printf ("\n     -d* = Set debug level to (*)");
#else
	printf ("\n     +d  = Some debug information");
#endif
	printf ("\n     -a  = Plot LDSS-2 avoids");
	printf ("\n     -i  = Show XY locations in inches");
	printf ("\n     -t  = omit report cutting times");
	printf ("\n     -z  = omit report of Z excursions");
//	printf ("\n     -p  = PostScript file needed");
//	printf ("\n     -w  = Plot in a window");
	printf ("\n     -h  = Display this help");
	printf ("\n     ");
	printf ("\n   Dump(s) written to standard output.");
/*	printf ("\n   Output file name is prompted for."); */
	printf ("\n \n");
	kr = 1;
	goto  main_finish;
//	return (1);
    }


//    opt.dbug = dodebug;

    if (avplot) pav = xav;	// Plot ldss avoids

/*  -- Loop to process arguments, or prompt if no arguments -- */

/* Insert a default file name here, so as to not need to prompt the
  user if no file name is given on command line. */
//    if (head == NULL) head = pushtxt (COMFILE, head);


    for (k=n=m=0; k >= 0; ) {	// Forever loop
	if (head == NULL) {	// No arguments, or looped out
#ifdef  PROMPTLOOP
	    if (m > 0) break;	// Had real arguments, looped out
#else
	    if (n > 0) break;	// No loop after any counted argument
#endif
	    askmestr (infile, 110, "Command File:");
	    if (*infile == AN) break;	// Exit on any null reply
	    filename = dynamstr (infile);
	} else {		// We have a command line argument given
	    filename = pop_name (&head);
	    if ( filename == NULL) continue;
	    if (*filename == AN) continue;
	    m++;	// Count number of command line arguments given
	}
	n++;		// Count number of arguments requested
//	k += smdftran (filename, opt);	// Count number processed
// The main subroutine call goes here
	printf (" Processing command file:  %s\n", filename);
	mltop = Data_Processor (filename, opt);
	if (mltop != NULL) {
	    k++;
//	    showmlq (mltop);
	    pwin (mltop, opt);
	}
	free (filename);	// MUST be dynamic
	while (mltop != NULL) mltop = freeml(mltop);
    }

/* All done, write a message about how it came out... */
    printf ("\n Processed %d of %d files.\n", k, n);
    kr = 0;

main_finish:
#ifdef  USE_CDAT
    CDhead = kill_cdat (CDhead);
#endif
#ifdef  USE_GSTO
    while (CDhead != NULL) CDhead = kill_gsto (CDhead);
#endif

#ifdef  USE_BSTO
    CDhead = btree_kill (CDhead, kill_kvm);
#endif
    opt.cp = kill_pvfs (opt.cp);
    return  kr;
}


/* Revision History:

    2011/10-26 -- Version 1.24
	-- On SHFTCLK, the feature that a center click of the
	mouse may be achieved by a shift-click is added.  This maps all
	clicks with shift held to center click.  This is not a 
	required minimum, but just a nice feature for some users.

    2008/10-17 -- Version 1.23
	-- Add support for "getopt" parsing of command line
	-- Put "getting mask cutting data" message on debug level 1
	-- Changed debug level prints

    2008/07-22 -- Version 1.22
	-- Decode multiple origin/version comments in nc file, with modified
	ncsource in ncputil.c and parsing in ncrutil.c.
	--> Minimum version at release 2.06 to handle new data

    2008/04-28 -- Version 1.21
	-- Add checksum file security support (in ncputil.c).

    2008/03-12 -- Version 1.20
	-- Add Program (Vers) structure for new mgutils requirement.
	-- Reject operation of obsolete software; warn on obsolete data
	-- Add decoding of source comments

    2008/02-08 -- Version 1.19
	-- Add support for WFCCD masks; fix report of mm/inch in min/max Z
	-- Add debug output of Acon value
	-- Add fix in ncrutil.c to reading of .nc file with malformed
	comment giving mask index before instrument name.
	--> This version will be minimum acceptable at release 2.04
	or any release implementing WFCCD

    2008/02-01 -- Version 1.18
	-- Add a -i flag for inch output of center and scale bar features.

    2007/10-26 -- Version 1.17
	-- Fix in ncputil.c code to make sure getconfigdata is done with a
	valid instrument index.  Otherwise we get very bad values and the
	times and Z values can go way wrong.  Also, detect the need to fix
	Acon, and give a message if that happens.
	-- Some debugs on DEBUG_PLT have been added here for future use.
	-- Moved version check above the dohelp display to allow for its
	output before the display of options.
	--> This version will be minimum acceptable at release 2.03

    2007/10-19 -- Version 1.16
	-- Add version control support.
	-- Minimum Version for release 2.0 and version control.

    2007/08-02 -- Version 1.15
	-- Fix mis-idendification of LDSS-3 instrument in ncputil.c
	-- Fixed some non application of ldss3 status

    2007/03-20 -- Version 1.14
	-- Increase digits of center location

    2006/10-24 -- Version 1.13
	-- Add support for gismo mask type

    2006/06-13 -- Version 1.12
	-- Add support for cdata table, with b-tree or gsto reading

    2006/05-18 -- Version 1.11
	-- Make ncrutil.c a separate .o file, refer to ncrutil.h only.

    2006/05-17 -- Version 1.10
	-- Reorgainze ncputil.c and ncrutil.c subroutines.

    2006/05-16 -- Version 1.08
	-- Get acceleration constant from configuration file.

    2006/05-11 -- Version 1.07
	-- Add acceleration parameter code to the time computation.

    2006/04-13 -- Version 1.06
	-- Change font to lucida-sans (lucidatypewriter), bold.

    2005/12-02 -- Version 1.05
	-- Includes new complot.c instead of cwplot.c, but that will
	be renamed soon.
	-- complot renamed back to cwplot.c

    2005/11-18 -- Version 1.03
	-- Support new method of configuration storage
	-- Add clear of dynamic configuration data

    2005/11-11 -- Version 1.02
	-- Put configuration read into mgutils

    2005/05-25 -- Version 1.01
	-- Omit plotting if data file is not found.

    2005/04-07 -- Version 1.00
	-- Change plotting to use new plot library.

    2005/01-28 -- Version 0.95
	-- Add parameter 4 for roughcut (edge of multi-masks)
	Also color brown for rough cut, and add to statistics.

    2004/08-13 -- Version 0.94
	-- Automatically look for xxxm.nc files in addition to
	the xxxi.nc files.  Look for root name first.
	-- Display warning message if metric measure found.

    2004/03-29 -- Version 0.93
	-- Allow M0 (program stop) code to be in a file

    2004/01-13 -- Version 0.90
	-- Enhance error messages to include line number in input file
	(in ncputil.c actually)...

    2003/10-09 -- Version 0.80
	-- Also look for maskcut.cfg if ncplot.cfg not found.

    2003/07-09 -- Version 0.75
	-- Fixing up configure file search

    2003/02-26 -- Version 0.70
	-- Add error messages for some nc programming problems
	-- Also, allow line numbering in .nc file

    2002/12-06 -- Version 0.60
	-- Fixed bug in vertical motion times
	-- Use new file find subroutine for diretory search

    2002/12-04 -- Version 0.55
	-- Introduce robust directory search for configuration file

    2002/12-02 -- Version 0.50
	-- Support M19, autofocus release command
	-- Better computation of autofocus miss distance

    2002/11-27 -- Version 0.45
	-- Support M18, autofocus hold command.

    2002/11-26 -- Version 0.40
	-- Change IMACS inner flange radius to 313, that of the
	mask, not the cutting fixture.

    2002/11-07 -- Version 0.38
	-- Always plot blue bound for LDSS mask
	-- Admit the -a flag in the help output

    2002/10-31 -- Version 0.35
	-- Make some comment output optional as extra debug.

    2002/8-30 -- Version 0.32
	-- Add decode of instrument comment
	-- Specific ldss avoidance detection

    2002/8-29 -- Version 0.30
	-- Add avoidance circles

    2002/8-26 -- Version 0.25
	-- Add time report feature

    2002/8-16 -- Version 0.20
	-- Finds .nc extension automatically
	-- No default file name, prompts for one

    2002/8-15 -- Version 0.15
	-- Decide colors based on status
	-- Detect IMACS mask automatically

    2002/8-14 -- Version 0.10
	-- Initial version

..*/
