/*  skymap.h  ==  Definitions for mapping  */

/* Copied from map.h from gmap project... */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2010 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef INCLUDE_SKYMAP_H
#define INCLUDE_SKYMAP_H

/* ---------------------------------------------------------------------- */

#ifdef IS_SKYMAP_C
#define EXT_SKYMAP 
#else
#define EXT_SKYMAP extern
#endif

/* DEFINEs -------------------------------------------------------------- */


/* INCLUDEs ------------------------------------------------------------- */

#include <X11/Xlib.h>

#include "catalog.h"	/* Includes kdcutil.h for vect3 */

/* TYPEDEFs ------------------------------------------------------------- */

/* Bit-map of mode values here... */
#define	GDR_POLAR	1
#define	GDR_SH		2

struct gstr {
	int	star;	/* Selection subscript */
	int	srad;	/* Star radius */
	int	x, y;	/* Window position, as plotted */
	float	xc, yc;	/* Positions relative to center */
	int	i;	/* Debug subscript */
};
typedef struct gstr gstr;

struct guiderdata {
	int	mode;		/* Bit-mapped */
	float	xc, yc;		/* Center offset */
	union	{
	    struct {
		float	xlo, xhi;
		float	ylo, yhi;
		float	xp, yp;
	    } rect_data;
	    struct {
		float	rlo, rhi;
		float	alo, ahi;
		float	rp, ap;
	    } polr_data;
	} position_u;
	float	xg, yg;		/* Guiding spot location */
	int	px;		/* Pixels on a side */
	float	psize;		/* Size of pixel interval */
	float	grot;		/* Camera rotation constant */
	float	scale;		/* Number of (x,y,r) units in radian */
	int	moving;		/* Flag for motion */

	int	sel;	/* Selection value */
	int	used;
	int	size;
	gstr	*tbl;

/* Magnitude limits for guider activity... */
	float	mag;
	float	bmag;

/* Data for tcs method position correction code (handle_ctbut) */
	float	bdec;	/* Box declination, radians */
	float	bdha;	/* Box h.a. from map_ca, radians */
	int	bsel;	/* Box selected flag */
};
typedef struct guiderdata guiderdata;

#ifdef  CACHECAT

// Otherwise defined for skymap.c only...
#define  Ncache  16
// Above goes with typedef for cache; below 2 defs only needed in skymap
#define  FNform  "OCIW_cat%c.dat"
#define  CacheDir  "CATCACHE"	// environment variable for directory.

/* Bits in cacheindx status */
#define  CI_exist (1 <<  0)	// set when file exists
#define  CI_good  (1 <<  1)	// set if header is good

struct	cacheindx {
	double	alpha;		// r.a. center in hours
	double	delta;		// dec. center in degrees
	double	frad;		// Radius in arc seconds
	int	cat;		// catalog subscript
	int	status;		// status present/empty/other
	time_t	date;		// file date  (st_ctime)
	int	stars;		// number of stars
	char	*filename;	// full file name
	};
typedef struct cacheindx cacheindx;
/* File date is set when the file is used in a skymap.  */

/* Bits in cachecon status */
#define  CS_init  (1 <<  0)	// set when initialized
#define  CS_skip  (1 <<  1)	// set to skip all caching

typedef	struct	cachecon {
	int	status;		// index filled/not filled; no caching; other;
	int	index;		// active index in ndx array
	cacheindx	ndx[Ncache];
	char	*directory;	// from environment variable or ./ default
	} cachecon;

#endif

/* GLOBALs -------------------------------------------------------------- */

EXT_SKYMAP          CATstar* maplist;
EXT_SKYMAP          double   map_ca,map_cd;
/* EXT_SKYMAP       vect3    map_xf[3]; */    /* Transform matrix */
EXT_SKYMAP          matrix3  map_xf;          /* Transform matrix */
EXT_SKYMAP          double   map_eq,cur_epoch;  /* Needed to show zenith */
EXT_SKYMAP          double   map_pa;
EXT_SKYMAP          int      map_nv; 		   /* Feature flags */
EXT_SKYMAP          int      map_uv, map_zd;	   /* Feature flags */
EXT_SKYMAP          int      map_pn;		   /* Feature flags */
EXT_SKYMAP          int      map_sc, map_cf;	   /* Feature flags */
EXT_SKYMAP          int      map_pg, map_sg, map_cg;  /* Feature flags */
EXT_SKYMAP	    int	     map_op, map_ap, map_kd;  /* Feature flags */
EXT_SKYMAP          int      map_savn;             /* Star saving */

EXT_SKYMAP volatile int      map_nstars,mapsize,map_sel;
EXT_SKYMAP volatile double   map_scale;
EXT_SKYMAP volatile int      map_w,map_h;
EXT_SKYMAP volatile double   map_pcx, map_pcy;

#ifdef  CACHECAT
EXT_SKYMAP  cachecon  Cache;
#endif

/* function prototype(s) */

// EXT_SKYMAP void   map_init             (Display*,double,Bool);
EXT_SKYMAP void   map_init             (MainWindow*,double,Bool);
EXT_SKYMAP void   map_free             (Display*);
// EXT_SKYMAP void   map_circle_star      (Display*,Window,int,float);
//sunifdef < #ifndef  ALLMSTAR
//sunifdef < EXT_SKYMAP int    map_getlist          (double,double,double,double,float);
//sunifdef < #else
EXT_SKYMAP int    map_getlist          (double,double,double,double);
//sunifdef < #endif
EXT_SKYMAP int    map_get_closest_star (int, int);
EXT_SKYMAP double map_dns              (int star);
EXT_SKYMAP long   map_get_x_coord      (double);
EXT_SKYMAP long   map_get_y_coord      (double);
// EXT_SKYMAP void   map_ad               (double*, double*, int, int);
EXT_SKYMAP int    map_get_radius       (float,float);
#ifdef  GSDATA
EXT_SKYMAP int    check_guidestar      (gpos* gs);
//void	validate_guiders (Obs* ob);
#endif
EXT_SKYMAP void   map_redraw           (Display*,Window,float);
#ifdef  GSDATA
// EXT_SKYMAP int    check_star           (int star);
EXT_SKYMAP void   map_guideset         (int star, int tog);
#ifdef  AUTOGS
EXT_SKYMAP int    get_autogs           (Obs* ob);
#endif
#endif
//EXT_SKYMAP double par_angle            (double, double, double);

EXT_SKYMAP vect2  xyproj               (vect3);
EXT_SKYMAP vect3  unproj               (vect2);

EXT_SKYMAP void	map_pltxy (int *x, int *y, vect2 v);
EXT_SKYMAP vect2	map_fpxy (int x, int y);

/* Temporary externals for testing */
// EXT_SKYMAP void  map_coords (int *x, int *y, double al, double de);

#ifdef  GUIDEWIN
EXT_SKYMAP void handle_gwinsel (Display*, Window, guiderdata*, int, int, int);

void  draw_guiderwin (Display*, Window, guiderdata*);
#endif

/* ---------------------------------------------------------------------- */

#endif  /* INCLUDE_SKYMAP_H */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
