/*  charplt.h == Character Plotting utility header */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2002 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef INCLUDE_CHARPLT_H
#define INCLUDE_CHARPLT_H

/*  #include	"kdcutil.h" */
#include	KDCUTILH

/*  ====  FUNCTION PROTOTYPES  ====  */

typedef void	drawing_subr (vect2 p, int pen, void* parameter);

/* Above allows the drawing subroutine to be prototyped correctly
  in the charplt package, as well as in user packages. */

/* To use this package, a drawing subroutine must be provided by
  the user.  The "parameter" is passed directly to this program,
  for use in describing plotting output unit or other things.
  The "pen" parameter may be -1, 0 or 1.  If -1, the position "P"
  is in absolute coordinates and represents a starting location.
  No drawing is done on this call.  Pen 0 is "up" and 1 is "down",
  and motion is to be from the last to this location with the pen
  either up or down.  These locations are relative to the last
  absolute location given with pen -1.  If vectors relative to
  the previous location are desired, the location must be retained
  in a local static varible in the subroutine. */


double	strwid (char* c);
/* Returns width of character "c" in units of the height
  of a capital character.  */


double	strplot (vect2 start,  vect2 base,
	double  curv,  double slant,
	char* string,  drawing_subr , void* parameter);
/* Start plotting at "start" in direction "base", which has length
  equal to 1 character height.  Curvature of plot line is "curv",
  and slant angle is "slant", limited to -15 to +30 degrees.  The
  data plotted is "string".  Use "drawing" program (user supplied),
  and pass the "parameter" pointer to the drawing program.
  Returns the total length of the data plotted.
.. */


#endif  /* INCLUDE_CHARPLT_H */





